package MyPAMLInfo;

#################################################
# Extract information from PAML programs runs   #
#           ------------------                  #
#   Written by Emanuel Maldonado - Mar 2015     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  MyPAMLInfo.pm module is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  MyPAMLInfo.pm is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MyPAMLInfo.pm is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with MyPAMLInfo.pm.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use strict;
use warnings;
use Data::Dumper;
use File::Spec::Functions qw(:ALL);

use MyUtil qw(:IPAML :CONST);

use constant {
    CODEML => 'codeml',
    CODEMLCTRLFLN => 'codeml.ctl', #this is dependent of CODEMLCTRLOPTF and vice-versa
    CODEMLDFLTOPTP => 'l:codeml', #mmap.pl default option p
    CODEMLCTRLOPTF => 'e:ctl', #mmap.pl default option f
    COMMONRESFLN => 'R', #default for mmap.pl  and imap.pl ;; must be same from templates files; if changed in templates (outfile); option -R must be always specified or changed here for default behaviour.
    PAMLTEMPLATES => 'templates',
    SM => 'SiteModels',
    SME => 'M012378',
    SMC => 'M8a',
    BSM => 'BranchSiteModels',
    BSMA => 'Alt',
    BSMN => 'Null',
    BM => 'BranchModels',
    BMA => 'TrC',
    BMN => 'M0',
    BMC => 'TrU',
    CM => 'CladeModels',
    CMA => 'CmC',
    CMN => 'M2a_rel',
};

use constant CODEML_OMEGA_VALS => qw(0.0 0.001 0.01 0.1 0.25 0.5 0.75 1 1.5 2);
use constant CODEML_KAPPA_VALS => qw(0.2 2 5);

BEGIN {
    require Exporter;

    our $VERSION = "1.0.0 Sep 25th, 2015";

    our @ISA = qw(Exporter);
    
    our @EXPORT = qw(&pinew &extract_cml_info &ei_sitemodels &ei_branchmodels &ei_branchsitemodels &ei_clademodels);

    our @EXPORT_OK = qw(CODEML COMMONRESFLN CODEMLCTRLFLN CODEMLCTRLOPTF CODEMLDFLTOPTP CODEML_OMEGA_VALS CODEML_KAPPA_VALS %C_MTYPESTAGS %C_MTYPESNAMES $C_CONTROLTMPL);
    our %EXPORT_TAGS = (
     	CODEML => [qw(&pinew &extract_cml_info &ei_sitemodels &ei_branchmodels &ei_branchsitemodels &ei_clademodels)],
	ICODEML => [qw(%C_MTYPESTAGS %C_MTYPESNAMES $C_CONTROLTMPL)],
	CONSTANTS  => [qw(CODEML COMMONRESFLN CODEMLCTRLFLN CODEMLCTRLOPTF CODEMLDFLTOPTP CODEML_OMEGA_VALS CODEML_KAPPA_VALS)], 
	);
};


our $OUTFILE;
our %Parameters;

our %C_MTYPESTAGS = ( c => CM, b => BM, w => BSM, s => SM );
our %C_MTYPESNAMES = ( c => [CMA,CMN], b => [BMN,BMA,BMC], w => [BSMA,BSMN], s => [SME,SMC] );
our $C_CONTROLTMPL = catdir($LMAPlibloc,PAMLTEMPLATES);


sub pinew {
    $OUTFILE = shift;
    if(-e $OUTFILE && -s $OUTFILE) {
	truncatefl($OUTFILE);
    }
    %Parameters = ();
}

#------------------------------------
sub extract_cml_info {
    my $ifl = shift @_;
    my $opt = shift @_;

    my @CTT = readac($ifl);

    if(!-e $ifl || -z $ifl || $CTT[-1] !~ m/Time\s+used/i) {
	print ":: Execution for file $ifl, has not finished! Please, check for errors.\n";
	return 0;
    }

    if($opt =~ m/^s:(.*)/) {
	ei_sitemodels($ifl,$1,\@CTT);
    } elsif($opt eq "b") {
	ei_branchmodels($ifl,\@CTT);
    } elsif ($opt eq "w") {
	ei_branchsitemodels($ifl,\@CTT);
    } elsif($opt eq "c") {
	ei_clademodels($ifl,\@CTT);
    }

    return 1;
}


#------------------------------------SM
#------------------------------------SM
sub ei_sitemodels {
    my $ifl = shift @_;
    my $opts = shift @_;
    my @CTT = @{$_[0]};

    my @selmodels = split(/,/,$opts);
    @selmodels = grep {!/8a/} @selmodels;
    @selmodels = sort {$a <=> $b} @selmodels;
    push @selmodels, "8a";

#    my %models = (0 => 1, 1 => 2,  2 => 3, 3 => 3, 7 => 10, 8 => 11, "8a" => 9); #model => n categories

    %Parameters = ();

    my @CSV;
    my $title = "File,Model,np,Parameters,treeLength,kappa,lnl,BEBsection";

    my $j = 0;

    my $r;
    my $fh = openOUTFLa($title."\n",0,$OUTFILE);
    for(my $j = 0 ; $j < @selmodels ; $j++) {
	for(my $i = 0 ; $i < @CTT ; $i++) {

	    if($selmodels[$j] eq "8a") {
		if($CTT[$i] =~ m/^Site-class\smodels:\s*beta\&w>1\s*/) {
		    $r = getSModelInfo("8a", $i+1, \@CTT);
		    $r =~ s/,$//;
		    push @CSV, $ifl.",".$r."\n";
		    last;		    
		}
	    } elsif($CTT[$i] =~ m/^Model\s+$selmodels[$j]:/) {
		$r = getSModelInfo($selmodels[$j], $i+1, \@CTT);
		$r =~ s/,$//;
		push @CSV, $ifl.",".$r."\n";
		last;
	    }
	}
    }
    print $fh @CSV;
    close $fh;
}

#------------------------------------SM
sub getSModelInfo {
    my $m = shift @_;
    my $i = shift @_;
    my @CTT = @{$_[0]};

    my ($lnl,$np,$tl,$k,@omega,@prop,$ompr,@BEB);

    $m = 9 if($m eq "8a");

    $ompr = "";
    for(my $j = $i ; $j < @CTT ; $j++) {
	if($CTT[$j] =~ m/^lnL\(ntime:\s*\d+\s+np:\s*(\d+)\):\s*(\-?\d+\.\d+).*/) {
	    $np = $1;
	    $lnl = $2;
	}
	if($CTT[$j] =~ m/^tree\s+length\s+=\s+(\d+\.\d+)/) {
	    $tl = $1;
	}
	if($CTT[$j] =~ m/^kappa\s+\(ts\/tv\)\s+=\s+(\d+\.\d+)/) {
	    $k = $1;
	}
    	if($m == 0) {
	    if($CTT[$j] =~ m/^omega\s+\(dN\/dS\)\s+=\s+(\d+\.\d+)/) {
		$ompr = "$1,n/a";
		$Parameters{$m} = "w0 = ".$1;
	    }
	} elsif($m == 1) {
	    if($CTT[$j] =~ m/^w:\s+(\d+\.\d+\s+\d+\.\d+)/) {
		@omega = split(/\s+/,$1);
	    }
	    if($CTT[$j] =~ m/^p:\s+(\d+\.\d+\s+\d+\.\d+)/) {
		@prop = split(/\s+/,$1);
	    }
	} elsif($m == 2) {
	    if($CTT[$j] =~ m/^w:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@omega = split(/\s+/,$1);
	    }
	    if($CTT[$j] =~ m/^p:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@prop = split(/\s+/,$1);
		@BEB = getBEB($j,\@CTT);
	    }
	} elsif($m == 3) {
	    if($CTT[$j] =~ m/^w:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@omega = split(/\s+/,$1);
	    }
	    if($CTT[$j] =~ m/^p:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@prop = split(/\s+/,$1);
	    }
	} elsif($m == 7) {
	    if($CTT[$j] =~ m/^Parameters\s+in\s+M$m/i) {
		$Parameters{$m} = $CTT[$j+1].$CTT[$j+2];
	    }
	    if($CTT[$j] =~ m/^w:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@omega = split(/\s+/,$1);
	    }
	    if($CTT[$j] =~ m/^p:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@prop = split(/\s+/,$1);
	    }
	} elsif($m == 8) {
	    if($CTT[$j] =~ m/^Parameters\s+in\s+M$m/i) {
		$Parameters{$m} = $CTT[$j+1].$CTT[$j+2];
	    }
	    if($CTT[$j] =~ m/^w:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@omega = split(/\s+/,$1);
	    }
	    if($CTT[$j] =~ m/^p:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@prop = split(/\s+/,$1);
		@BEB = getBEB($j,\@CTT);
	    }
	} elsif($m == 9) { #M8a
	    if($CTT[$j] =~ m/^Parameters\s+in\s+M8\s+.*/i) {
		$Parameters{'8a'} = $CTT[$j+1].$CTT[$j+2];
	    }
	    if($CTT[$j] =~ m/^w:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@omega = split(/\s+/,$1);
	    }
	    if($CTT[$j] =~ m/^p:\s+(\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+\s+\d+\.\d+)/) {
		@prop = split(/\s+/,$1);
	    }
	}

	last if(@omega && @prop || $ompr);
    }
    if(@omega && @prop) {
	    $ompr .= "$omega[$_],$prop[$_]," foreach (0..scalar(@prop)-1);
	    if(!defined $Parameters{$m}) {
		$Parameters{$m} .= "w$_ = $omega[$_] p$_ = $prop[$_] " foreach (0..scalar(@prop)-1);
	    }
    }

    $m = "8a" if($m == 9);

    my $beb = "";
    $beb = "\"".join("\n",@BEB)."\"" if(@BEB);
    $beb =~ s/,/;/g;
    $Parameters{$m} =~ s/\s*//;

    return "M$m,$np,$Parameters{$m},$tl,$k,$lnl,$beb";
}


#------------------------------------SM
sub getBEB {
    my $j = shift @_;
    my @CTT = @{$_[0]};

    my @beb;
    for(my $i = $j+1; $i < @CTT; $i++) {
	if($CTT[$i] =~ m/^Bayes\s+Empirical\s+Bayes/) {
	    push @beb, $CTT[$i++] while($CTT[$i] !~ m/^The\s+grid/);
	    last;
	}
    }

    my $yes = 0;
    foreach (@beb) { #confirm there are sites...
	if($_ =~ m/\s+\d+\s+\w.*/) {
	    $yes =1;
	    $_ =~ s/^\s+//;
	    $_ =~ s/\s+$//;
	    $_ =~ s/\s+/;/g;
	}
    }
    if(!$yes) {
	return ();
    }

    pop @beb if($beb[-1] =~ m/^\s*/); 

    return @beb;
}


#------------------------------------BM
#------------------------------------BM
sub ei_branchmodels {
    my $ifl = shift @_;
    my @CTT = @{$_[0]};

    my @CSV;

    my ($i, $np, $tl, $k, $lnl, @proportions, %sclasses);
    $i = 0;
    $sclasses{"0"} = [];
    $sclasses{"1"} = [];
    $sclasses{"2a"} = [];
    $sclasses{"2b"} = [];

    my @lin0 = grep {/Model:/} @CTT;
    my $model;
    my $beb = "";

    if ($lin0[0] =~ m/One.*ratio.*/) {
	$model = BMN;	
    } if ($lin0[0] =~ m/several.*ratios.*omega.*fixed/) {
	$model = BMA;
    } elsif($lin0[0] =~ m/several.*ratios.*/) {
	$model = BMC;
	if(my @BEB = getBEB(0,\@CTT)) {
	    $beb = "\"".join("\n",@BEB)."\"";
	    $beb =~ s/,/;/;
	}
    }

    my $ompr;
    my $omegas;
    foreach my $line (@CTT) {
	if($line =~ m/^lnL\(ntime\:\s*\d+\s+np\:\s*(\d+)\)\:\s*(\-?\d+\.\d+).*/) {
	    $np = $1;
	    $lnl = $2;
	}
	if($line =~ m/^tree\s+length\s+=\s+(\d+\.\d+)/) {
	    $tl = $1;
	}
	if($line =~ m/^kappa\s+\(ts\/tv\)\s+=\s+(\d+\.\d+)/) {
	    $k = $1;
	}

	if($model eq BMN) {
	    $ompr = "$1" if($line =~ m/^omega\s+\(dN\/dS\)\s+=\s+(\d+\.\d+)/);
	} elsif($model eq BMA) {
	    $omegas = "$1" if($line =~ m/^w\s+\(dN\/dS\)\s+for\s+branches:((\s+(\d+\.\d+))+)/);
	} elsif($model eq BMC) {
	    $omegas = "$1" if($line =~ m/^w\s+\(dN\/dS\)\s+for\s+branches:((\s+(\d+\.\d+))+)/);
	}
    }

    if($omegas) {
        $omegas =~ s/^\s+//g;
        my @ooomm = split(/\s/,$omegas);
        foreach (0..scalar(@ooomm)-1) {
            $ompr .= "w$_ = $ooomm[$_];" if($ooomm[$_]);
        }
    }

    $ompr =~ s/;$//;
    my $fh = openOUTFLa("File,Model,np,treeLength,kappa,omegas,lnl\n",0,$OUTFILE); #no BEB here
    push @CSV, "$ifl,$model,$np,$tl,$k,$ompr,$lnl\n";
    print $fh @CSV;
    close $fh;

}


#------------------------------------BSM
#------------------------------------BSM
sub ei_branchsitemodels {
    my $ifl = shift @_;
    my @CTT = @{$_[0]};

    my @CSV;

    my ($i, $np, $tl, $k, $lnl, @proportions, %sclasses);
    $i = 0;
    $sclasses{"0"} = [];
    $sclasses{"1"} = [];
    $sclasses{"2a"} = [];
    $sclasses{"2b"} = [];

    my @lin0 = grep {/Model:/} @CTT;
    my @lin1 = grep {/Site-class\s+models:/} @CTT;
    my $model;
    my $beb ="";

    if ($lin0[0] =~ m/several.*omega.*fixed/) {
	$model = BSMN;
    } elsif($lin0[0] =~ m/several\s+dN\/dS\s+ratios/) {
	$model = BSMA;
	if(my @BEB = getBEB(0,\@CTT)) {
	    $beb = "\"".join("\n",@BEB)."\"";
	    $beb =~ s/,/;/;
	}
    }

    foreach my $line (@CTT) {
	if($line =~ m/^lnL\(ntime\:\s*\d+\s+np\:\s*(\d+)\)\:\s*(\-?\d+\.\d+).*/) {
	    $np = $1;
	    $lnl = $2;
	}
	if($line =~ m/^tree\s+length\s+=\s+(\d+\.\d+)/) {
	    $tl = $1;
	}
	if($line =~ m/^kappa\s+\(ts\/tv\)\s+=\s+(\d+\.\d+)/) {
	    $k = $1;
	}

	if($line =~ m/^(proportion)/) {
	    $line =~ s/proportion\s+//;
	    @proportions = split(/\s+/,$line);
	} elsif ($line =~ m/^(background\s+w)/) {
	    $line =~ s/background\s+w\s+//;
	    my @btype = split(/\s+/,$line);
	    push @{$sclasses{"0"}}, $btype[0];
	    push @{$sclasses{"1"}}, $btype[1];
	    push @{$sclasses{"2a"}}, $btype[2];
	    push @{$sclasses{"2b"}}, $btype[3];
	    $i++;
	} elsif ($line =~ m/^(foreground\s+w)/) {
	    $line =~ s/foreground\s+w\s+//;
	    my @btype = split(/\s+/,$line);
	    push @{$sclasses{"0"}}, $btype[0];
	    push @{$sclasses{"1"}}, $btype[1];
	    push @{$sclasses{"2a"}}, $btype[2];
	    push @{$sclasses{"2b"}}, $btype[3];
	    $i++;
	}
    }

    my (@sc0, @sc1, @sc2a, @sc2b) = ((),(),(),());
    @sc0 = reverse @{$sclasses{"0"}} if(defined $sclasses{"0"}); 
    @sc1 = reverse @{$sclasses{"1"}} if(defined $sclasses{"1"}); 
    @sc2a = reverse @{$sclasses{"2a"}} if(defined $sclasses{"2a"}); 
    @sc2b = reverse @{$sclasses{"2b"}} if(defined $sclasses{"2b"}); 

    my $fh = openOUTFLa("File,Model,np,treeLength,kappa,(SC0) fg w,(SC0) bg w,(SC0) p,(SC1) fg w,(SC1) bg w,(SC1) p,(SC2a) fg w,(SC2a) bg w,(SC2a) p,(SC2b) fg w,(SC2b) bg w,(SC2b) p,lnl,BEBsites\n",0,$OUTFILE);
    push @CSV, "$ifl,$model,$np,$tl,$k,".join(",",@sc0).",$proportions[0],".join(",",@sc1).",$proportions[1],".join(",",@sc2a).",$proportions[2],".join(",",@sc2b).",$proportions[3],$lnl,$beb\n";
    print $fh @CSV;
    close $fh;

}




#------------------------------------CM
#------------------------------------CM
sub ei_clademodels { #CmC or CmD and M1a and/or M2a_rel
    my $ifl = shift @_;
    my @CTT = @{$_[0]};

    my @CSV;

    my ($i, $np, $tl, $k, $lnl, @proportions, %sclasses, $beb);
    $i = 0;
    $sclasses{0} = [];
    $sclasses{1} = [];
    $sclasses{2} = [];

    my  @lin0 = grep {/Model:/} @CTT;
    my  @lin1 = grep {/Site-class\s+models:/} @CTT;
    my $model;

    if($lin0[0] =~ m/NSbranch3\s+/) {
	if($lin1[0] =~ m/\s+PositiveSelection/) { 
	    $model = CMA;
	} elsif ($lin1[0] =~ m/\s+discrete\s+.*/ ) {
	    $model = "CmD";
	}
    } elsif ($lin0[0] =~ m/One\s+dN\/dS\s+ratio\s+/) {
	if($lin1[0] =~ m/NearlyNeutral/) {
	    $model = "M1a";
	} elsif ($lin1[0] =~ m/\s+M2a_rel\s+/) {
	    $model = CMN;
	}  elsif ($lin1[0] =~ m/\s+discrete\s+/) {
	    $model = "M3";
	}
    }

    $beb = "";
    $beb = "\"".join("\n",getCMBEB($ifl))."\"" if($model eq CMA);

    foreach my $line (@CTT) {
	if($line =~ m/^lnL\(ntime\:\s*\d+\s+np\:\s*(\d+)\)\:\s*(\-?\d+\.\d+).*/) {
	    $np = $1;
	    $lnl = $2;
	}
	if($line =~ m/^tree\s+length\s+=\s+(\d+\.\d+)/) {
	    $tl = $1;
	}
	if($line =~ m/^kappa\s+\(ts\/tv\)\s+=\s+(\d+\.\d+)/) {
	    $k = $1;
	}

	if($model eq CMA || $model eq "CmD") {
	    if($line =~ m/^(proportion)/) {
		$line =~ s/proportion\s+//;
		@proportions = split(/\s+/,$line);
	    } elsif ($line =~ m/^(branch\s+type\s+\d+)/) {
		$line =~ s/branch\s+type\s+\d+:\s+//;
		my @btype = split(/\s+/,$line);
		push @{$sclasses{0}}, $btype[0];
		push @{$sclasses{1}}, $btype[1];
		push @{$sclasses{2}}, $btype[2];
		$i++;
	    }
	} elsif($model eq "M1a" || $model eq CMN || $model eq "M3") {
	    if($line =~ m/^p\:/) {
		$line =~ s/^p\:\s+//;
		@proportions = split(/\s+/,$line);
	    } elsif($line =~ m/^w\:/) {
		$line =~ s/^w\:\s+//;
		my @btype = split(/\s+/,$line);
		push @{$sclasses{0}}, $btype[0] if(defined $btype[0]);
		push @{$sclasses{1}}, $btype[1] if(defined $btype[1]);
		push @{$sclasses{2}}, $btype[2] if(defined $btype[2]);
		$i++;
	    }
	}
    }

    my (@sc0, @sc1, @sc2) = ((),(),());
    @sc0 = @{$sclasses{0}} if(defined $sclasses{0}); 
    @sc1 = @{$sclasses{1}} if(defined $sclasses{1}); 
    @sc2 = @{$sclasses{2}} if(defined $sclasses{2}); 

    my $fh = openOUTFLa("File,Model,np,treeLength,kappa,(SC0) w0,(SC0) p0,(SC1) w1,(SC1) p1,(SC2) wX,(SC2) pX,lnl,BEBsites\n",0,$OUTFILE);
    push @CSV, "$ifl,$model,$np,$tl,$k,".join("/",@sc0).",$proportions[0],".join("/",@sc1).",$proportions[1],".join("/",@sc2).",$proportions[2],$lnl,$beb\n";
    print $fh @CSV;
    close $fh;

}


#------------------------------------CM
sub getCMBEB {
    my $fl = shift @_;

    my $inputfl = (splitpath($fl))[2];

    $fl =~ s/($inputfl)$/rst/;
    my @CTT = readac($fl);

    my $j = 0;
    my @beb;
    for( ; $j < @CTT ; $j++) {
	if($CTT[$j] =~ m/^Bayes\s+Empirical\s+Bayes.*/) {
	    last;
	}
    }

    my $yes= 0;
    push @beb, $CTT[$j++];
    push @beb, $CTT[$j];
    for(my $i = $j; $i < @CTT; $i++) {
	if($CTT[$i] =~ m/3\)$/) {
	    $yes = 1;
	    my $bebline = $CTT[$i];
	    if($bebline =~ m/\s+\d\s+\w.*/) {
		$bebline =~ s/^\s+//;
		$bebline =~ s/\s+$//;
		$bebline =~ s/\s+/;/g;
	    }
	    push @beb, $bebline;
	}
    }
    if(!$yes) { #if no sites, then, empty!
	return ();
    }

    return @beb;
}

END {};

1;
