package MyPhylo;

#################################################
# Enable tree editing for PAML hypothesis       #
#           ------------------                  #
#   Created by Emanuel Maldonado - Mar 2015     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  MyPhylo.pm module is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado

#  MyPhylo.pm is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.

#  MyPhylo.pm is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with MyPhylo.pm.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use strict;
use warnings;
use Data::Dumper;
use Bio::TreeIO;
use IO::String;
use File::Path qw(make_path);
use File::Spec::Functions qw(:ALL);

use constant {
    DOWN => '/',
    UP => '\\',
    CONN => '|',
    REND1 => ":",
    REND2 => "|",
    SPACE => " ",
    BLS => 'b',
    BL => 'bl',
    BS => 'bs',
    DFLT_TREEDIR => 'lmap_savedtrees',
    NEWICK => 'newick',
    BRLEN => 100,
    INODELEN => 2
};

BEGIN {
    require Exporter;

    our $VERSION = "1.0.0 Nov 19th, 2015";

    our @ISA = qw(Exporter);

    our @EXPORT = qw(&pnew &draw_ascii &isTreeRooted &branchLabelNodeID &writeTreeToFile &writeTreeAsTxt2File &writeUnrTree2File &gsBrLength &gsINodelength &modifyDfltTreeDir &removeBootstraps &removeBranchlengths &nodesverify &renameTips &getNumNodesLabeled &pruneTree &rootUnrTree &rerootTree);

    our @EXPORT_OK = qw(BLS BL BS DFLT_TREEDIR);

    our %EXPORT_TAGS = (
	DRAWTREE => [qw(&pnew &draw_ascii &isTreeRooted &branchLabelNodeID &writeTreeToFile &writeTreeAsTxt2File &writeUnrTree2File &gsBrLength &gsINodelength &modifyDfltTreeDir &removeBootstraps &removeBranchlengths &nodesverify &renameTips &getNumNodesLabeled &pruneTree &rootUnrTree &rerootTree)],
	CONSTANTS => [qw(BLS BL BS DFLT_TREEDIR)],
	);
};


# exported package globals
our $TFILE;

our $SHOW_BL;
our $SHOW_BS;

our $_dirdone;

our %_NodeLevel = ();
our @_INodeOrder = ();
our %_PrintedNodes = ();

our $_treeio;
our $_tree;
our $_treeasstr;

our $_RootNodeID;
our $_InodeStrLen;

our $_dlftTreeDir;

our $_initBlen = BRLEN;
our $_initINodelen = INODELEN;
our $_blenDec;
our $_InodeFactor;
our $_maxlenblen = 0;
our $_maxlenblent;
our $_blsinc = 3; #max 3 chars for boots (#\d\d)

our @_LoRows = ();

sub pnew {  
    $TFILE = shift;
    _reset();

    $_treeio = Bio::TreeIO->new(-format => NEWICK, -file => $TFILE);
    $_tree = $_treeio->next_tree();

    _getsetBS();

    $SHOW_BS = $SHOW_BL = 0;
    $_InodeStrLen = 0;
    $_RootNodeID = "";
    $_dirdone = 0;
    $_dlftTreeDir = DFLT_TREEDIR;
    _inodeDistance(1);
}

sub draw_ascii {
    my $show = shift;
    my $bls = 0; #account for colon
    if(defined $show) {
	$SHOW_BS = $show eq BS ? 1 : 0;
	$SHOW_BL = $show eq BL ? 1 : 0;
	if($show eq BLS) {
	    $bls = $SHOW_BL = $SHOW_BS = 1;
	}
    }

    if($SHOW_BS && $SHOW_BL) {
	$_maxlenblen += $_blsinc;
	$_blsinc = 0;
    } else {
	$_maxlenblen = $_maxlenblent;
	$_blsinc = 3;
    }

    my $rootnode = $_tree->get_root_node;
    $_RootNodeID = $rootnode->internal_id();

    @_INodeOrder = ();
    push @_INodeOrder, $_RootNodeID;
    _inodeOrder($rootnode,1);

    %_NodeLevel = ();
    $_NodeLevel{$_RootNodeID} = 0;
    _nodelevels($rootnode,1);

    %_PrintedNodes = ();
    my ($sbs,$sbl) = ($SHOW_BS,$SHOW_BL);
    $SHOW_BS = $SHOW_BL = 0;
    $_InodeStrLen = length(_getINodeToPrint($rootnode))+$bls; #invariable length!
    ($SHOW_BS,$SHOW_BL) = ($sbs,$sbl);
    @_LoRows = ();
    _draw_tree($rootnode,0,$_initBlen);
    _draw_connections();
}

sub gsBrLength {
    my $blen = shift;

    $_initBlen = $blen if(defined $blen);
    
    return $_initBlen;
}

sub gsINodeLength {
    my $inlen = shift;

    $_initINodelen = $inlen if(defined $inlen);

    return $_initINodelen;
}

#every tree has a root, even unrooted trees by design;
#the only difference relies in the number of descendensts (dichotomy ->rooted ; trichotomy -> unrooted)
sub isTreeRooted {
    return 1 if(scalar(($_tree->get_root_node())->each_Descendent()) == 2);
    return 0;
}

sub rootUnrTree {
    my $nodes = shift;
    
    $nodes =~ s/[AL]0*//g;
    my @nodes = split(/,/,$nodes);
    my $root = $_tree->get_root_node(); 

    my $check = 0;
    foreach my $nid (@nodes) {
	foreach my $rch ($root->each_Descendent()) {
	    if($rch->internal_id() == $nid) {
		$check++;
	    }
	}
    }
    return "Could not root tree: specified nodes are not 'root' (A"._formatNodeID($root).") descendents!" if($check != 2);

    my $newnode = Bio::Tree::Node->new();
    $newnode->add_Descendent(_getNodeWithID($nodes[0]));
    $newnode->add_Descendent(_getNodeWithID($nodes[1]));
    $newnode->ancestor($root);
    _updatetree(0);
    return "Tree is now rooted!";
}

sub pruneTree {
    my $nodes = shift;
    
    my @nodes = ($nodes);
    @nodes = split(",",$nodes) if($nodes =~ m/,/);    
    
    my $cnt = 0;
    foreach my $nid (@nodes) {
	if($nid =~ m/L/) { #removeLeaf
	    $nid =~ s/L0*//;
	    for my $n ( $_tree->get_nodes ) {
		if($n->is_Leaf() && $n->internal_id() == $nid) {
		    if(defined (($n->ancestor())->ancestor())) {
			_reconnNodes($n,$nid);
		    } else {
			_rerootUndAnc($n,$nid);
		    }
		    ($n->ancestor())->remove_Descendent($n); #rem leaf
		    $cnt++;
		    last;
		}
	    }
	} else { #removeAncestor all descendents
	    $nid =~ s/A0*//;
	    for my $n ( $_tree->get_nodes ) {
		if($n->internal_id() == $nid ) {
		    if(defined (($n->ancestor())->ancestor())) {
			_reconnNodes($n,$nid);
		    } else {
			_rerootUndAnc($n,$nid);
		    }
		    $n->remove_all_Descendents();
		    $cnt++;
		    last;
		}
	    }
	}
    }
    _updatetree(0);
    return "Removed $cnt nodes!" if($cnt);
    return "Node does not exist!";
}

sub _reconnNodes { #connect sister (leaf/anc) node to the ancestral of the ancestral
    my $n = shift;
    my $nid = shift;
    my @snodes = grep {$_->internal_id() != $nid} ($n->ancestor())->each_Descendent(); #sister nodes of the target to be removed
    foreach (@snodes) {
	(($n->ancestor())->ancestor())->add_Descendent($_); #set new anc child (sister of target)
	$_->ancestor(($n->ancestor())->ancestor()); #set child (sister of target) new anc
    }
    (($n->ancestor())->ancestor())->remove_Descendent($n->ancestor()); #rem target ancestor
}

sub _rerootUndAnc {
    my $n = shift; #node with undef grandfather
    my $nid = shift;
    
    my @snodes = grep {$_->internal_id() != $nid} ($n->ancestor())->each_Descendent(); #sister nodes of the target to be removed

    my $newroot = shift @snodes;
    my $newrootid = $newroot->internal_id();
    my $oldroot = $_tree->get_root_node();

    for my $n ( $_tree->get_nodes ) { #remove oldroot having one single descendent (Ax-ancestral)
	if($n->internal_id() == $newrootid ) {
	    $oldroot->remove_Descendent($n);
	    $n->add_Descendent($oldroot);
	    $_tree->set_root_node($n);
	    $n->remove_Descendent($oldroot);
	    last;
	}
    }
}

sub getTreeUnrooted {
    my @Basalnodes = (); #trichotomy
    my $lin = 0;
    my $ain = 0;
    my $childValues;
    my $root = $_tree->get_root_node();
    my @rootChildren = $root->each_Descendent();
    my %map = _determineChildrenMap(@rootChildren); #trichotomy - nodes to employ
    @rootChildren = sort {$map{$a->internal_id()} <=> $map{$b->internal_id()}} @rootChildren; #sort increasing placing more importance in nodes with map = 1 than 2

    #collect basal nodes
    foreach my $rootch (@rootChildren) {
	if($map{$rootch->internal_id()} == 0) {  #=== root child is leaf
	    push @Basalnodes, $rootch;
	} else {                                 #=== root child is ancestor which has...
	    if($map{$rootch->internal_id()} == 1) {                       #one leaf and one clade... or
		$childValues = _getValuesInfo($rootch);
		push @Basalnodes, $rootch->each_Descendent();
	    } elsif($map{$rootch->internal_id()} == 2) {                  #two clades
		push @Basalnodes, $rootch if(@Basalnodes + 1 == 3);
		push @Basalnodes, $rootch->each_Descendent() if(@Basalnodes + 2 == 3);
	    }
	}
    }

    #get subtrees
    my @unrootedtree = ();
    foreach (@Basalnodes) {
	if($_->is_Leaf()) {
	    local $SHOW_BS = 1; local $SHOW_BL = 1;
	    push @unrootedtree, $_->id()." ".(_getValuesInfo($_));
	} else {
	    $_tree->set_root_node($_); #cause $_tree->clone($node) seems not to work correcly (gives always undef)
	    local $_treeasstr = "";
	    _updatetree(1); #1 to also account for labelings
	    my $subtree = $_treeasstr;
	    $subtree =~ s/;$//;
	    push @unrootedtree, $subtree; #as_text requer Bio::Tree::TreeI e $_tree  desse tipo
	}
    }

    $_tree->set_root_node($root);
    _updatetree(1);
    return "(".join(",",@unrootedtree).")".($childValues eq ":" ? "" : $childValues).";";
}

sub _determineChildrenMap {
    my @rchildren = @_;

    my $i = 0;
    my $mem = 0;
    my $lin = 0; #leaf
    my $ain = 0; #ances
    my %brmap = ();
    my %map = ();
    foreach my $rootch (@rchildren) {
	my $riid = $rootch->internal_id();
	if($rootch->is_Leaf()) {
	    $map{$riid} = 0;
	    ++$i;
	    next;
	}
	foreach my $rootGrSon ($rootch->each_Descendent()) {
	    if($rootGrSon->is_Leaf()) {
		local $SHOW_BL = 1; local $SHOW_BS = 0;
		$brmap{$riid} = _getValuesInfo($rootGrSon) || $riid; #see exceptional case 3: save brlen; in case there is no brlen on tree then there is no logical way of deciding, hence any decision may make sense.
		$lin = 1;
	    } else { 
		$ain++;
	    }
	}
	if($mem + $ain == 4) { #in exceptional case 1: were both rootchildren have all descendents as clades (Ntaxa <= 4 clades)
	    $ain = 1;
	} 
	elsif($ain == 0 && $i) { #in exceptional case 2: were both rootchildren have all descendents as leafs (4taxa <= 4 leafs) ($lin = 1 x2[$i=0,1]) => ($ain ==0)
	    $ain = -1;
	} 
	elsif($mem + $lin == 2 && $i) { #in exceptional case 3: were both rootchildren are identical i.e. both falling in "category 1": one leaf and one clade. Here decision is made based on leafs maximum brlen.
	    my @riids = sort {$brmap{$b} <=> $brmap{$a}} keys %brmap; # determine which of the 2 rootch has the leaf with the max brlen 
	    $map{$riids[0]} = 1; #max: falls in category 1: assume this rootch as the best choice for category 1 (leaf and clade)
	    $map{$riids[1]} = 2; #min: falls in category 2: assume this rootch as ancestor of 2 clades
	    $ain = $map{$riid};  #ensure below map assignment is congruent with these last assignments/updates
	}
	$map{$riid} = 2 if($ain > 1);               #category 2: node in pos $i has 2 children not leafs
	$map{$riid} = 1 if($ain == 1 || $ain == 0); #category 1: node in pos $i has 1 children not leafs [ or 2 children leafs (exceptional case 1) ]
	$map{$riid} = 0 if($ain < 0);               #category 0: exceptional case 2
	$mem = $map{$riid};
	$lin = $ain = 0;
	$mem = 0 if($i); #after 2 iter
	$i++;
    }
    return %map;
}

sub renameTips {
    my $tips = shift;
    
    my $cnt = 0;
    if($tips =~ m/\w+\.csv/) { #rename by CSV file OLDname,NEWname
	if(-e $tips && -s $tips && -f $tips) {
	    open(my $FT, "<", $tips) or return "Could not open specified CSV file! : $!";
	    my @filetips = <$FT>;
	    close $FT;
	    shift @filetips; #no title
	    chomp @filetips;
	    
	    if(my $hm = _somearenotTreeNames(\@filetips))  {
		return "Found $hm taxa not found in tree, please verify your CSV file!";
	    }
	    if(my $hm = _repeatedNamesFromUser(0,\@filetips)) {
		return "Found $hm repeated names in 2nd column, please verify your CSV file!";
	    }
	    if(my $hm = _repeatedNamesInAll(0,\@filetips)) {
		return "Found $hm repeated names compared to tree tips, please verify your CSV file!";
	    }

	    foreach (@filetips) {
		my ($tipname,$destname) = split(/,/,$_,2); #comma separated values!
		for my $node ( $_tree->get_nodes ) { #remove from savings...
		    if(defined $node->id()) {
			if($node->id() eq $tipname) {
			    $node->id($destname);
			    $cnt++;
			}
		    }
		}
	    }
	} else {
	    return "Could not open specified file! Please check your CSV file.";
	}
    } else { #rename by Node tags
	$tips =~ s/L0*//g;
	my @tipsnames = ();
	$tipsnames[0] = $tips;
	if(index($tips, ",") != -1) {
	    @tipsnames = split(/,/,$tips); #tip
	}
	if(my $hm = _repeatedNamesFromUser(1,\@tipsnames)) {
	    return "Found $hm repeated names, please verify your entry!";
	}
	if(my $hm = _repeatedNamesInAll(1,\@tipsnames)) {
	    return "Found $hm repeated names compared to tree tips, please verify your entry!";
	}
	foreach my $tipname (@tipsnames) {
	    my ($tip,$name) = split(/=/,$tipname);
	    for my $node ( $_tree->get_nodes ) { #remove from savings...
		if(defined $node->id()) {
		    if($node->is_Leaf() && $node->internal_id() == $tip) {
			$node->id($name);
			$cnt++;
		    }
		}
	    }
	}
    }
    _updatetree(0);
    return "Changed $cnt tips names!";
}

sub _somearenotTreeNames { #verify if all sources specified are in tree
    my @filetips = @{$_[0]};

    my @tnames;
    foreach (@filetips) {
	push @tnames, (split(/,/,$_))[0]; 
    }

    my @inTree;
    foreach (@tnames) {
	for my $node ( $_tree->get_nodes ) {
	    if(defined $node->id()) {
		if($node->is_Leaf ) {
		    if($node->id() eq $_) {
			push @inTree, $_;
		    }
		}
	    }
	}
    }

    return @tnames - @inTree;
}

sub _repeatedNamesFromUser { #verify if all dest names are not repeated
    my $mode = shift;
    my @tipsnames = @{$_[0]};
    
    my @names;
    foreach (@tipsnames) {
	push @names, (split(/=/,$_))[1] if($mode); 
	push @names, (split(/,/,$_))[1] if(!$mode); 
    }
    my %check = map { $_ => 1 } @names;
    my $c = scalar(keys %check);
    return abs(@names - $c);
}

sub _repeatedNamesInAll { #verify if all dest names are not repeated in tree
    my $mode = shift;
    my @tipsnames = @{$_[0]};
    
    my @names;
    foreach (@tipsnames) {
	push @names, (split(/=/,$_))[1] if($mode); 
	push @names, (split(/,/,$_))[1] if(!$mode);
    }

    my @tnames;
    for my $node ( $_tree->get_nodes ) {
	if(defined $node->id()) {
	    if($node->is_Leaf) {
		push @tnames, $node->id();
	    }
	}
    }

    my @allnames = (@names,@tnames);
    my %check = map { $_ => 1 } @allnames;
    my $c = scalar(keys %check);
    return abs(@allnames - $c);
}

sub removeBootstraps {
    for my $node ( $_tree->get_nodes ) {
	$node->bootstrap("");
    }   

    for my $node ( $_tree->get_nodes ) { 
	my $bl = $node->branch_length();
	next if(!$bl);
	if($bl =~ m/^(\d+(\.\d+)?)$/) {                        # 'BL'   branch length -The one and only
	} elsif($bl =~ m|^(\s*[\$\#]\d+):$|) {                 # 'L:'   (=>boots label) (anc or leaf by space)
	    $node->branch_length(undef);
	} elsif($bl =~ m|^(\d+(\.\d+)?):$|) {                  # 'N:'   (=>boots num)
	    $node->branch_length(undef); 
	} elsif($bl =~ m|^\s*([\$\#]\d+):(\d+(\.\d+)?)$|) {    # 'L:BL' (=>boots label) (anc or leaf by space)
	    $node->branch_length($2);
	} elsif($bl =~ m|^(\d+(\.\d+)?):(\d+(\.\d+)?)$|) {     # 'N:BL' (=>boots num)
	    $node->branch_length($3); 
	}
    }

    _updatetree(0);
}

sub removeBranchlengths {
    for my $node ( $_tree->get_nodes ) {
	my $bl = $node->branch_length();
	next if(!$bl);
	if($bl =~ m/^(\d+(\.\d+)?)$/) {                        # 'BL'   branch length -The one and only
	    $node->branch_length(undef);
	} elsif($bl =~ m|^(\s*([\$\#]\d+):)$|) {               # 'L:'   (=>boots label) (anc or leaf by space)
	    $node->branch_length($1);
	} elsif($bl =~ m|^((\d+(\.\d+)?):)$|) {                # 'N:'   (=>boots num)
	    $node->branch_length($1);
	} elsif($bl =~ m|^(\s*([\$\#]\d+):)(\d+(\.\d+)?)$|) {  # 'L:BL' (=>boots label) (anc or leaf by space)
	    $node->branch_length($1);
	} elsif($bl =~ m|^((\d+(\.\d+)?):)(\d+(\.\d+)?)$|) {   # 'N:BL' (=>boots num)
	    $node->branch_length($1);
	}
    }
    _updatetree(0);
}

sub nodesverify {
    my $nodes = shift;

    if($nodes =~ m/A\d+(,A\d+)?/) {
	$nodes =~ s/A0*//g;
	my %alnodes = map { $_ => 1} split(/,/,$nodes);

	for my $node ( $_tree->get_nodes ) {
	    if($node->is_Leaf()) {
		if(exists $alnodes{$node->internal_id()}) {
		    return (0,$node->internal_id(),'ancestral');
		}
	    }
	}
    } elsif($nodes =~ m/L\d+(,L\d+)?/) {
	$nodes =~ s/L0*//g;
	my %alnodes = map { $_ => 1} split(/,/,$nodes);

	for my $node ( $_tree->get_nodes ) {
	    if(!$node->is_Leaf()) {
		if(exists $alnodes{$node->internal_id()}) {
		    return (0,$node->internal_id(),'leaf (or tip)');
		}
	    }
	}
    }

    return (1,0,0);
}

sub branchLabelNodeID {
    my $nodeTag = shift;
    my $brlabel = shift;

    my $numnodes = 0;
    my $leafs = 0;
    $nodeTag =~ s/A0*//;
    $leafs = 1 if($nodeTag =~ m/,?L/g);
    $nodeTag =~ s/L0*//;
    my $blen = "";

    #boostrap() for display purposes
    #branch_length() for setting, saving and displaying values
    for my $node ( $_tree->get_nodes ) {
	$blen = $node->branch_length();
	$blen =~ s/ ?[\#\$]\d+:?//; #... to replace labels
	if($nodeTag == $node->internal_id()) {
	    if($blen) {
		$node->bootstrap($brlabel); #does not save value...
		if($node->is_Leaf) {
		    $brlabel = " ".$brlabel.":".$blen;
		} else {
		    $brlabel = $brlabel.":".$blen;
		}
		$node->branch_length($brlabel);
	    } elsif($node->is_Leaf) {
		$node->branch_length(" ".$brlabel.":");
		$node->bootstrap($brlabel); #does not save value...
	    } else {
		$node->branch_length($brlabel.":");
		$node->bootstrap($brlabel); #does not save value...
	    }
	    $numnodes++;
	}
    }
 
    _updatetree(1);   
    return $numnodes;
}

sub getNumNodesLabeled {
    my $cnt = 0;
    for my $node ( $_tree->get_nodes ) {
	if(defined $node->bootstrap) {
	    if($node->bootstrap =~ m/[\#\$]\d+/) {
		$cnt++;
	    }
	}
    }
    return $cnt;
}

sub modifyDfltTreeDir {
    my $newtd = shift;
    $_dlftTreeDir = $newtd ? $newtd : DFLT_TREEDIR;
    $_dirdone = 0;
    return $_dlftTreeDir;
}

sub writeUnrTree2File {
    my $filename = shift;
    my $unrooted = getTreeUnrooted();
    _createDfltTreeDir() if(!$_dirdone);
    my $fl = catfile($_dlftTreeDir,$filename);
    open(FH,">",$fl) or die "ERR::MyPhylo:: Could not open file for writing: $!";
    print FH $unrooted,"\n";
    close FH;
    return $fl;   
}

sub writeTreeAsTxt2File {
    my $filename = shift;
    _createDfltTreeDir() if(!$_dirdone);
    my $fl = catfile($_dlftTreeDir,$filename);
    open(FH,">",$fl) or die "ERR::MyPhylo:: Could not open file for writing: $!";
    print FH $_treeasstr,"\n";
    close FH;
    return $fl;
}


sub writeTreeToFile {
    my $filename = shift;
    _createDfltTreeDir() if(!$_dirdone);
    my $fl = catfile($_dlftTreeDir,$filename);
    my $out = new Bio::TreeIO(-file => ">$fl", -format => NEWICK);
    $out->write_tree($_tree);
    return $fl;
}

sub _updatetree {
    my $doalso = shift;

    $_treeasstr = $_tree->as_text(NEWICK);
    $_treeasstr =~ s/:(\d+(\.\d+)?):([,\)])/$1$3/g; #orig boots nums no blens
    $_treeasstr =~ s/:(\d+(\.\d+)?)(:\d+(\.\d+)?)/$1$3/g; #orig boots nums with blens
    $_treeasstr =~ s/:(\s*[\$\#]\d+):?([,\)])/$1$2/g; # removed blens ; keep only labels; anc and leafs (nospace and space respect.)

    if($doalso) {
	$_treeasstr =~ s/:(\s?[\$\#])/$1/g; #right after labeling; anc and leafs (nospace and space respect.)
	$_treeasstr =~ s/(\w+)([\#]\d+)/$1 $2/g; #make space for labels in leafs
    }
        
    my $iostr = IO::String->new($_treeasstr);
    $_treeio = Bio::TreeIO->new(-fh => $iostr,
				-format => NEWICK);
}

sub _reset {
    undef $_dirdone;
    
    undef %_NodeLevel = ();
    undef @_INodeOrder = ();
    undef %_PrintedNodes = ();
    
    undef $_treeio;
    undef $_tree;
    undef $_treeasstr;
    
    undef $_RootNodeID;
    undef $_InodeStrLen;
}

sub _getsetBS() {

    for my $node ( $_tree->get_nodes ) {
	if($node->id()) {
	    if($node->id() =~ m/(\w*)\s*([\$\#]\d+)/) { #bootstraps as labels
		$node->id($1); #keep name/id only
		$node->bootstrap($2);
		$node->branch_length($2.":".$node->branch_length());
	    } elsif ($node->id() =~ m/^(\d+(\.\d+)?)$/ ) { #boostraps as numbers
		$node->id("");
		$node->bootstrap($1);
		my $brlen = defined $node->branch_length() ? $node->branch_length() : "";
		$node->branch_length($1.":".$brlen);
	    }
	}
	my $bl = $node->branch_length();
	if($bl) {
	    if(length($bl) > $_maxlenblen) {
		$_maxlenblen = length($bl);
	    }
	}
    }
    $_maxlenblent = ++$_maxlenblen; #plus ':' char
}

sub _createDfltTreeDir {
    $_dlftTreeDir =~ s:\/$::;
    make_path($_dlftTreeDir,{mode=>0777}); #make full path from currdir
    print ":: Directories created!\n";
    $_dirdone = 1;
}

sub _inodeOrder {
    my $rnode = shift;
    my $level = shift;

    foreach my $node ( $rnode->each_Descendent ) { 
	if($node->is_Leaf) {
	    my $done = 0;
	    do {
		$node = $node->ancestor;
		if($node) {
		    if($node->internal_id() != $_RootNodeID) {
			if(!exists $_PrintedNodes{$node->internal_id()}) {
			    push @_INodeOrder, $node->internal_id();
			    $_PrintedNodes{$node->internal_id()} = 1;
			    $done = 1;
			}
		    }
		}
	    } while (!$done && defined $node);
	} else {
	    _inodeOrder($node,$level+1);
	}
    }
}

sub _nodelevels {
    my $rnode = shift;
    my $level = shift;
    foreach my $node ( $rnode->each_Descendent ) { 
	$_NodeLevel{$node->internal_id()} = $level;
	_nodelevels($node,$level+1);
    }
}

sub _draw_tree {
    my $rnode = shift;
    my $level = shift;
    my $blen = shift;

    foreach my $node ( $rnode->each_Descendent ) {
	if($node->is_Leaf) {
	    my $n = _formatNodeID($node);
	    my $lnode = _getLNodeToPrint($node);
	    if($n == $_RootNodeID -1) {
		push @_LoRows,
		sprintf("%s%s%s%s%s%s%s%s%s%s","|","-"x($_initBlen-$blen+$_InodeStrLen-1),"-","-"x($blen-$_InodeStrLen)," ","L".$n,$lnode,"-",$node->id(),"\n"); #exception case
	    } else {
		push @_LoRows,
		sprintf("%s%s%s%s%s%s%s%s%s%s","|"," "x($_initBlen-$blen+$_InodeStrLen-1),"|","-"x($blen-$_InodeStrLen)," ","L".$n,$lnode,"-",$node->id(),"\n");
	    }

	    $_PrintedNodes{$node->internal_id()} = 1;
	    my $done = 0;
	    do {
		$node = $node->ancestor;
		if($node) {
		    if($node->internal_id() != $_RootNodeID) {
			if(!exists $_PrintedNodes{$node->internal_id()}) {
 			    my $level = $_NodeLevel{$node->internal_id()} * $_InodeFactor;
			    my $inode = _getINodeToPrint($node);
			    $inode = " ".$inode if($SHOW_BS && $SHOW_BL);
			    push @_LoRows,
			    sprintf("%s%s%s%s%s","|"," "x($level/2)," "x($level*2),$inode,"\n");
			    $_PrintedNodes{$node->internal_id()} = 1;
			    $done = 1;
			}
		    } else {
			if(!exists $_PrintedNodes{$node->internal_id()}) {
			    my $inode = _getINodeToPrint($node);
			    push @_LoRows, sprintf("%s%s",$inode,"\n");
			    $_PrintedNodes{$node->internal_id()} = 1;
			    $done = 1;
			}
		    }
		}
	    } while (!$done && defined $node);
	} else {
	    _draw_tree($node,$level+1,$blen-$_blenDec);
	}
    }
}

sub _draw_connections {
    my @LoL = ();
    foreach (@_LoRows) {
	push @LoL, [split("",$_)];
    }
    my $trow;
    for(my $row = 0 ; $row < @LoL ; $row++) {
	for(my $col = 0; $col < @{$LoL[$row]}; $col++) {
	    if($LoL[$row][$col] eq DOWN || $LoL[$row][$col] eq UP) {
		if(($trow = _traversingDownHitsNode($row+1, $col, \@LoL))) { #first check then do
		    $LoL[$row][$col] = DOWN;
		    @LoL = _writeConDown($trow-1,$row+1,$col, \@LoL);
		    last;
		} else {
		    $LoL[$row][$col] = UP;
		    @LoL = _writeConUp($row-1,$col,\@LoL);
		    last;
		}
	    }
	}
    }
    @LoL = _processRoot(\@LoL);
    _printTree(\@LoL);
}

sub _printTree {
    my @LoL = @{$_[0]};
    for(my $r = 0 ; $r < @LoL ; $r++) {
	print join("",@{$LoL[$r]});
    }
}

sub _processRoot {
    my @LoL = @{$_[0]};

    for(my $row = 0 ; $row < @LoL ; $row++) {
	if($LoL[$row][$#{$LoL[$row]}-1] eq REND1 || $LoL[$row][$#{$LoL[$row]}-1] eq REND2) { #root end?
	    my $col = $#{$LoL[$row]}-1;
	    my $rhypnode = _getHypNode($row,$col/2,\@LoL);
	    if($rhypnode =~ m:[\|-]+(A\d+)[\|-]+:) { #yes root!
		if(defined $LoL[$row-1][0]) { #is there row before?
		    my $hypnodeUp = _getHypNode($row-1,$col,\@LoL);
		    if($hypnodeUp =~ m:\s+|-+: && $LoL[$row-1][$col] eq SPACE && $LoL[$row-1][$col+1] eq CONN) {
			$LoL[$row-1][$col] = DOWN;
			$LoL[$row][$col] = DOWN;			
		    }
		}
		if(defined $LoL[$row+1][0]) { #is there row after?
		    my $hypnodeDn = _getHypNode($row+1,$col,\@LoL);
		    if($hypnodeDn =~ m:\s+|-+: && $LoL[$row+1][$col] eq SPACE && $LoL[$row+1][$col+1] eq CONN) {
			$LoL[$row+1][$col] = UP;
			$LoL[$row][$col] = UP;			
		    }
		}
	    }
	}
    }

    return @LoL;
}

sub _writeConDown {
    my $trow = shift;
    my $srow = shift;
    my $scol = shift;
    my @LoL = @{$_[0]};
    
    for(my $r = $srow ; $r < @LoL ; $r++) {
	if($LoL[$r][$scol] eq SPACE) {
	    $LoL[$r][$scol] = CONN;
	}
	last if($r == $trow);
    }
    return @LoL;
}

sub _writeConUp {
    my $row = shift;
    my $col = shift;
    my @LoL = @{$_[0]};

    my $hypnode;
    my $margin = 4; #for joining around the hypnode
    for(my $r = $row ; $r >= 0 ; $r--) {
	if($LoL[$r][$col] eq SPACE) {
	    $LoL[$r][$col] = CONN;
	}
	$hypnode = join("",(@{$LoL[$r]}[($col-$margin)..($col+$margin)]));
	last if($hypnode =~ m:[\|-]+(A\d+)[\|-]+:);
    }
    return @LoL;
}

sub _traversingDownHitsNode {
    my $row = shift;
    my $col = shift;
    my @LoL = @{$_[0]};
    my $margin = 4; #for joining around the hypnode    
    my $hypnode;
    for(my $r = $row ; $r < @LoL ; $r++) {
	last if(!defined $LoL[$r][$col]);
	next if($LoL[$r][$col] eq SPACE);
	my $start = $col - $margin;
	my $end = $col + $margin;
	$end = $col if(!defined $LoL[$r][$end]);
	$start = $col if(!defined $LoL[$r][$start]);
	$hypnode = join("",(@{$LoL[$r]}[$start..$end]));
	if($hypnode =~ m/[\|-]+(A\d+)[-\|]+/) {
	    return $r;
	} else {
	    return 0;
	}
    }
    return 0;
}

sub _getHypNode {
    my $row = shift;
    my $col = shift;
    my @LoL = @{$_[0]};
    my $margin = 4; #for joining around the hypnode    
    my $start = $col - $margin;
    my $end = $col + $margin;
    $end = $col if(!defined $LoL[$row][$end]);
    $start = $col if(!defined $LoL[$row][$start]);
    return "" if(!defined @{$LoL[$row]}[$start..$end]);
    my $hnode = join("",(@{$LoL[$row]}[$start..$end]));
    return $hnode;
}

sub _getNodeWithID {
    my $nid = shift;

    for my $n ( $_tree->get_nodes ) {
    	if($n->internal_id() == $nid ) {    
	    return $n;
    	}
    }

    return 0;   
}

sub _inodeGetAppChar {
    my $currnode = shift;

    for(my $i =0, my $j =1; $j < @_INodeOrder ; $i++, $j++) {
	if($_INodeOrder[$i] == $currnode) {
	    if($_INodeOrder[$i] < $_INodeOrder[$j]) {
		return DOWN;
	    } else {
		return UP;
	    }
	}
    }
    return UP;
}

sub _getINodeToPrint {
    my $node = shift;

    my $n = _formatNodeID($node);
    my $char = _inodeGetAppChar($node->internal_id());
    $char = $node->internal_id() == $_RootNodeID ? '|' : $char;
    my $show_info = _getValuesInfo($node);
    my $inode = $char.("-"x$_initINodelen).'|A'.$n.'|'.("-"x$_initINodelen).'|'.$show_info;
    return $inode;
}

sub _getLNodeToPrint {
    my $node = shift;

    my $n = _formatNodeID($node);
    my $show_info = _getValuesInfo($node);
    my $format = " (%".$_maxlenblen."s) ";
    my $lnode = $show_info ? sprintf($format,$show_info) : "";
    return $lnode;
}


sub _getValuesInfo {
    my $node = shift;

    my $tbrlen = $SHOW_BL ? (defined $node->branch_length() ? $node->branch_length() : '') : '';
    if($tbrlen =~ m/:?(\d+(\.\d+)?)$/) { #BL, L:BL, N:BL
	$tbrlen = $1;
    } elsif($tbrlen =~ m/^([\$\#]\d+)|(\d+(\.\d+)?):$/) { #L: , N:
	$tbrlen = "";
    }
    my $tboots = $SHOW_BS ? (defined $node->bootstrap() ? $node->bootstrap() : '') : '';
    my $show_info = $SHOW_BL && $SHOW_BS ? ($tboots.":".$tbrlen) : $SHOW_BL ? $tbrlen : $SHOW_BS ? $tboots : '';    
    return $show_info;
}

sub _inodeDistance {
    my $incfactor = shift;
    
    $incfactor = 1 if(!defined $incfactor);

    $_blenDec = 5 * $incfactor; #both need be at same proportion 5*2 then 2*2 5*3 then 2*3
    $_InodeFactor = 2 * $incfactor;
}

sub _formatNodeID {
    my $node = shift;
    my $lenrid = length($_RootNodeID);
    return sprintf("%0".$lenrid."d",$node->internal_id());
}

END {}

1;



