package MyUtil;

#################################################
# Extract information from PAML programs runs   #
#           ------------------                  #
#   Written by Emanuel Maldonado - Feb 2015     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  MyUtil.pm module is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  MyUtil.pm is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MyUtil.pm is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with MyUtil.pm.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use strict;
use warnings;
use FindBin;
use Data::Dumper;
use MIME::Base64;
use Term::ANSIColor qw(:constants); 
use File::Spec::Functions qw(:ALL);

use constant {
    PREFSFL => ".lmapp",
    OPAMLPREFS => ".omapd",
    LMAPCOOKIE => ".lmapcooky",
    BIN => "bin",
    LMAPLIB => "LMAPlib",
    CITATION => "Maldonado E., Almeida D., Escalona T., Khan I., Vasconcelos V. and Antunes A. (2016)\n    LMAP: Lightweight Multigene Analyses in PAML\n    BMC Bioinformatics, 17:354.\n    DOI: 10.1186/s12859-016-1204-5\n",
    PROJWEBSITE => "HOME PAGE: http://lmapaml.sourceforge.net"
};

BEGIN {
    require Exporter;

    our $VERSION = "1.0.0 Mar 22nd, 2017";

    our @ISA = qw(Exporter);

    our @EXPORT = qw(&unew &getPFL &readPFL &getPBinPath &getPTBL &getPKeys &wantColor &enableTCOLOR &disableTCOLOR &getDirContent &getPAMLRTDefsFl &clear &readac &readacd &openOUTFLa &openOUTFLb &getonecd &list_directory &getcd &printToFile &cleanOpts &printLicense &setCookyLoc &modifyLMAPCooky &readLMAPCooky &modifyLMAPCooky2 &truncatefl);

    our @EXPORT_OK = qw($LMAPlibloc LMAPLIB $HOMEBINPATH CITATION PROJWEBSITE);

    our %EXPORT_TAGS = (
	LMAP => [qw(&unew &wantColor &enableTCOLOR &disableTCOLOR &cleanOpts &getPBinPath &printLicense &setCookyLoc &modifyLMAPCooky)],
	GPAML => [qw(&unew &wantColor &enableTCOLOR &disableTCOLOR &readacd &readac &clear &list_directory &cleanOpts &printLicense)],
	CPAML => [qw(&unew &wantColor &enableTCOLOR &disableTCOLOR &readacd &readac &cleanOpts &getDirContent &printToFile &printLicense)],
	RPAML => [qw(&unew &getPBinPath &wantColor &enableTCOLOR &disableTCOLOR &getDirContent &clear &cleanOpts &printLicense &setCookyLoc &readLMAPCooky &modifyLMAPCooky2)],
	IPAML => [qw(&unew &wantColor &enableTCOLOR &disableTCOLOR &getDirContent &readac &openOUTFLa &printToFile &cleanOpts &printLicense &truncatefl)],
	OPAML => [qw(&unew &getPBinPath &getPAMLRTDefsFl &clear &readac &openOUTFLb &wantColor &enableTCOLOR &disableTCOLOR &list_directory &cleanOpts &printLicense)],
	NOTIFY => [qw(&getPFL &readPFL &getPTBL &printLicense)],
	INST => [qw(&getPKeys &getPFL &getonecd &getPTBL &getcd &printLicense)],
	CONST => [qw($LMAPlibloc LMAPLIB $HOMEBINPATH CITATION PROJWEBSITE)]
	);
};


our $TERMCOLOR = 0;
our $OTHERPAMLBINPATH = 0;
our $LMAPlibloc = catdir("$FindBin::Bin",LMAPLIB);
our $HOMEBINPATH = catdir($ENV{HOME},BIN,"");

our %_TCOLOR = ();
our $_PFFL = catfile($ENV{HOME},PREFSFL);
our $_PRTDEFS = catfile($ENV{HOME},OPAMLPREFS);
our $_DFLTPAMLBINPATH = $HOMEBINPATH;
our %_EPREFTBLKeys = (
    h => "SMTPHOSTNAME",
    p => "SMTPPORT",
    e => "SMTPHELO",
    s => "SMTPSSL",
    u => "SMTPAUTH_USER",
    w => "SMTPAUTH_PASSWD",
    f => "FIXEDNOTIFYTO",
    );
our %_GPREFTBLKeys = (
    b => "OTHERPAMLBINPATH",
    c => "USETERMCOLOR",
    );

our $_cookyloc = "";
our @_FDONE = ();

sub unew {
    _initTCOLOR();
    _readGP();
}

sub getPAMLRTDefsFl {
    return $_PRTDEFS;
}

sub getPTBL {
    return (%_GPREFTBLKeys,%_EPREFTBLKeys);
}

sub getPFL {
    return $_PFFL;
}

sub getPBinPath {
    return $OTHERPAMLBINPATH ? $OTHERPAMLBINPATH : $_DFLTPAMLBINPATH;
}

sub getPKeys {
    return (values %_GPREFTBLKeys, values %_EPREFTBLKeys);
}

sub wantColor {
    return $TERMCOLOR;
}

sub enableTCOLOR {
    _initTCOLOR();
    return %_TCOLOR;
}

sub disableTCOLOR {
    foreach (keys %_TCOLOR) {
        $_TCOLOR{$_} = "";
    }
    return %_TCOLOR;
}

sub _initTCOLOR {
    $_TCOLOR{underln} = UNDERLINE;
    $_TCOLOR{bold} = BOLD;

    $_TCOLOR{qunln} = BRIGHT_RED UNDERLINE; #quiting screens and options
    $_TCOLOR{chscunln} = BRIGHT_RED UNDERLINE; #change screens task/run options
    $_TCOLOR{menu} = BOLD GREEN; #menu colors, bars, header..
    $_TCOLOR{tblttl} = BRIGHT_CYAN;

    $_TCOLOR{bggreen} = BLACK ON_GREEN;
    $_TCOLOR{bgwhite} = BLACK ON_WHITE;
    $_TCOLOR{bgcyan} = BLACK ON_CYAN;
    $_TCOLOR{Bred} = BOLD RED;
    $_TCOLOR{Bcyan} = BOLD CYAN;
    $_TCOLOR{Blgreen} = BOLD BRIGHT_GREEN;
    $_TCOLOR{Blblue} = BOLD BRIGHT_BLUE;
    $_TCOLOR{red} = RED;
    $_TCOLOR{green} = GREEN;
    $_TCOLOR{blue} = BLUE;
    $_TCOLOR{cyan} = CYAN;
    $_TCOLOR{yellow} = YELLOW;
    $_TCOLOR{magenta} = MAGENTA;
    $_TCOLOR{lmagenta} = BRIGHT_MAGENTA;
    $_TCOLOR{lcyan} = BRIGHT_CYAN;
    $_TCOLOR{lyellow} = BRIGHT_YELLOW;
    $_TCOLOR{lgreen} = BRIGHT_GREEN;
    $_TCOLOR{lred} = BRIGHT_RED;
    $_TCOLOR{lblue} = BRIGHT_BLUE;
    $_TCOLOR{endcolor} = RESET;
    $_TCOLOR{eopen} = $_TCOLOR{eclose} = ""; #empty
}

sub _readGP {
    my @PREF = split(/\n/,readPFL());
    my $i = 2;
    foreach (@PREF) {
	if(!$TERMCOLOR && $_ =~ m/^($_GPREFTBLKeys{c})/) {
	    $TERMCOLOR = (split("=",$_,2))[1];
	    $TERMCOLOR =~ s/\<no\>/0/;
	    $TERMCOLOR =~ s/\<yes\>/1/;
	    $i--;
	}
	if(!$OTHERPAMLBINPATH && $_ =~ m/^$_GPREFTBLKeys{b}/) {
	    $OTHERPAMLBINPATH = (split("=",$_,2))[1];
	    $OTHERPAMLBINPATH =~ s/^\<no\>\s*//;
	    chomp $OTHERPAMLBINPATH;
	    $i--;
	}
	last unless($i);
    }
}

sub clear {
    my $usecase = shift;
#http://apple.stackexchange.com/questions/31872/how-do-i-reset-the-scrollback-in-the-terminal-via-a-shell-command
    if(!$usecase) {
	system("tput clear ; printf '\ec\e[3J'"); #screen and scrollback
	if($? == -1) {
	    die "ERR:: An error occured during the execution of terminal clear program: $!\n";
	} elsif($? & 127) {
	    printf "ERR:: terminal clear unexpectedly terminated with signal %d, %s coredump\n",($? & 127),($? & 128) ? 'with':'without';
	}
    } elsif ($usecase == 1) { #debugging
	system("tput clear");
	if($? == -1) {
	    die "ERR:: An error occured during the execution of clear program: $!\n";
	} elsif($? & 127) {
	    printf "ERR:: clear unexpectedly terminated with signal %d, %s coredump\n",($? & 127),($? & 128) ? 'with':'without';
	}
    }
}

sub readac {
    my $inputfl = shift;
    open my $fl, "<", $inputfl or die "ERR::MyUtil:: Could not open file for reading: $!";
    my @CTT = <$fl>;
    close $fl;
    chomp @CTT;
    return (@CTT);
}

sub readacd {
    my $indir = shift;
    opendir(my $dh,$indir) or die "ERR::MyUtil:: Could not open directory: $!";
    my @ctts = grep {!/^\./ && -f catfile($indir,$_)} readdir($dh);
    closedir($dh);
    return @ctts;
}

sub truncatefl  {
    my $outfile = shift;

    open(my $F, ">", $outfile) or die "ERR::MyUtil:: Could not truncate results file: $!";
    close $F;
}

sub openOUTFLa {
    my $title = shift @_;
    my $newfl = shift @_;
    my $fl = shift @_;

    if($newfl) {
	$fl = $newfl.$fl;
    }

    my $fh;
    if(!-e $fl || -z $fl) {
	open($fh, ">", $fl) or die "ERR::MyUtil:: Could not open results file for writting: $!";
	print $fh $title if($title);
    } elsif(-s $fl) {
	open($fh, ">>", $fl) or die "ERR::MyUtil:: Could not open results file for appending: $!"; #append
    }
    return *$fh;
}

sub openOUTFLb {
    my $title = shift @_;
    my $newfl = shift @_;
    my $fl = shift @_; #OUTFILE

    if($newfl) {
	$fl = $newfl;
    }

    my $fh;
    if(!-e $fl) {
	open($fh, ">", $fl) or die "ERR::MyUtil:: Could not open results file for writting: $!";
	print $fh $title if($title);
    } elsif(-s $fl) {
	open($fh, ">>", $fl) or die "ERR::MyUtil:: Could not open results file for appending: $!";
    }
    return *$fh;
}

sub setCookyLoc {
    my $loc = shift;
    $_cookyloc = $loc;
}

sub modifyLMAPCooky {
    my @models = @{$_[0]};
    return if(!$_cookyloc);

    my $FH = undef;
    my $cooked = catfile($_cookyloc,LMAPCOOKIE);
    if(-e $cooked && -s $cooked) {
        open($FH, "+<", $cooked) or die "ERR::MyUtil:: Could not recook cookie $cooked: $!\n";
        my @COOKY = <$FH>;
        for(my $i = 0; $i < (@COOKY) ; $i++) {
            if($COOKY[$i] =~ m/^DONE:\t([bwsc](,[bwsc]){0,3})/ && $i == 1) {
                print $FH "RUN:\t".join(",",@models)."\n"; #i = 2
            }
        }
    } else {
        open($FH, ">", $cooked) or die "ERR::MyUtil:: Could not cook cookie $cooked: $!\n";
        print $FH "#LMAP cookie. Do not edit! Date created: ".scalar(localtime())."\n";
        print $FH "RUN:\t".join(",",@models)."\n";
    }
    close $FH;
}

sub readLMAPCooky {
    return () if(!$_cookyloc);

    my @torun = ();
    my @done = ();
    my ($a,$b);
    my @COOKY;
    my $FH = undef;
    my $cooked = catfile($_cookyloc,LMAPCOOKIE);
    if(-e $cooked && -s $cooked) {
        open($FH, "<", $cooked) or die "ERR::MyUtil:: Could not eat cookie $cooked: $!\n";
        @COOKY = <$FH>;
        close $FH;
        for(my $i = 0; $i < (@COOKY) ; $i++) {
            if($COOKY[$i] =~ m/^DONE:\t([bwsc](,[bwsc]){0,3})/) {
                if(index($COOKY[$i], ",") != -1) {
                    @done = split(/,/,$1);
                } else {
                    $done[0] = $1;
                }
                $a = $i;
            }
            if($COOKY[$i] =~ m/^RUN:\t([bwsc](,[bwsc]){0,3})/) {
                if(index($COOKY[$i], ",") != -1) {
                    @torun = split(/,/,$1);
                } else {
                    $torun[0] = $1;
                }
                $b = $i;
            }
        }
    } else {
        return ();
    }

    @_FDONE = (@torun,@done);
    my %h = map {$_ => 1} @_FDONE;
    @_FDONE = keys %h;

    if($a && $b) {
        if($a < $b) {
            return @torun;
        }
    } elsif($b) {
        return @torun;
    }
#when last entry is not Run  
    undef $_cookyloc;
    return ();
}

sub modifyLMAPCooky2 {
    return if(!$_cookyloc);

    my $cooked = catfile($_cookyloc,LMAPCOOKIE);
    if(-e $cooked && -s $cooked) {
	open(my $CF, ">", $cooked) or die "ERR::MyUtil:: Could not recook cookie $cooked: $!\n"; #trunc
	print $CF "#LMAP cookie. Do not edit! Date updated: ".scalar(localtime())."\n";
	print $CF "DONE:\t".join(",",@_FDONE)."\n";
	close $CF;
    }
}


sub getDirContent {
    my $inputdir = $_[0];
    my $DH;
    opendir($DH, $inputdir) or die "ERR::MyUtil:: Could not open directory [$inputdir]: $!\n";
    my @content = grep {!/^\./} readdir($DH) or die "ERR::MyUtil:: Could not read directory [$inputdir]: $!\n";
    close $DH;
    my @ncont;
    for(my $i = 0; $i < @content ; $i++) {
	if($content[$i] ne ".." and  $content[$i] ne ".") {
	    push @ncont, catfile($inputdir,$content[$i]);
	}
    }
    return sort @ncont;
}

sub readPFL {
    my ($p, $d);
    open PFL,"<",$_PFFL or die "ERR::MyUtil:: Could not find preferences file: please run the LMAP configure.pl application!";
    $d = <PFL>;
    {
	local $/ = undef;
	$p = <PFL>;
    }
    close PFL;
    if($d =~ m/\w{3}\s+(\w{3})\s+(\d{1,2})\s+/) {
	foreach (0..$2) {
	    $p = decode_base64($p);
	}
    }
    return $p;
}

sub list_directory {
    my $two = shift;
    my $paintbiofls = shift;
    my $dir = $two ? $two : ".";
    my $ncols = 3;
    my %colmaxlen;
    if(opendir(my $dh, $dir)) {
	my @dl = grep {!/^\./} readdir($dh);
	@dl = sort {$a cmp $b} @dl;
	for(my $k = 0; $k < $ncols ; $k++) {
	    my @col;
	    for(my $j = $k; $j < @dl ; $j+=$ncols) {
		push @col, $dl[$j];
	    }
	    $colmaxlen{$k} = (sort {$b <=> $a} (map {length($_)} @col))[0];
	}

	for(my $i = 0; $i < @dl ; $i+=$ncols) {
	    my $e = $i + $ncols-1;
	    $e = $#dl if($e > $#dl);
	    for(my $j = $i ; $j <= $e ; $j++) {
		my $h = $j;
		$h = $j - $i if($j > $ncols -1);
		my $format = "%-".$colmaxlen{$h}."s%s";
		$format = $_TCOLOR{lmagenta}."%-".$colmaxlen{$h}."s".$_TCOLOR{endcolor}."%s" if($paintbiofls & $dl[$j] =~ /\.(nwk|newick|fst|fas|fasta|ph|phy|phylip)$/ && (-f $dl[$j] || -f catfile($dir,$dl[$j])));
		$format = $_TCOLOR{lyellow}."%-".$colmaxlen{$h}."s".$_TCOLOR{endcolor}."%s" if($dl[$j] =~ /\.csv$/ && (-f $dl[$j] || -f catfile($dir,$dl[$j])));
		$format = $_TCOLOR{Blgreen}."%-".$colmaxlen{$h}."s".$_TCOLOR{endcolor}."%s" if(-x $dl[$j] || -x catfile($dir,$dl[$j]));
		$format = $_TCOLOR{Blblue}."%-".$colmaxlen{$h}."s".$_TCOLOR{endcolor}."%s" if(-d $dl[$j] || -d catfile($dir,$dl[$j]));
		printf($format,$dl[$j]," "x2);
	    }
	    print "\n";
	}
    } else {
	my $err = "$!";
	$err =~ s/\sat[.]*//;
	print "\n\t$err!\n";
    }
    print "\nPress ENTER to continue: ";
    <STDIN>;
}


sub cleanOpts {
    my %opts = %{$_[0]};
    foreach (keys %opts) {
	delete $opts{$_} unless (defined $opts{$_});
    }
    return %opts;
}

sub printToFile {
    my $filename = shift;
    my @Outfile = @{$_[0]};

    open "FL", ">", $filename || die "ERR::MyUtil:: Could not create new file $filename : $!\n";
    foreach (@Outfile) {
	print FL $_,"\n";
    }
    close FL;
    print ":: Content printed to <$filename>\n";
}

sub getcd {
    my $DECFL = shift;
    my $y;
    my $t = scalar(localtime());
    if($t =~ m/\w{3}\s+\w{3}\s+(\d{1,2})\s+/) {
	foreach (0..$1) {
	    $DECFL = encode_base64($DECFL);
	}
    }
    return "LMAP Preferences generated: ".$t."\n".$DECFL;
}

sub getonecd {
    my $op = shift;
    my $CTT = shift;
    if($op eq "c") {
	$CTT = encode_base64($CTT);
    } elsif($op eq "d") {
	$CTT = decode_base64($CTT);
    }
    return $CTT;
}

sub printLicense {
    my $specific = shift;
    if(defined $specific) {
	if($specific =~ m/.*([lgcmio]map\.pl)$/) {
	    $specific = $1;
	}
    }
    my $dfltall = "install.pl, configure.pl, lmap.pl, gmap.pl, cmap.pl, mmap.pl, imap.pl, omap.pl, MyUtil.pm, MyNotify.pm, MyPhylo.pm, MyPAMLInfo.pm";
    my $show = $specific || $dfltall;
    my @License = <<LIC;
LMAP package ($show)
Copyright (C) 2015 Emanuel Maldonado
This program comes with ABSOLUTELY NO WARRANTY; for details see `LICENSE_GPL.txt' file.
This is free software, and you are welcome to redistribute it
under certain conditions; see `LICENSE_GPL.txt' file for details.

LIC

print @License;
    return;
}

END {}

1;



