#!/usr/bin/perl -w

#################################################
# Change PAML control files parameters values   #
#           ------------------                  #
#   Written by Emanuel Maldonado - Jan 2014     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  cmap.pl is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  cmap.pl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  cmap.pl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with cmap.pl.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################


use Data::Dumper;
use strict;
use warnings;

use Getopt::Long;
Getopt::Long::Configure ("bundling");

use vars qw(%TCOLOR);
use File::Spec::Functions qw(:ALL);

#My modules
use FindBin;
use lib "$FindBin::Bin/LMAPlib";
use MyUtil qw(:CPAML :CONST);


use constant {
    CHANGEDFLNM => "changedfiles.txt",
    VERSION => "1.0.0 Nov 20th, 2015"
};

unew();
if(wantColor()) {
    %TCOLOR = enableTCOLOR();
} else {
    %TCOLOR = disableTCOLOR();
}

my %opts;

GetOptions(
    "d=s" => \$opts{d}, 
    "f=s" => \$opts{f}, 
    "g=s" => \$opts{g}, 
    "L=s" => \$opts{L}, 
    "r!" => \$opts{r}, 
    "v!" => \$opts{v}, 
    "h!" => \$opts{h}, 
    ) or HELP_MESSAGE("*"x30);

%opts = cleanOpts(\%opts);

HELP_MESSAGE("") if($opts{h} || !%opts);
VERSION_MESSAGE() if($opts{v});

if(!$opts{d}) {
    HELP_MESSAGE("ERR:: Missing input directory: use -d option followed by path to directory.\n");
}

my $inputdir = $opts{d};
die "ERR:: Directory [$inputdir] does not exist!" if(!-d $inputdir);
$inputdir =~ s:/$::;

if(!$opts{f}) {
    HELP_MESSAGE("ERR:: Required identification of control files to be changed: use option -f with argument.");
}

if(!$opts{g}) {
    HELP_MESSAGE("ERR:: Target parameters/lines identification required: use option -g with argument.\n");
}

if(!$opts{L}) {
    HELP_MESSAGE("ERR:: Text/value required for replacement in parameters: use -L option followed by text in quotes.\n");
}

printLicense($0);
changePAMLParams($opts{f},$opts{g},$opts{L},$opts{r});

exit 0;



sub changePAMLParams {
    my $ctltag = shift;
    my $target_line = shift;
    my $repl_line = shift;
    my $recurse = shift;

    my @DCOLL; 
    my @content = [];
    @content = getDirContent($inputdir);

    my ($tag, $thisctl) = split(":",$ctltag);

    my $flcnt=0;
    my @changedfls;
    my $flag = 0;
    while(1) {
	while($_  = shift @content) { 
	    if(-f $_) { 
		my $ctlflname = (splitpath($_))[2];
		if($tag eq 'f' && $thisctl eq $ctlflname) {
		    $flag = 1;
		} elsif($tag eq 's' && $ctlflname =~ m:^($thisctl):) {
		    $flag = 1;
		} elsif($tag eq 'e' && $ctlflname =~ m:($thisctl)$:) {
		    $flag = 1;
		}
		if($flag) {
		    $flag = 0;
		    print ":z: Handling File: $_...\n";
		    if(modifyControlFile($_,$repl_line,$target_line)) {
			$flcnt++;
			push @changedfls, $_;
		    }
		}
	    } elsif (-d $_) {
		print ":z: Adding Subdirectory $_\n";
		push @DCOLL, $_;
	    }
	    if($recurse && !@content) { #after all content is passed through
		while(my $currdir = shift(@DCOLL)) {  # remove at the beggining
		    print ":: Going into directory $currdir...\n";
		    @content = getDirContent($currdir);
		    last if(@content);
		    print ":: Directory $currdir empty!\n" if(!@content);
		}
	    }
	}
	last if(!@content);
    }

    print ":z: Changed $flcnt control files\n";
    printToFile(catfile(".",CHANGEDFLNM), \@changedfls);
}

sub modifyControlFile {
    my $inputfl = $_[0];
    my $new_repl_line = $_[1];
    my $target_line = $_[2];

    my $changed = 0;
    my $changedlines = 0;
    my @FILE = readac($inputfl);

    if(!@FILE) {
	return $changed;
    }

    my ($param, $value, $offs, $eqidx, $staridx);
    my ($tag, $word) = split(/:/,$target_line, 2);
    for(my $i = 0; $i < @FILE ; $i++) {
	next if($FILE[$i] =~ m/^\s*\*/);
	if(index($FILE[$i],"=") != -1) {
	    my $param = (split("=",$FILE[$i]))[0];
	    $staridx = index($FILE[$i],"\*");
	    $eqidx = index($FILE[$i],"=");
	    $offs = ($staridx)-($eqidx)-1;
	    if($staridx > 0) {
		$value = substr($FILE[$i], $eqidx+1, $offs);
	    } else {
		$value = substr($FILE[$i], $eqidx+1);
	    }
	    $param =~ s/\s*//g;
	    if($tag eq "f" && $param eq $word) { #complete
		$FILE[$i] =~ s:\Q$value\E: $new_repl_line :;
		$changedlines++;
	    } elsif($tag eq "s" && $param =~ m/^($word)/) { #beggining
		$FILE[$i] =~ s:\Q$value\E: $new_repl_line :;
		$changedlines++;
	    } elsif($tag eq "c" && $param =~ m/($word)/) { #middle
		$FILE[$i] =~ s:\Q$value\E: $new_repl_line :;
		$changedlines++;
	    } elsif($tag eq "e" && $param =~ m/($word)$/) { #end
		$FILE[$i] =~ s:\Q$value\E: $new_repl_line :;
		$changedlines++;
	    }
	    $changed=1;
	}
    }
    $changed = $changedlines ? 1 : 0;

    print "\t:: Modified $changedlines lines!\n";
    printToFile($inputfl, \@FILE);

    return $changed;
}



sub HELP_MESSAGE {
    print shift;
    my $citation = CITATION;
    my $pweb = PROJWEBSITE;
    my $v = VERSION;
    my @HELP = <<HELP;

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    cmap.pl - Change PAML control files specific parameters.
   
  $TCOLOR{bold}SYNOPSIS:$TCOLOR{endcolor}
    
    cmap.pl -d [directory] -f [e:ctl] -g f:param -L "value" {-r} 

  $TCOLOR{bold}DESCRIPTION:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | This application is intended to change specific parameters values in several
    | PAML control files already located in an organized directory structure.
    | This is a helper application which is meant to be used, if necessary, after
    | creating an organized directory structure, done by gmap.pl and before running
    | with mmap.pl, so that, any required changes can still take effect.

  $TCOLOR{bold}OPTIONS:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | -d [dir]        Input directory of the final structure containing all PAML input files.
    | -f [tag:part]   Choose control files through a given filename portion, where tag is one of:
    |                     f: filename equals given part
    |                     s: filename starts with part
    |                     e: filename ends with part
    | -g [tag:ptext]  Control file target parameter line identified by its name eg.: -g f:icode
    |                 Where tag is one of the following:
    |                     f: Complete/full line
    |                     s: Starting with text  (may affect several lines/parameters)
    |                     c: Containing text     (may affect several lines/parameters)
    |                     e: Ending with text    (may affect several lines/parameters)
    | -L [paramval]   Value or text to be replaced in selected line(s)/parameter(s) from option -g.
    |                 ----------------------- not mandatory: ----------------------
    | -r              Work recursively.
    |                 -----------------------------------
    | -h              This help
    | -v              Application version

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Jan 2014.

  $TCOLOR{bold}CITATION AND DOWNLOAD:$TCOLOR{endcolor}
    
    $citation
    $pweb

  $TCOLOR{bold}LICENSE AND COPYRIGHT:$TCOLOR{endcolor}

    LMAP: Lightweight Multigene Analyses in PAML
    cmap.pl is part of LMAP package
    Copyright (C) 2015 Emanuel Maldonado

  cmap.pl is free software: you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  cmap.pl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with cmap.pl.  If not, see <http://www.gnu.org/licenses/>.

HELP
print @HELP;
    exit 1;
}

sub VERSION_MESSAGE {
    my $v = VERSION;
    my @VERS = <<VERS;
Use option -h for help and other information.

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    cmap.pl - Change PAML control files specific parameters.

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Jan 2014.

VERS

print @VERS;
    exit 1;
}

