#!/usr/bin/perl -w

#################################################
# Create necessary directory structures.        #
#           ------------------                  #
#   Written by Emanuel Maldonado - Mar 2015     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  gmap.pl is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  gmap.pl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  gmap.pl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with gmap.pl.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use Data::Dumper;
use strict;
use warnings;

use Getopt::Long;
Getopt::Long::Configure ("bundling");

use vars qw($PAML_PROG %KAPPAVALS %OMEGAVALS $ALINDIR $TINDIR $TFILE $PREPMODELS $BASEDIR @DIRLIST $PROJID $TERM
%ALIGNLOC %TREELOC @SITEMODELS %HYPOTHESIS %GENENAMES %CTLLOC %CTLFILES %TCOLOR $DEMIDFLTDIR $SAVED $HAVETREES);

use sigtrap qw(handler do_rexit INT QUIT TERM TSTP HUP);

use File::Copy;
use File::Spec::Functions qw(:ALL);

#My modules
use FindBin;
use lib "$FindBin::Bin/LMAPlib";
use MyUtil qw(:GPAML :CONST);
use MyPAMLInfo qw(:ICODEML :CONSTANTS);
use MyPhylo qw(:DEFAULT :CONSTANTS);

use constant {
    PROJID => "LMAP_PROJ",
    PHELPFILE => "gmap_promptHelp.txt",
    VERSION => "1.0.0 Nov 20th, 2015"
};

$OMEGAVALS{b} = [CODEML_OMEGA_VALS]; #defaults
$OMEGAVALS{c} = [CODEML_OMEGA_VALS]; #defaults
$KAPPAVALS{s} = [CODEML_KAPPA_VALS]; #defaults
$KAPPAVALS{w} = [CODEML_KAPPA_VALS]; #defaults

$DEMIDFLTDIR = DFLT_TREEDIR;

@DIRLIST = ();

$TERM = 0;

unew();
if(wantColor()) {
    %TCOLOR = enableTCOLOR();
} else {
    %TCOLOR = disableTCOLOR();
}

my %opts;

GetOptions(
    "A=s" => \$opts{A}, 
    "T=s" => \$opts{T}, 
    "t=s" => \$opts{t}, 
    "d=s" => \$opts{d}, 
    "m=s" => \$opts{m}, 
    "j=s" => \$opts{j}, 
    "O=s" => \$opts{O}, 
    "K=s" => \$opts{K}, 
    "help" => \$opts{help}, 
    "v!" => \$opts{v}, 
    "h!" => \$opts{h}, 
    ) or HELP_MESSAGE("*"x30);

%opts = cleanOpts(\%opts);

HELP_MESSAGE("") if($opts{h} || !%opts);
VERSION_MESSAGE() if($opts{v});

dumpPromptHelp() if($opts{help});

if(!$opts{A} || !-d $opts{A}) {
    HELP_MESSAGE("ERR:: Missing input directory containing alignments: use -A option followed by path to directory.\n");
}

if(!$opts{d} || !-d $opts{d}) {
    HELP_MESSAGE("ERR:: Missing input directory where to locate the directory struture: use -d option followed by path to directory.\n");
}

if(! defined $opts{T} && ! defined $opts{t}) {
    HELP_MESSAGE("ERR:: Missing input tree option: use either -T followed by path to directory or -t followed by path to newick file.\n");
}

if(defined $opts{T}) {
    if(!-d $opts{T}) {
	HELP_MESSAGE("ERR:: Missing input directory containing newick trees: use -T option followed by path to directory.\n");
    }
}

if(defined $opts{t}) {
    if(!-f $opts{t}) {
	HELP_MESSAGE("ERR:: Missing input file newick tree: use -t option followed by path to file.\n");
    }
}

if(!$opts{m}) {
    HELP_MESSAGE("ERR:: Missing input specification of models to be prepared: use -m followed by required model specification\n");
}

if(defined $opts{m}) {
    if($opts{m} !~ m/^(b|w|c|(s\[([0-8]|8a)(:([0-8]|8a))*\]))(,(b|w|c|(s\[([0-8]|8a)(:([0-8]|8a))*\])))*$/) {
	HELP_MESSAGE("ERR: Models specification should be separated by commas and\nfor site models, models are separated by colon and in brackets.\n");
    }
}

if($opts{m} =~ m/,?b,?|,?c,?/) { #branch and clade
    if($opts{O}) { #then new omega values should be:
	my $o = $opts{O};
	my $one = ($o =~ s/b//g);
	my $two = ($o =~ s/c//g);
	if($one > 1 || $two > 1) {
	    HELP_MESSAGE("ERR: There should be only one occurrence for each method branch (b) and/or clade (c) with specified values.\n");
	}
        my @movals;
        if((index $opts{O}, ":") != -1) {
            @movals = split(":",$opts{O});
            if(@movals > 2) {
                HELP_MESSAGE("ERR: There should be at most two groups of values specified; for branch (b) and/or clade (c): c[vals]:b[vals] or b[vals] or c[vals].\n");
            }
        } else {
	    @movals = ($opts{O},"");
        }
	if($movals[0] =~ m/b\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
	    @{$OMEGAVALS{b}} = split(",",$1);
	    if($movals[1] =~ m/c\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
		@{$OMEGAVALS{c}} = split(",",$1);
	    }
	} elsif($movals[0] =~ m/c\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
	    @{$OMEGAVALS{c}} = split(",",$1);
	    if($movals[1] =~ m/b\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
		@{$OMEGAVALS{b}} = split(",",$1);
	    }
	} else {
	    die "ERR:: Option -O argument is not correctly specified: should be one or both of b[x,...,z] or c[x,...,z] were x,...,z are any values.\n";
	}
    }
}

if($opts{m} =~ m/,?s\[[\d:]+\],?|,?w,?/) { #branch-site and site
    if($opts{K}) { #then new omega values should be:
	my $k = $opts{K};
	my $one = ($k =~ s/s//g);
	my $two = ($k =~ s/w//g);
	if($one > 1 || $two > 1) {
	    HELP_MESSAGE("ERR: There should be only one occurrence for each method branch-site (w) and/or site (s) with specified values.\n");
	}

        my @mkvals;
        if(index($opts{K}, ":") != -1) {
            @mkvals = split(":",$opts{K});
            if(@mkvals > 2) {
                HELP_MESSAGE("ERR: There should be at most two groups of values specified; for branch-site (w) and/or site (s): w[vals]:s[vals] or s[vals] or w[vals].\n");
            }
        } else {
	    @mkvals = ($opts{K},"");
        }
	if($mkvals[0] =~ m/s\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
	    @{$KAPPAVALS{s}} = split(",",$1);
	    if(defined $mkvals[1] && $mkvals[1] =~ m/w\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
		@{$KAPPAVALS{w}} = split(",",$1);
	    }
	} elsif($mkvals[0] =~ m/w\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
	    @{$KAPPAVALS{w}} = split(",",$1);
	    if(defined $mkvals[1] && $mkvals[1] =~ m/s\[((\d+(\.\d+)?)(,\d+(\.\d+)?)*)\]/) {
		@{$KAPPAVALS{s}} = split(",",$1);
	    }
	} else {
	    die "ERR:: Option -K argument is not correctly specified: should be one or both of s[x,...,z] or w[x,...,z] were x,...,z are any values.\n";
	}
    }
}

printLicense($0);

$BASEDIR = $opts{d};
$BASEDIR =~ s:/$::;
$ALINDIR = $opts{A};
$ALINDIR =~ s:/$::;
my $ans;
if($opts{t}) {
    $TFILE = $opts{t};
    if(interactiveTreeLabel()) {
	$opts{T} = $DEMIDFLTDIR;
    } else {
	exit 0;
    }
}
if($opts{T}) {
    $TINDIR = $opts{T};
    $TINDIR =~ s:/$::;
    getTrees();
}

$PREPMODELS = $opts{m};
$PAML_PROG = CODEML;
$PROJID = $opts{j} || createDefaultProjID();

getCtlFiles();
getAligns();
buildDirStructure();
createDataPaths();

exit 0;


############################
################ FUNCTIONS:
#######################

sub interactiveTreeLabel {

    my $show_mode = BLS;
    my $op;
    my $numnodes = 0;
    my $msg = "";
    my $tfl = $TFILE;
    $tfl = (splitpath($TFILE))[2] if(index($TFILE,"/") != -1);
    my $ctfl = $TCOLOR{Bcyan}.$tfl.$TCOLOR{endcolor};
    my $barsz = gsBrLength();
    my $unrtag = "";
    $SAVED = 1;
    $HAVETREES = 0;
    pnew($TFILE);
    if(!isTreeRooted()) {
	return 0 if(!decideOnTreeType());
	$unrtag = "$TCOLOR{Bred}\[UNROOTED\]$TCOLOR{endcolor}";
	$msg = "$TCOLOR{red}Unrooted tree: tree representation maybe wrong...Root the tree with command \'rootunr Na,Nb\'.$TCOLOR{endcolor}";
    }
    my $cmdcnt = 0;
    my $ph = 1;
    for (;;) {
	printCanvas($barsz,$unrtag,$ctfl,$tfl,$show_mode) if($ph);
	printStatusBars($show_mode,$numnodes,$barsz) if($ph);
	printPrompt($msg,$cmdcnt);
	$op = <STDIN>;
	chomp $op;
	$op = "" if(STDIN->error);
	$op =~ s/^\s+//;
	$msg = "";
	$ph = 1;
	if($op =~ m/^(A\d+(,A\d+)*)=([\$\#]\d+)\s*$/ || $op =~ m/^(L\d+(,L\d+)*)=([\#]\d+)\s*$/) {
	    my @r = nodesverify($1);
	    if($r[0]) {
		$numnodes += branchLabelNodeID($_,$3) foreach (split(",",$1));
		$msg = "Nodes $1 labeled as $3";
	    } else {
		$msg = "Nodes $1 NOT labeled: node ".$r[1]." is not ".$r[2];
	    }
	    $SAVED = 0;
	} elsif($op =~ m:^ls(\s)*((([~\/\.]{1}[\/]?)|(((([~\.]?\/)?)[\w\d\.-])+[\/]?))\s*)?$:) {
	    my $d = $2;
	    $d =~ s:^~:$ENV{HOME}: if($d);
	    list_directory($d,1);
	} elsif($op =~ m/^rbl\s*$/) {
	    removeBranchlengths();
	    $SAVED = 0;
	} elsif($op =~ m/^rbs\s*$/) {
	    removeBootstraps();
	    $numnodes = 0;
	    $SAVED = 0;
	} elsif($op =~ m/^rtaxa\s+(L\d+=[\w\ ]+(,L\d+=[\w\ ]+)*)\s*$/ || $op =~ m/^rtaxa\s+(([~\.]\/)?.*\.csv)\s*$/) {
	    my $d = $1;
	    $d =~ s:^~:$ENV{HOME}: if($d);
	    $msg = renameTips($d);
	    $SAVED = 0;
	} elsif($op =~ m/^prune\s+([LA]\d+(,[LA]\d+)*)\s*$/) {
	    $msg = pruneTree($1);
	    $SAVED = 0;
	} elsif($op =~ m/^rootunr\s+([LA]\d+,[LA]\d+)\s*$/) {
	    $msg = rootUnrTree($1);
	    $unrtag = "" if(isTreeRooted());
	    $SAVED = 0;
	} elsif($op =~ m/^saveunr\s*(.+\..+)\s*$/) {
	    $msg = "Saved unrooted tree to ".writeUnrTree2File($1);
	    $SAVED = 1;
	    $HAVETREES = 1;
	} elsif($op =~ m/^save\s*(.+\..+)\s*$/) {
	    $msg = "Saved to ".writeTreeAsTxt2File($1);
	    $SAVED = 1;
	    $HAVETREES = 1;
	} elsif($op =~ m/^savetodir\s+([\.\w\/]+)\s*$/) {
	    $DEMIDFLTDIR = modifyDfltTreeDir($1); 
	    $msg = "Saving future newick trees to ".$DEMIDFLTDIR;
	} elsif($op =~ m/^reset\s*$/) {
	    pnew($TFILE);
	    $numnodes = 0;
	    $SAVED = 0;
	} elsif($op =~ m/^brlen\s+(\d+)\s*$/) { 
	    $barsz = gsBrLength(int($1));
	} elsif($op =~ m/^show\s*(bs|bl|b|n)\s*$/) {
	    $show_mode = $1;
	    $msg = "Showing none.";
	    $msg = "Showing branch lengths." if($1 eq BL);
	    $msg = "Showing bootstraps." if($1 eq BS);
	    $msg = "Showing branch lengths and bootstraps." if($1 eq BLS);
	} elsif($op =~ m/\?\s*$/) {
	    promptHelp();
	    $ph = 0;
	} elsif($op =~ m/^[dD](one)?\s*$/ || $TERM) {
	    last if(rdone());
	} else {
	    $msg = "Command '$op' not found!" if($op);
	}
	$cmdcnt++;
    }

    return proceed() if($HAVETREES);
    return 0;
}


sub do_rexit { #signal-handler
    $TERM = 1;
    STDIN->flush;
    STDIN->close() || warn "WARNING: STDIN close failed: $!";
}

sub rdone {
    clear(0);
    local $SIG{INT} = 'IGNORE';
    local $SIG{QUIT} = 'IGNORE';
    local $SIG{TERM} = 'IGNORE';
    local $SIG{TSTP} = 'IGNORE';
    print "\n\n";
    print $TCOLOR{menu},"EXIT screen - Really Done?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x11,"\n",$TCOLOR{endcolor},"\n";
    open(STDIN,"<","/dev/tty") or die "ERR:: Could not reopen STDIN: $!";
    my $ans = "";
    do {
	if(!$SAVED) {
	    print "\tPhylogenetic tree is modified, do you want to exit labeling mode, without saving? ([$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	} else {
	    print "\tDo you really want to exit? ([$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	}
	$ans = <STDIN>;
	chomp $ans;
    } while ($ans !~ m/^[yYnN]/);
    if($ans =~ m/[yY](es)?/) {
	print "\nLeaving... Goodbye!\n" if(!$HAVETREES);
	return 1;
    }
    $TERM = 0;
    return 0;
}

sub proceed {
    clear(0);
    local $SIG{INT} = 'IGNORE';
    local $SIG{QUIT} = 'IGNORE';
    local $SIG{TERM} = 'IGNORE';
    local $SIG{TSTP} = 'IGNORE';
    my $cwd = "";
    print "\n\n";
    print $TCOLOR{menu},"CONTINUE screen - Proceed?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x15,"\n",$TCOLOR{endcolor},"\n";
    do {
	if($DEMIDFLTDIR !~ m/^\//) {
	    $cwd = catfile(".","");
	}
	print "\n\tWould you like to proceed to directory creation with the newly created trees in $cwd".$DEMIDFLTDIR."? (Type [$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	$ans = <STDIN>;
    } while ($ans !~ m/^[YynN]/);
    if($ans =~ m/[yY]/) {
	return 1;
    } else {
	print "\nLeaving... Goodbye!\n";
	return 0;
    }
}

sub decideOnTreeType {
    clear(0);
    local $SIG{INT} = 'IGNORE';
    local $SIG{QUIT} = 'IGNORE';
    local $SIG{TERM} = 'IGNORE';
    local $SIG{TSTP} = 'IGNORE';
    my $cwd = "";
    print "\n\n";
    print $TCOLOR{menu},"UNROOTED TREE screen - Proceed?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor},"\n";
    do {
	if($DEMIDFLTDIR !~ m/^\//) {
	    $cwd = catfile(".","");
	}
	print "\n\tThe tree was found to be unrooted (trichotomy), would you like to proceed and root the tree in gmap or exit? (Type [$TCOLOR{qunln}P$TCOLOR{endcolor}]roceed/[$TCOLOR{qunln}E$TCOLOR{endcolor}]xit)\n\t> ";
	$ans = <STDIN>;
    } while ($ans !~ m/^[PpEe]/);
    if($ans =~ m/[Pp]/) {
	return 1;
    } else {
	print "\nLeaving... Goodbye!\n";
	return 0;
    }
}

sub printCanvas {
    my $barsz = shift;
    my $unrtag= shift;
    my $ctfl = shift;
    my $tfl = shift;
    my $show_mode = shift;

    clear(0);
    printf "%".$barsz."s","= ".$ctfl." ".$unrtag." "."="x(($barsz-3)-length($tfl.$unrtag))."\n\n";
    print $TCOLOR{cyan};
    draw_ascii($show_mode);
    print $TCOLOR{endcolor};
    print "\n","="x$barsz,"\n";
}

sub printStatusBars {
    my $show_mode = shift;
    my $numnodes = shift;
    my $barsz = shift;


    my $show = $show_mode eq BLS ? 'BOTH' : $show_mode eq BL ? 'BRLEN' : $show_mode eq BS ? 'BOOTS' : 'NONE';

    my $statusbar1 = "Nodes Modified: [".$numnodes."], Nodes Labeled: [".getNumNodesLabeled()."], Show Mode: [".$show."], ASCII brlen: [".$barsz."]\n";

    print $TCOLOR{menu},"="x(length($statusbar1)),"\n",$TCOLOR{endcolor};
    print $TCOLOR{Bcyan},$statusbar1,$TCOLOR{endcolor};
    print $TCOLOR{menu},"="x(length($statusbar1)),"\n",$TCOLOR{endcolor};

    my $statusbar2 = "Saving to folder: [".catfile(".",$DEMIDFLTDIR)."]\n";

    print $TCOLOR{bold},$statusbar2,$TCOLOR{endcolor};
    print $TCOLOR{menu},"="x(length($statusbar2)),"\n",$TCOLOR{endcolor};
}

sub printPrompt {
    my $msg = shift;
    my $ccnt = shift;

    if($msg) {
	$msg = " \{".$TCOLOR{cyan}.$msg.$TCOLOR{endcolor}.$TCOLOR{bold}."\}";
    } else {
	$msg = $TCOLOR{bold};
    }
    print "( type '?' for help ; 'd' when done )\n";
    print $TCOLOR{bold},"[$ccnt GMAP$msg PROMPT]\$ ",$TCOLOR{endcolor};
}

##########

sub buildDirStructure {

    my $max;
    my %selmodels = map { $_ => 1 } getRequiredModels();

#location/projname #1 
    push @DIRLIST, catfile($BASEDIR,$PROJID);

    $max = getMaxLenDirList();

#location/projname/modeltypes #2
    my @mlist = map {$C_MTYPESTAGS{$_}} keys %selmodels;
    makeDirList($PROJID,$max,\@mlist);

    $max = getMaxLenDirList();

    my $btag = catfile("",$C_MTYPESTAGS{b});
    my $ctag = catfile("",$C_MTYPESTAGS{c});
    my $wtag = catfile("",$C_MTYPESTAGS{w});
    my $stag = catfile("",$C_MTYPESTAGS{s});

#location/projname/modeltypes/values #3
    makeDirList($btag,$max,$OMEGAVALS{b}) if(exists $selmodels{b});
    makeDirList($ctag,$max,$OMEGAVALS{c}) if(exists $selmodels{c}); 
    makeDirList($stag,$max,$KAPPAVALS{s}) if(exists $selmodels{s});
    makeDirList($wtag,$max,$KAPPAVALS{w}) if(exists $selmodels{w});

    $max = getMaxLenDirList();

#location/projname/modeltypes/values/hypothesis #4
    makeDirList($btag,$max,$HYPOTHESIS{b}) if(exists $selmodels{b});
    makeDirList($ctag,$max,$HYPOTHESIS{c}) if(exists $selmodels{c});
    makeDirList($wtag,$max,$HYPOTHESIS{w}) if(exists $selmodels{w});

    $max = getMaxLenDirList();
    my $smax = getMaxLenDirList($stag);

#location/projname/modeltypes/values/(hypothesis/)modelnames #5
    makeDirList($btag,$max,$C_MTYPESNAMES{b}) if(exists $selmodels{b});
    makeDirList($ctag,$max,$C_MTYPESNAMES{c}) if(exists $selmodels{c}); 
    makeDirList($stag,$smax,$C_MTYPESNAMES{s}) if(exists $selmodels{s});
    makeDirList($wtag,$max,$C_MTYPESNAMES{w}) if(exists $selmodels{w}); 
    
    $max = getMaxLenDirList();
    $smax = getMaxLenDirList($stag);

#location/projname/modeltypes/values/(hypothesis/)modelnames/genenames #6
    makeDirList($btag,$max,$GENENAMES{b}) if(exists $selmodels{b});
    makeDirList($ctag,$max,$GENENAMES{c}) if(exists $selmodels{c}); 
    makeDirList($stag,$smax,$GENENAMES{s}) if(exists $selmodels{s});
    makeDirList($wtag,$max,$GENENAMES{w}) if(exists $selmodels{w}); 

}

sub createDataPaths {
    my $dcnt = 0;
    print ":z: Going to create ",scalar(@DIRLIST)," directories...\n";
    @DIRLIST = sort @DIRLIST;
    my %markExistent = (s => 0, w => 0,  c => 0,  b => 0);
    foreach (@DIRLIST) {
    	if(mkdir $_) {
	    print "::Directory [ ",$_," ] created!\n";
	    $dcnt++;
    	    if(isDataPath($_)) {
		print "\t::Placing alignment, ";
    		#copy aln
    		my $gene = getGeneMatchPath($_);
    		if(copy($ALIGNLOC{$gene}, catfile($_,(splitpath($ALIGNLOC{$gene}))[2]) )) {
		    #print "Alignment file of $gene copied!\n";
		} else {
		    die "ERR:: Could not copy file $gene into $_: $!\n";
		}
		print "phylogenetic tree ";
    		#copy tree
    		my $tree = getTreeMatchPath($_);
    		if(copy($TREELOC{$tree}, catfile($_,(splitpath($TREELOC{$tree}))[2]) )) {
		    #print "Tree file of $tree copied!\n";
		} else {
		    die "ERR:: Could not copy file $tree into $_: $!\n";
		}
		print "and control file... ";
    		#modify and create control file for this case
    		createctl($_,$ALIGNLOC{$gene},$TREELOC{$tree});
		print "Done!\n";
    	    }
    	} elsif($_ =~ m:\/($PROJID)$:) {
	    warn("WARNING:: Could not create directory $_. Directory already exists!\n");
	    next;
	} elsif($_ =~ m:\/$PROJID\/(Site|Clade|BranchSite|Branch)Models$:) { #PROJID exists already...
	    my $m = $1;
	    $m = 'w' if($m =~ m/BranchSite/);
	    $markExistent{lc($m)} = 1;
	    warn("WARNING:: Could not create directory $_. Directory already exists!\n");
	    next;
	} else {
	    next if($markExistent{s} || $markExistent{w} || $markExistent{c} || $markExistent{b}); #if any modeltype already exists skip and try create next
    	    die "ERR:: Could not create directory $_: $!\n";
    	}
    }
    if($dcnt == scalar(@DIRLIST)) {
	print ":z: All directories were created successfully!\nYou may now run mmap.pl for running all these analyses.\n";
    } elsif($markExistent{s} && $markExistent{w} && $markExistent{c} && $markExistent{b}) {
	print ":z: Could not create directories for any model types: Please create a new project ID (use option -j).\n";
    }
}

sub createctl {
    my $destpath = shift;
    my $genefl = shift;
    my $treefl = shift;

    my ($modeltype) = ($destpath =~ m:\/((Clade|Branch|BranchSite|Site)Models)\/:);
    my ($valok) = ($destpath =~ m:\/(\d+(.\d+)?)\/:);
    $treefl = (splitpath($treefl))[2];
    $genefl = (splitpath($genefl))[2];
    my ($icode) = ($genefl =~ m:_[swcb]{0,4}(\d{1,2})\.\w+$:);
    $icode = defined $icode ? $icode : 0; #default universal code

    if($modeltype eq $C_MTYPESTAGS{b}) {
	modelSelector('b',$destpath,$genefl,$treefl,$icode,$valok);
    } elsif($modeltype eq $C_MTYPESTAGS{c}) {
    	modelSelector('c',$destpath,$genefl,$treefl,$icode,$valok);
    } elsif($modeltype eq $C_MTYPESTAGS{w}) {
    	modelSelector('w',$destpath,$genefl,$treefl,$icode,$valok);
    } elsif($modeltype eq $C_MTYPESTAGS{s}) {
    	modelSelector('s',$destpath,$genefl,$treefl,$icode,$valok);
    }
}


sub modelSelector {
    my $mtype = shift;
    my $destpath = shift;
    my $genefl = shift;
    my $treefl = shift;
    my $icode = shift;
    my $valok = shift;

    my $nssites = "";
    if($mtype eq 's') {
	$nssites = join(" ",grep {!/8a/} @SITEMODELS);
    }

    foreach (@{$C_MTYPESNAMES{$mtype}}) {
	if($destpath =~ m:\/$_\/:i) {
	    my $loc = $CTLLOC{$CTLFILES{$mtype}{$_}};
	    makeControlfl($destpath,$loc,$genefl,$treefl,$icode,$valok,$nssites);
	}
    }
}

sub makeControlfl {
    my $destloc = shift;
    my $loc = shift;
    my $gene = shift;
    my $tree = shift;
    my $icode = shift;
    my $value = shift;
    my $nssites = shift;

    my @CTT = readac($loc); #controlfl
    foreach (@CTT) {
	$_ =~ s/<align.fas>/$gene/;
	$_ =~ s/<tree.nwk>/$tree/;
	$_ =~ s/<icode>/$icode/;
	$_ =~ s/<valok>/$value/; #omega or kappa
	$_ =~ s/<nssites>/$nssites/ if($nssites); #site-models
    }

    open(CTL,">",catfile($destloc,CODEMLCTRLFLN));
    print CTL $_,"\n" foreach (@CTT);
    close CTL;
}

#If there were user choices as to which trees to use in modeltypes then there should be only those paths for these choices not for not chosen ones (which are useless-not to be run);
sub getTreeMatchPath { 
    my $p = shift;
    foreach my $h (keys %TREELOC) { #will try to match path to hypothesis and thus give correct file
	if($p =~ m:\/SiteModels\/:) {
	    if($p =~ m:\/\Q$h\E$:) { #case of site models were gene matching tree ends the path
		return $h;
	    }
	} elsif($p =~ m:\/\Q$h\E\/:) {
	    return $h; #hypothesis <=> tree file
	}
    }
    die "ERR:: No match found for given phylogenetic trees.\n";
}

#If there were user choices as to which genes to use in modeltypes then there should be only those paths for these choices not for not chosen ones (which are useless-not to be run);
sub getGeneMatchPath { 
    my $p = shift;
    foreach my $g (keys %ALIGNLOC) {
	if($p =~ m/\/\Q$g\E$/) {
	    return $g;
	}
    }
    die "ERR:: No match found for given alignments.\n";
}

sub isDataPath { #the path which will contain all files
    my $p = shift;
    foreach my $g (keys %ALIGNLOC) {
	if($p =~ m:\/\Q$g\E$:) {
	    return 1;
	}
    }
    return 0;
}

sub getLongestPaths {
    my $distinct = shift;
    my @dlist = @DIRLIST;
    if($distinct) {
	@dlist = grep {/$distinct/} @DIRLIST;
    }
    my $y;
    my %longest = map { $_ => $y = ($_ =~ s:\/:\/:g) } @dlist; #get size based on nber of '/'
    return %longest;
}

sub getMaxLenDirList {
    my $distinct = shift;
    my %longest = getLongestPaths($distinct);
    my $max = (sort { $b <=> $a } values %longest)[0];
    return $max;
}

sub makeDirList {
    my $distinct = shift;
    my $max = shift;
    my @dlist = @{$_[0]};

    my %longest = getLongestPaths($distinct);
    my @mydlist = map { $_ if($longest{$_} == $max); } keys %longest;
    my @mdlist = grep { $_ ne ""} @mydlist;
    @mdlist = grep { $_ =~ m/$distinct/ } @mdlist;

    foreach (@mdlist) {
	foreach my $newdir (@dlist) {
	    if($newdir) {
		push @DIRLIST, catdir($_,$newdir);
	    }
	}
    }
}

sub getRequiredModels {
    my $models = $PREPMODELS;
    $models =~ s/s\[(.)+\]/s/;
    my @models = split(",",$models);

    if($PREPMODELS =~ m/,?s\[([012378:a]+)\],?/) {
	@SITEMODELS = split(":",$1);
	@{$C_MTYPESNAMES{s}}[0] = (grep {!/8a/} @SITEMODELS) ? @{$C_MTYPESNAMES{s}}[0] : "";
	@{$C_MTYPESNAMES{s}}[1] = (grep {/8a/} @SITEMODELS) ? @{$C_MTYPESNAMES{s}}[1] : "";
    }

    return @models;
}

sub getCtlFiles {
    my @ctts = readacd($C_CONTROLTMPL);

    #make hash pointing "gene names" to "absolute location" for later copying
    %CTLLOC = map {$_ => catfile($C_CONTROLTMPL,$_)} @ctts;
    %{$CTLFILES{b}} = map {
	get_model_key($_,'b') => $_
    } (grep {/_$C_MTYPESTAGS{b}/i} @ctts);

    %{$CTLFILES{c}} = map {
	get_model_key($_,'c') => $_
    } (grep {/_$C_MTYPESTAGS{c}/i} @ctts);

    %{$CTLFILES{w}} = map {
	get_model_key($_,'w') => $_
    } (grep {/_$C_MTYPESTAGS{w}/i} @ctts);

    %{$CTLFILES{s}} = map {
	get_model_key($_,'s') => $_
    } (grep {/_$C_MTYPESTAGS{s}/i} @ctts);
}

sub get_model_key {
    my $v = shift;
    my $k = shift;
    my $model;
    foreach $model (@{$C_MTYPESNAMES{$k}}) {
	if( $v =~ m/$model/i) {
	    return $model;
	}
    }
}

sub getAligns { #modeltypes distinction ; genes => locations
    my @ctts = readacd($ALINDIR);

    #=> GeneCode_bwsc0.fas # discard files not name formatted acc to requirements
    @ctts = grep {/(_[bwsc]+\d{1,2})(\.(fst|fas|fasta|ph|phy|phylip|txt))$/} @ctts; 

    %GENENAMES = ();

    %ALIGNLOC = map {s/_[bwsc]+\d{1,2}\.\w+$//r => catfile($ALINDIR,$_)} @ctts; 

    #grab files discriminating each modeltypes and respective icodes
    @{$GENENAMES{b}} = (grep {/_([swc]{0,3}b{1}[swc]{0,3})\d{1,2}\.\w+$/ } @ctts);
    @{$GENENAMES{b}} = map { s/_[bwcs]+\d{1,2}\.\w+$//r } @{$GENENAMES{b}};
    @{$GENENAMES{c}} = (grep {/_([swb]{0,3}c{1}[swb]{0,3})\d{1,2}\.\w+$/ } @ctts);
    @{$GENENAMES{c}} = map { s/_[bwcs]+\d{1,2}\.\w+$//r } @{$GENENAMES{c}};
    @{$GENENAMES{w}} = (grep {/_([sbc]{0,3}w{1}[sbc]{0,3})\d{1,2}\.\w+$/ } @ctts);
    @{$GENENAMES{w}} = map { s/_[bwcs]+\d{1,2}\.\w+$//r } @{$GENENAMES{w}};
    @{$GENENAMES{s}} = (grep {/_([bwc]{0,3}s{1}[bwc]{0,3})\d{1,2}\.\w+$/ } @ctts);
    @{$GENENAMES{s}} = map { s/_[bwcs]+\d{1,2}\.\w+$//r } @{$GENENAMES{s}};

    if(!%GENENAMES) {
	die "ERR:: Could not find alignment files properly named in $ALINDIR\n";
    }
}

sub getTrees { #modeltypes distinction ; hypothesis => locations
    my @ctts = readacd($TINDIR);
    
    %HYPOTHESIS = ();

    @ctts = grep {/(_[wcb]{1,3})|(_s)(\.(nwk|newick|txt))$/} @ctts; #=> GeneCode_s.nwk | HYP_bwc.nwk
    %TREELOC = map {s/((_[wcb]{1,3})|(_s))\.\w+$//r => catfile($TINDIR,$_)} @ctts;

    #grab files discriminated for each modeltypes
    @{$HYPOTHESIS{b}} = (grep {/_([wc]{0,2}b{1}[wc]{0,2})\.\w+$/} @ctts);
    @{$HYPOTHESIS{b}} = map {s/_[wcb]{1,3}\.\w+$//r } @{$HYPOTHESIS{b}};
    @{$HYPOTHESIS{c}} = (grep {/_([wb]{0,2}c{1}[wb]{0,2})\.\w+$/} @ctts);
    @{$HYPOTHESIS{c}} = map {s/_[wcb]{1,3}\.\w+$//r } @{$HYPOTHESIS{c}};
    @{$HYPOTHESIS{w}} = (grep {/_([bc]{0,2}w{1}[bc]{0,2})\.\w+$/} @ctts);
    @{$HYPOTHESIS{w}} = map {s/_[wcb]{1,3}\.\w+$//r } @{$HYPOTHESIS{w}};
    #exception for site-models Trees must be named like GENENAME_s.nwk
    @{$HYPOTHESIS{s}} = (grep {/_s\.\w+$/} @ctts); 
    @{$HYPOTHESIS{s}} = map {s/_s\.\w+$//r } @{$HYPOTHESIS{s}}; #this case it is not Hypothesis but genenames

    if(!%HYPOTHESIS) {
	die "ERR:: Could not find phylogenetic trees files properly named in $TINDIR\n";
    }
}


sub createDefaultProjID {
    return PROJID.int(rand(10))."".int(rand(10))."".int(rand(10))."".int(rand(10));
}


sub dumpPromptHelp {
    %TCOLOR = disableTCOLOR();
    open(FOUT, ">", catfile(".",PHELPFILE));
    select(FOUT); #switch from default STDOUT to FOUT
    promptHelp(); #print to FOUT
    close FOUT;
    select(STDOUT); #switch back to STDOUT
    %TCOLOR = enableTCOLOR();
    print ":: Prompt help printed to file ",catfile(".",PHELPFILE),"\n";
    exit 1;
}

sub promptHelp {
    clear(0) if(select() !~ m/FOUT/);
    printLicense($0);
    print "\t",$TCOLOR{menu},"GMAP PROMPT HELP - COMMANDS\n";
    print "="x80,$TCOLOR{endcolor},"\n\n";
    print $TCOLOR{bold},"Options in [ ] - means, options are not mandatory; | - means, alternative options; ( ) - means, grouping",$TCOLOR{endcolor},"\n\n";
    print $TCOLOR{Bcyan},"NOTE: The BioPerl modules here used, require that the inicial/input tree is rooted.\nSee http://www.bioperl.org/wiki/HOWTO:Trees",$TCOLOR{endcolor},"\n\n";
    printf("%s%-10s %-15s %-20s%s\n", $TCOLOR{bold},"Command","Options","Description",$TCOLOR{endcolor});
    printf("%-10s %-15s %-20s\n", "==========","===============","======================");
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"rbl",$TCOLOR{endcolor},$TCOLOR{yellow},"","Remove or clear branch length values, when available on the tree.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"rbs",$TCOLOR{endcolor},$TCOLOR{eopen},"","Remove or clear bootstrap values, when available on the tree.",$TCOLOR{eclose});
    printf("%s%s%-25s %s%s %-20s%s\n", $TCOLOR{bold},"","(Lx|Lx,...,Lz)=(\#)y",$TCOLOR{endcolor},$TCOLOR{yellow},"Specify one or more species nodes (leafs) to be labeled with PAML symbols for branch (\#),",$TCOLOR{endcolor});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","followed by any required number eg.: L11=\#1 or L07,L09,L11=\#1.",$TCOLOR{endcolor});
    printf("%s%s%-25s %s%s %-20s%s\n", $TCOLOR{bold},"","(Ax|Ax,...,Az)=(\$|\#)y",$TCOLOR{endcolor},$TCOLOR{eopen},"Specify one or more ancestral nodes to be labeled with PAML symbols for branch (\#) or for clade(\$),",$TCOLOR{eclose});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{eopen},"","followed by any required number eg.: A11=\#1 or A07,A09,A11=\$1.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"ls",$TCOLOR{endcolor},$TCOLOR{yellow},"[path|dir]","List contents on given directory. If no argument is given, lists current directory.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"save",$TCOLOR{endcolor},$TCOLOR{eopen},"filename.nwk","Save tree with current configuration to filename.nwk. Default directory is ./".$DEMIDFLTDIR,$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"saveunr",$TCOLOR{endcolor},$TCOLOR{yellow},"filename.nwk","Save rooted tree as unrooted with current configuration to filename.nwk. Default directory is ./".$DEMIDFLTDIR,$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"savetodir",$TCOLOR{endcolor},$TCOLOR{eopen},"path|dir","Set default directory where to save (labeled) newick trees.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{eopen},"","Will change default directory ./".$DEMIDFLTDIR." to the one given as argument.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"show",$TCOLOR{endcolor},$TCOLOR{yellow},"bs|bl|b|n","Show values/labels on the tree. Options are bs: bootstraps ; bl: branch-lengths ; b: both ; n:none.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"brlen",$TCOLOR{endcolor},$TCOLOR{eopen},"n","Set branch length for display, n number of characters. By default is set to 100.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"rootunr",$TCOLOR{endcolor},$TCOLOR{yellow},"Na,Nb","Root an unrooted tree, by specifying two nodes. This command creates a new node having Na and Nb as descendents",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","and the current 'root' as ancestor. This new node will hence be located below the current 'root' and above the specified descendents.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","Nodes Na and Nb are any two nodes, leafs or ancestrals, connected to the current 'root'. eg.: rootunr A16,L01",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","Note: the word root was employed in quotes ('root'), since the node presented as most close to the left is not a true root.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","In fact, by BioPerl design, every tree must be rooted and every unrooted tree has thus an additional root node,",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","which in this case becomes evident. Hence, the root displayed is in fact, a false root.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"prune",$TCOLOR{endcolor},$TCOLOR{eopen},"(Nx|Nx,...,Nz)","Prune one or more nodes specified from phylogenetic tree, without changing the tree relationships.",$TCOLOR{eclose});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{eopen},"","Nodes can be tips (leafs) or ancestrals, in which case, the entire clade or descendent branches are removed,",$TCOLOR{eclose});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{eopen},"","including the specified ancestral node. eg.: prune A13 (ancestral) or prune L11 (leaf) or prune L11,L12,A17 (leafs and ancestral).",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"rtaxa",$TCOLOR{endcolor},$TCOLOR{yellow},"(Lx=N,...,Lz=N)","Rename phylogenetic tree taxa. This can be done with two alternative options: either by supplying an option",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"|(file.csv)","consisting of 1 or more pairs of node/taxa (Lnumber) and associated new name (N), separated by commas, as in",$TCOLOR{endcolor});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","(L01=Name1,L02=Name2,...); or an option indicating any titled CSV file listing two columns, containing at",$TCOLOR{endcolor});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","the left column, the original name and at the right column, the new name.",$TCOLOR{endcolor});
    printf("%s%s%s %s%-25s %-20s%s\n", $TCOLOR{bold},"",$TCOLOR{endcolor},$TCOLOR{yellow},"","eg.: rtaxa L11=Ted or rtaxa L11=Ed,L78=Tom,L80=Joe or rtaxa path/to/filename.csv.",$TCOLOR{endcolor}); 
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"reset",$TCOLOR{endcolor},$TCOLOR{eopen},"","Reread tree from file, thus forgetting any previous changes made.",$TCOLOR{eclose});
    printf("%-10s %-15s %-20s\n", "----------","---------------","----------------------");
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"d[one]",$TCOLOR{endcolor},$TCOLOR{yellow},"","Verifies if tree was saved, before exiting.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"?",$TCOLOR{endcolor},$TCOLOR{eopen},"","This Help.",$TCOLOR{eclose});
    printf("%-10s %-15s %-20s\n", "==========","===============","======================");

    print "\nVERSION: ".VERSION."\n";
    print $TCOLOR{menu},"="x80,$TCOLOR{endcolor},"\n\n";
    print "Type any command or press \[$TCOLOR{qunln}ENTER$TCOLOR{endcolor}\] to go back to tree.\n\n";
}


sub HELP_MESSAGE {
    print shift;
    my $citation = CITATION;
    my $pweb = PROJWEBSITE;
    my $o = join(", ",@{$OMEGAVALS{b}});
    my $k = join(", ",@{$KAPPAVALS{s}});
    my $v = VERSION;
    my $phelp = PHELPFILE;
    my @HELP = <<HELP;

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    gmap.pl - Generate an organized directory structure.
   
  $TCOLOR{bold}SYNOPSIS:$TCOLOR{endcolor}
    
    gmap.pl -A [dirfiles] -T [dirfiles] -d [location] -m [s[models],b,w,c] {-j [projname]} {-O b[x,y,...,z]} {-K s[x,y,...,z]}
    gmap.pl -A [dirfiles] -t [inittree.nwk] -d [location] -m [s[models],b,w,c] {-j [projname]} {-O c[x,y,...,z]} {-K w[x,y,...,z]}

  $TCOLOR{bold}DESCRIPTION:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | This application is intended to generate an organized directory structure, containing all
    | necessary files for the execution of codeml from PAML package. All files, alignments and phylogenetic trees,
    | can be supplied by the user in separate directories. These files need to be named specially to the
    | purpose of distributing files to the correct locations in the final directory. Final control files are obtained
    | through the templates supplied with the package, which are automatically modified according to the options selected by
    | the user and placed along with the respective alignment and phylogenetic tree.
    | A specific option allows the user to make several interactive edits in an initial rooted phylogenetic tree.
    | Among the possible editions, the most relevant is the placement of PAML specific signs for branch (\#) and clade (\$),
    | thus enabling the labeling of phylogenetic trees, required for branch, branch-site and clade models.
    | An interactive help is provided, to help the user to perform these and other modifications.
    | The resulting directory structure can then be passed into mmap.pl application for running.

  $TCOLOR{bold}OPTIONS:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | -A [algndir]    Input directory containing all the alignment files distinguished by its name.
    |                 The name for each alignment file must contain gene name abbreviation (eg.: ND5)
    |                 followed by the models indication (see option -m below) for which this alignment will be used and followed by
    |                 one or two digits indicating the PAML icode parameter for each alignment. Gene name must be separated by '_'
    |                 (underscore) from the rest. Formats supported: FASTA (.fst, .fas, .fasta) and PHYLIP (.ph, .phy, .phylip).
    |                    eg.: [algndir]/ND5_sbc0.fas ; with icode 0 (universal code) will be employed in site, branch and clade models.
    | -T [treedir]    Input directory containing all the phylogenetic tree files (in NEWICK format) to be applied in several model types
    |                 (see option -m below). These are distinguished by its name, which must be given as the name of each hypothesis to
    |                 test or otherwise, for the only case of site-models, simply the matching gene names.
    |                    eg.: [treedir]/H_bc.nwk ; the hypothesis is named as 'H' and this tree is to be employed in branch and
    |                         clade models ('b' and 'c' options are according to option -m arguments found below)
    |                    eg.: [treedir]/ND5_s.nwk ; this tree is to be used for the case of site models for ND5 gene/alignment (in option -A).
    | -t [tree.nwk]   Input single tree file, in NEWICK format, for labeling in interactive mode. Use instead of the option -T.
    |                 It enables the user to edit/label and save phylogenetic trees for multiple hypothesis testing, for branch,
    |                 branch-site and clade models. PAML tags can be inserted for branches (\#) or clades (\$).
    | -d [dir]        Input directory which will contain the final directory structure containing all input files.
    | -m [typmodels]  One or more of the following options in any order, separated by commas. eg.: -m c,s[0:1:2:3:7:8:8a],b,w.
    |    s[a:...:z]      Prepare to run the specified site models. Where a:...:z are any values/models from \(0,1,2,3,7,8,8a\),
    |                     separated by ':' and enclosed in brackets '[]'.
    |    b               Prepare to run branch models. Will run models M0, TrC and TrU.
    |    w               Prepare to run branch-site models. Will run models A (Alternative) and A1 (Null).
    |    c               Prepare to run clade models. Will run models CmC (Alternative) and M2a_rel (Null).
    |                 ----------------------- not mandatory: ----------------------
    | -j [projname]   Specify the project name. This is part of the directory structure to be created and will be 
    |                 the base directory name. Hereafter, it can be used to build additional runs for other
    |                 model types at this location. By default, a generic name will be created and used.
    | -O [b[o]:c[o]]  Define the omega values to be tested for each model in option -m. By default, the omega values used in 
    |                 the case of branch and clade models are: $o
    |                 The values can be specified for one or both model types (b or c), in any order. To specify which values
    |                 to run for each model, type as in following examples: -O b[x,y,...,z]:c[x,y,...,z] or -O c[x,y,...,z]
    | -K [s[k]:w[k]]  Define the kappa values to be tested for each model in option -m. By default, the kappa values used in
    |                 the case of branch-site and site models are: $k
    |                 The values can be specified for one or both model types (s or w), in any order. To specify which values
    |                 to run for each model, type as in following examples: -K s[x,y,...,z]:w[x,y,...,z] or -K w[x,y,...,z]
    |                 -----------------------------------
    | --help          Dump the interactive commands help to a file located in current directory, named '$phelp'.
    | -h              This help
    | -v              Application version

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2015.

  $TCOLOR{bold}CITATION AND DOWNLOAD:$TCOLOR{endcolor}
    
    $citation
    $pweb

  $TCOLOR{bold}LICENSE AND COPYRIGHT:$TCOLOR{endcolor}

    LMAP: Lightweight Multigene Analyses in PAML
    gmap.pl is part of LMAP package
    Copyright (C) 2015 Emanuel Maldonado

  gmap.pl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  gmap.pl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with gmap.pl.  If not, see <http://www.gnu.org/licenses/>.

HELP
print @HELP;
    exit 1;
}

sub VERSION_MESSAGE {
    my $v = VERSION;
    my @VERS = <<VERS;
Use option -h for help and other information.

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    gmap.pl - Generate an organized directory structure.

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2015.

VERS

print @VERS;
    exit 1;
}


