#!/usr/bin/perl -w


#################################################
# Obtain information from multiple codeml runs  #
#           ------------------                  #
#   Written by Emanuel Maldonado - Mar 2012     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  imap.pl is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  imap.pl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  imap.pl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with imap.pl.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use Data::Dumper;
use strict;
use warnings;

use Getopt::Long;
Getopt::Long::Configure ("bundling");

use File::Spec::Functions qw(:ALL);
use vars qw($OUTFILE $PAML_PROG %TCOLOR);

#My modules
use FindBin;
use lib "$FindBin::Bin/LMAPlib";
use MyUtil qw(:IPAML :CONST);
use MyPAMLInfo qw(:CODEML :CONSTANTS);

use constant {
    FINISHEDFL => "correctly_finished.txt",
    UNFINISHEDFL => "uncorrectly_finished.txt",
    VERSION => "1.0.0 Nov 20th, 2015"
};

unew();
if(wantColor()) {
    %TCOLOR = enableTCOLOR();
} else {
    %TCOLOR = disableTCOLOR();
}

my %opts;

GetOptions(
    "d=s" => \$opts{d}, 
    "R=s" => \$opts{R}, 
    "s=s" => \$opts{s}, 
    "b!" => \$opts{b}, 
    "w!" => \$opts{w}, 
    "c!" => \$opts{c}, 
    "r!" => \$opts{r}, 
    "o=s" => \$opts{o},
    "v!" => \$opts{v}, 
    "h!" => \$opts{h}, 
    ) or HELP_MESSAGE("*"x30);

%opts = cleanOpts(\%opts);

HELP_MESSAGE("") if($opts{h} || !%opts);
VERSION_MESSAGE() if($opts{v});

if(!$opts{d} || !-d $opts{d}) {
    HELP_MESSAGE("ERR:: Missing input directory: use -d option followed by path to directory.\n");
}

if(!$opts{o}) {
    HELP_MESSAGE("ERR:: Missing output file: use -o option followed by path to file.\n");
}

if(!$opts{b} && !$opts{c} && !$opts{s} && !$opts{w}) {
    HELP_MESSAGE("ERR:: Missing analyses type: use -b/-c/-w/-s options.\n");
}

if(!exists $opts{R}) { 
    $opts{R} = COMMONRESFLN; #default
} elsif (!defined $opts{R}) {
    HELP_MESSAGE("ERR:: Missing common results file: use -R option followed by filename common to all runs.\n");
}

$PAML_PROG = CODEML;
my $inputdir = $opts{d};
$inputdir =~ s:/$::;
$OUTFILE = $opts{o};

my $option;
$option = "s:".$opts{s} if($opts{s});
$option = "b" if($opts{b});
$option = "w" if($opts{w});
$option = "c" if($opts{c});

pinew($OUTFILE);
printLicense($0);
startPAMLInfo($opts{R},$inputdir,$option,$opts{r});

exit 0;


############################
################ FUNCTIONS:
#######################

sub startPAMLInfo {
    my $resfl = shift;
    my $inputdir = shift;
    my $option = shift;
    my $recurse = shift;
    my @DCOLL;
    my @content = [];
    @content = getDirContent($inputdir);
    
    my $flcnt=0;
    my @changedfls;
    my $eflcnt=0;
    my @errorfls;
    
    while(1) {
	while($_  = shift @content) { 
	    if(-f $_ && $_ =~ m/($resfl)$/) {
		print ":: Handling File: $_\n";
		if(extract_info($_,$option)) {
		    $flcnt++;
		    push @changedfls, $_;
		} else {
		    $eflcnt++;
		    push @errorfls, $_;
		}
	    } elsif (-d $_) {
		print ":: Adding Subdirectory $_\n";
		push @DCOLL, $_;
	    }
	    
	    if(!@content && $recurse) { #after all content is passed through
		print ":: No more content on current directory\n";
		while(my $currdir = shift(@DCOLL)) {  # remove at the beggining
		    print ":: Going into directory $currdir...\n";
		    @content = getDirContent($currdir);
		    last if(@content);
		    print ":: Directory $currdir empty!\n" if(!@content);
		}
	    }
	}
	last if(!@content);
    }
    
    print "\n\n:: Extraction results:\n";
    print "\n:z: $flcnt files corresponding runs were correctly finished.\n" if($flcnt);
    my $opt = $option eq 'b' ? 'BM' : $option eq 'c' ? 'CM' : $option eq 'w' ? 'BSM' : 'SM';
    my @idirs = splitdir($inputdir);
    pop @idirs; #rem models dir
    $inputdir = "";
    $inputdir = pop @idirs if(@idirs); #get previous dirname
    $inputdir .= "_" if($inputdir);
    printToFile(catfile(".",$inputdir.$opt."_".FINISHEDFL), \@changedfls) if($flcnt);
    print "\n:z: $eflcnt files corresponding runs are NOT correctly finished or do not exist.\n" if($eflcnt);
    printToFile(catfile(".",$inputdir.$opt."_".UNFINISHEDFL), \@errorfls) if($eflcnt);
    print "\n";
}

sub extract_info {
    my $ifl = shift @_;
    my $opt = shift @_;
    
    if($PAML_PROG eq CODEML) {
	extract_cml_info($ifl,$opt);
    } else {
	die "ERR:: PAML program name not correct: should be one of (...)\n";
    }
}

sub HELP_MESSAGE {
    print shift;
    my $dfltcommonresfl = COMMONRESFLN;
    my $citation = CITATION;
    my $pweb = PROJWEBSITE;
    my $v = VERSION;
    my @HELP = <<HELP;

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    imap.pl - Extract PAML information from results files.
   
  $TCOLOR{bold}SYNOPSIS:$TCOLOR{endcolor}
    
    imap.pl -d [dirstruct] [-s [models] | -b | -w | -c] -o [outfile.csv] {-R [res.filename]} {-r} 

  $TCOLOR{bold}DESCRIPTION:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | This application is intended to extract PAML (codeml) information from the results files located 
    | in an organized directory struture, which contain data regarding any models previously run.
    | The information extracted is conveniently placed in CSV tables consisting of several important parameters/values.
    | The resulting file can then be passed as input into omap.pl, for screening and final LRT calculations.

  $TCOLOR{bold}OPTIONS:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | -d [dir]        Input directory containing all and only the directories that contain the codeml files.
    | -s [0,...,n]    Extract information from the specified site models results.
    | -b              Extract information from branch models results.
    | -w              Extract information from branch-site models results.
    | -c              Extract information from clade models results.
    | -o [outfile]    Choose a name for the resulting CSV file to create or append results.
    |                 ----------------------- not mandatory: ----------------------
    | -r              Work recursively.
    | -R [filename]   Results filename common to all PAML runs. By default, is the same occurring in templates files: $dfltcommonresfl
    |                 -----------------------------------
    | -h              This help
    | -v              Application version

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2012.

  $TCOLOR{bold}CITATION AND DOWNLOAD:$TCOLOR{endcolor}
    
    $citation
    $pweb

  $TCOLOR{bold}LICENSE AND COPYRIGHT:$TCOLOR{endcolor}

    LMAP: Lightweight Multigene Analyses in PAML
    imap.pl is part of LMAP package
    Copyright (C) 2015 Emanuel Maldonado

  imap.pl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  imap.pl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with imap.pl.  If not, see <http://www.gnu.org/licenses/>.

HELP
print @HELP;
    exit 1;
}

sub VERSION_MESSAGE {
    my $v = VERSION;
    my @VERS = <<VERS;
Use option -h for help and other information.

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    imap.pl - Extract PAML information from results files.

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2012.

VERS

print @VERS;
    exit 1;
}


