#!/usr/bin/perl -w

#################################################
#          Main LMAP application                #
#           ------------------                  #
#   Written by Emanuel Maldonado - Jun 2015     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  lmap.pl is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  lmap.pl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  lmap.pl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with lmap.pl.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use Data::Dumper;
use strict;
use warnings;

use Getopt::Long;
Getopt::Long::Configure ("bundling");

use File::Spec::Functions qw(:ALL);
use vars qw(%OMEGAVALS %KAPPAVALS %TCOLOR);

#My modules
use FindBin;
use lib "$FindBin::Bin/LMAPlib";
use MyUtil qw(:LMAP :CONST);
use MyPAMLInfo qw(:ICODEML :CONSTANTS);
use MyNotify qw(:DEFAULT);

use constant {
    VERSION => "1.0.0 Nov 20th, 2015"
};



$OMEGAVALS{b} = [CODEML_OMEGA_VALS]; #defaults                                                          
$OMEGAVALS{c} = [CODEML_OMEGA_VALS]; #defaults                                                          
$KAPPAVALS{s} = [CODEML_KAPPA_VALS]; #defaults                                                          
$KAPPAVALS{w} = [CODEML_KAPPA_VALS]; #defaults  

unew();
if(wantColor()) {
    %TCOLOR = enableTCOLOR();
} else {
    %TCOLOR = disableTCOLOR();
}

my %opts;


GetOptions(
    "A=s" => \$opts{A},    
    "T=s" => \$opts{T},
    "d=s" => \$opts{d},
    "m=s" => \$opts{m},
    "j=s" => \$opts{j},
    "g=s" => \$opts{g}, #optional: #cmap.pl app specific options (g and L)
    "L=s" => \$opts{L},
    "R=s" => \$opts{R},
    "f=s" => \$opts{f},
    "p=s" => \$opts{p},
    "O=s" => \$opts{O},
    "K=s" => \$opts{K},
    "n=i" => \$opts{n},
    "e:s" => \$opts{e},
    "no-omap" => \$opts{nomap},
    "v!" => \$opts{v},
    "h!" => \$opts{h},
    ) or HELP_MESSAGE("*"x30);

%opts = cleanOpts(\%opts);

HELP_MESSAGE("") if($opts{h} || !%opts);
VERSION_MESSAGE() if($opts{v});

if(!$opts{A} || !-d $opts{A}) {
    HELP_MESSAGE("ERR:: Missing input directory containing alignments: use -A option followed by path to directory.\n");
}

if(!$opts{T} || !-d $opts{T}) {
    HELP_MESSAGE("ERR:: Missing input directory containing newick trees: use -T option followed by path to directory.\n");
}

if(!$opts{d} || !-d $opts{d}) {
    HELP_MESSAGE("ERR:: Missing input directory where to locate the directory struture: use -d option followed by path to directory.\n");
}

if(!$opts{m}) {
    HELP_MESSAGE("ERR:: Missing input specification of methods to be prepared: use -m followed by required method specification\n");
}

if(defined $opts{m}) {
    if($opts{m} !~ m/^(b|w|c|(s\[([0-8]|8a)(:([0-8]|8a))*\]))(,(b|w|c|(s\[([0-8]|8a)(:([0-8]|8a))*\])))*$/) {
	HELP_MESSAGE("ERR: Models specification should be separated by commas and\nfor site models, models are separated by colon and in brackets.\n");
    }
}

if(!$opts{j}) {
    HELP_MESSAGE("ERR:: Missing project name: use -j followed by name\n");
}


######not mandatory

if(defined $opts{O}){
    my $o = $opts{O};
    my $one = ($o =~ s/b//g);
    my $two = ($o =~ s/c//g);
    if($one > 1 || $two > 1) {
	HELP_MESSAGE("ERR: There should be only one occurrence for each method branch (b) and/or clade (c) with specified values.\n");
    }
    if($opts{O} !~ m/^([bc]\[(\d+(\.\d+)?)(,\d+(\.\d+)?)*\])(\:([bc]\[(\d+(\.\d+)?)(,\d+(\.\d+)?)*\]))?$/) { 
	HELP_MESSAGE("ERR:: Option -O argument is not correctly specified: should be one or both of b[x,...,z] or c[x,...,z] were x,...,z are any values.\n");
    }
}
if(defined $opts{K}){
    my $k = $opts{K};
    my $one = ($k =~ s/s//g);
    my $two = ($k =~ s/w//g);
    if($one > 1 || $two > 1) {
	HELP_MESSAGE("ERR: There should be only one occurrence for each method branch-site (w) and/or site (s) with specified values.\n");
    }
    if($opts{K} !~ m/^([sw]\[(\d+(\.\d+)?)(,\d+(\.\d+)?)*\])(\:([sw]\[(\d+(\.\d+)?)(,\d+(\.\d+)?)*\]))?$/) { 
	HELP_MESSAGE("ERR:: Option -K argument is not correctly specified: should be one or both of s[x,...,z] or w[x,...,z] were x,...,z are any values.\n");
    }
}

if(exists $opts{g} xor exists $opts{L}) { #enable cmap.pl
    HELP_MESSAGE("ERR:: To include/use cmap.pl application, please specify both options -g and -L.\n");
}

if(exists $opts{g}) {
    if(!$opts{g}) {
	HELP_MESSAGE("ERR:: Target parameters/lines identification required: use option -g with argument.\n");
    }
}

if(exists $opts{L}) {
    if(!$opts{L}) {
	HELP_MESSAGE("ERR:: Text/value required for replacement in parameters: use -L option followed by text in quotes.\n");
    }
}

if(!exists $opts{R}) { 
    $opts{R} = COMMONRESFLN; #default
} elsif (!defined $opts{R}) {
    HELP_MESSAGE("ERR:: Missing common results file: use -R option followed by filename common to all runs.\n");
}

if(!exists $opts{f}) {
    $opts{f} = CODEMLCTRLOPTF; #default
} elsif (!defined $opts{f}) {
    HELP_MESSAGE("ERR:: Argument missing in option -f. Please, see below. \n");
}

if(!exists $opts{p}) {
    $opts{p} = CODEMLDFLTOPTP; #default
} elsif(!defined $opts{p}) {
    HELP_MESSAGE("ERR:: Argument missing in option -p. Please see below. \n");
}

if(defined $opts{n}) {
    if(!int($opts{n})) {
        HELP_MESSAGE("ERR:: Argument of -n option is not a number ( > 0).\n");
    }
}

if(defined $opts{e}) {
    if(!isValidAddr($opts{e})) {
        HELP_MESSAGE("ERR:: Email address specified was not found correct, please try again...\n");
    }
}

printLicense;
my $projloc = catdir($opts{d},$opts{j});
my $thisapp = "lmap.pl";
setCookyLoc($projloc);

$SIG{INT} = 'IGNORE';
$SIG{QUIT} = 'IGNORE';
$SIG{TSTP} = 'IGNORE';

########################## LMAP WORKFLOW:
####### GMAP.pl

my $refapp = "gmap.pl";

my $gmap = "$FindBin::Bin/$refapp -A $opts{A} -T $opts{T} -d $opts{d} -m $opts{m} -j $opts{j} ";
$gmap .= "-O $opts{O} " if(defined $opts{O});
$gmap .= "-K $opts{K}" if(defined $opts{K});

print ":z: Executing: $gmap \n\n";
system($gmap);
if($? == -1) {
    die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
} elsif ($? >> 8 == 1) {
    print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
    exit 1;
} elsif($? >> 8 == 0) {
    print "::The program $refapp finished with success!\n\n";
} elsif($? & 127) {
    printf "::$refapp unexpectedly terminated with signal %d, %s coredump\n",($? & 127),($? & 128) ? 'with':'without';
    exit 1;
} else {
    die "ERR:: $refapp unknown error: $! - $?\n";
}

####### CMAP.pl

$refapp = "cmap.pl";
if(defined $opts{g} && defined $opts{L}) { #not mandatory; run by user choice    
    my $cmap = "$FindBin::Bin/$refapp -d $projloc -f $opts{f} -r ";
    $cmap .= "-g $opts{g} " if(defined $opts{g});
    $cmap .= "-L $opts{L}" if(defined $opts{L});
    
    print ":z: Executing: $cmap \n\n";
    system($cmap);
    if($? == -1) {
	die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
    } elsif ($? >> 8 == 1) {
	print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
	exit 1;
    } elsif($? >> 8 == 0) {
	print "::The program $refapp finished with success!\n\n";
    } elsif($? & 127) {
	printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
	exit 1;
    } else {
	die "ERR:: $refapp unknown error: $! - $?\n";
    }
}

modifyLMAPCooky([getModels($opts{m})]);
####### MMAP.pl

$refapp = "mmap.pl";

my $mmap = "$FindBin::Bin/$refapp -d $projloc -R $opts{R} -f $opts{f} -p $opts{p} -r -l ";
$mmap .= "-n $opts{n} " if(defined $opts{n});
$mmap .= "-e " if(exists $opts{e});
$mmap .= "$opts{e} " if(defined $opts{e});

print ":z: Executing: $mmap \n\n";
system($mmap);
if($? == -1) {
    die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
} elsif ($? >> 8 == 1) {
    print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
    exit 1;
} elsif($? >> 8 == 0) {
    print "::The program $refapp finished with success!\n\n";
} elsif($? & 127) {
    printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
    exit 1;
} else {
    die "ERR:: $refapp unknown error: $! - $?\n";
}


####### IMAP.pl (extract from every specified method)

my %imapFiles = (c => '', w => '', s => '', b =>'');

$refapp = "imap.pl";
my ($outputfl,$dirlocation);
my $imap = "$FindBin::Bin/$refapp -R $opts{R} -r ";
if ($opts{m} =~ m/,?s(\[[0-8:a]+\]),?/) {
    my $l = $1;
    $outputfl = $opts{j}."_SMresults.csv";
    $dirlocation = catdir($projloc,$C_MTYPESTAGS{s});
    $l =~ s/:/,/g;
    $l =~ s/\[|\]//g;
    my $imaps = $imap." -d $dirlocation -o $outputfl -s $l";
    print ":z: Executing: $imaps \n\n";
    system($imaps);
    if($? == -1) {
	die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
    } elsif ($? >> 8 == 1) {
	print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
	exit 1;
    } elsif($? >> 8 == 0) {
	print "::The program $refapp finished with success!\n\n";
	$imapFiles{s} = $outputfl;
    } elsif($? & 127) {
	printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
	exit 1;
    } else {
	die "ERR:: $refapp unknown error: $! - $?\n";
    }
}

if ($opts{m} =~ m/,?w,?/) {
    $outputfl = $opts{j}."_BSMresults.csv";
    $dirlocation = catdir($projloc,$C_MTYPESTAGS{w});
    my $imapw = $imap." -d $dirlocation -o $outputfl -w ";
    print ":z: Executing: $imapw \n\n";
    system($imapw);
    if($? == -1) {
	die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
    } elsif ($? >> 8 == 1) {
	print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
	exit 1;
    } elsif($? >> 8 == 0) {
	print "::The program $refapp finished with success!\n\n";
	$imapFiles{w} = $outputfl;
    } elsif($? & 127) {
	printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
	exit 1;
    } else {
	die "ERR:: $refapp unknown error: $! - $?\n";
    }
}

if ($opts{m} =~ m/,?b,?/) {
    $outputfl = $opts{j}."_BMresults.csv";
    $dirlocation = catdir($projloc,$C_MTYPESTAGS{b});
    my $imapb = $imap." -d $dirlocation -o $outputfl -b ";
    print ":z: Executing: $imapb \n\n";
    system($imapb);
    if($? == -1) {
	die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
    } elsif ($? >> 8 == 1) {
	print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
	exit 1;
    } elsif($? >> 8 == 0) {
	print "::The program $refapp finished with success!\n\n";
	$imapFiles{b} = $outputfl;
    } elsif($? & 127) {
	printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
	exit 1;
    } else {
	die "ERR:: $refapp unknown error: $! - $?\n";
    }
}


if ($opts{m} =~ m/,?c,?/) {
    $outputfl = $opts{j}."_CMresults.csv";
    $dirlocation = catdir($projloc,$C_MTYPESTAGS{c});
    my $imapc = $imap." -d $dirlocation -o $outputfl -c ";
    print ":z: Executing: $imapc \n\n";
    system($imapc);
    if($? == -1) {
	die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
    } elsif ($? >> 8 == 1) {
	print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
	exit 1;
    } elsif($? >> 8 == 0) {
	print "::The program $refapp finished with success!\n\n";
	$imapFiles{c} = $outputfl;
    } elsif($? & 127) {
	printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
	exit 1;
    } else {
	die "ERR:: $refapp unknown error: $! - $?\n";
    }
}

print "\n::$refapp information extracted into:\n";
foreach (keys %imapFiles) {
    if($imapFiles{$_}) {
	print "Branch models results in file: <$imapFiles{$_}>\n" if($_ eq 'b');
	print "Site models results in file: <$imapFiles{$_}>\n" if($_ eq 's');
	print "Branch-Site results models in file: <$imapFiles{$_}>\n" if($_ eq 'w');
	print "Clade models results in file: <$imapFiles{$_}>\n" if($_ eq 'c');
    }
}
print "\n"x2;
if(exists $opts{nomap}) { #premature end of workflow: by user choice
    print "\n:z: LMAP finished successfully!\n";
    exit 0;
}


####### OMAP.pl

$refapp = "omap.pl";
my $inputfl = -e $imapFiles{s} ? $imapFiles{s} : -e $imapFiles{w} ? $imapFiles{w} : -e $imapFiles{b} ? $imapFiles{b} : -e $imapFiles{c} ? $imapFiles{c} : "notfound";
$outputfl = $opts{j}."_FinalResults.csv";

if($inputfl eq "notfound") {
    die "ERR:: Could not run omap.pl: imap.pl result files not found!\n";
}

my $omap = "$FindBin::Bin/$refapp -i $inputfl -o $outputfl";
print ":z: Executing: $omap \n\n";
system($omap);
if($? == -1) {
    die "ERR:: An error occured during the initialization of the $refapp program: $!\n";
} elsif ($? >> 8 == 1) {
    print "ERR:: $thisapp workflow interrupted: The program $refapp did not finish correctly!\n";
    exit 1;
} elsif($? >> 8 == 0) {
    print "::The program $refapp finished with success!\n\n";
} elsif($? & 127) {
    printf "ERR:: %s unexpectedly terminated with signal %d, %s coredump\n",$refapp,($? & 127),($? & 128) ? 'with':'without';
    exit 1;
} else {
    die "ERR:: $refapp unknown error: $! - $?\n";
}

print "\n:z: LMAP finished successfully!\n";
exit 0;


###############################

sub getModels {
    my $models = shift;
    $models =~ s/s\[(.)+\]/s/; #site models only need s                                                                                                                                                                                                                       
    my @models = split(",",$models);
    return @models;
}


sub HELP_MESSAGE {
    print shift;
    my $efpath = getPBinPath();
    my $dfltcommonresfl = COMMONRESFLN;
    my $ctrlfl = CODEMLCTRLFLN;
    my $dfltoptf = CODEMLCTRLOPTF;
    my $dfltoptp = CODEMLDFLTOPTP;
    my $citation = CITATION;
    my $pweb = PROJWEBSITE;
    my $o = join(", ",@{$OMEGAVALS{b}});
    my $k = join(", ",@{$KAPPAVALS{s}});
    my $v = VERSION;
    my @HELP = <<HELP;

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    lmap.pl - Lightweight Multigene Analyses in PAML.
   
  $TCOLOR{bold}SYNOPSIS:$TCOLOR{endcolor}
    
    lmap.pl -A [dirfiles] -T [dirfiles] -d [location] -m [s[models],b,w,c] -j [projname] \
            {-f [e:ctl]} {-p [l:codeml]} {-R [resfilename]} {-n [integer]} {-e {emailaddr}} {-O [b[x,y,...,z]]} {-K [s[x,y,...,z]]}

  $TCOLOR{bold}DESCRIPTION:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | This is the main program of LMAP package and is a wrapper of all LMAP applications in the following order:
    |        'gmap.pl > (cmap.pl >) mmap.pl > imap.pl (> omap.pl)'
    |    (Further information can be found in any of the help messages of LMAP applications by using option -h).
    | The purpose is to run all the LMAP applications using only one program. The application omap.pl will always require 
    | intervention from the user to proceed with specific operations. At this point omap.pl will be running, ready to be used.
    | The cmap.pl application can be enabled (included in the workflow), by using two extra options (-g and -L).
    | The omap.pl application can be disabled (excluded from the workflow), by specifying the option --no-omap.
    | The use of lmap.pl does not hinder or entirely replace the use of the remaining applications.
    | Please note that many options from the LMAP applications (g/c/m/i/omap.pl) are not available in this context and 
    | some have become mandatory in this program, for example, option -j.

  $TCOLOR{bold}OPTIONS:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | -A [algndir]    Input directory containing all the alignment files distinguished by its name.
    |                 The name for each alignment file must contain gene name abbreviation (eg.: ND5)
    |                 followed by the models indication (see option -m below) in which this alignment will be used and followed by
    |                 one or two digits indicating the PAML icode parameter for each alignment. Gene name must be separated by '_'
    |                 (underscore) from the rest. Formats supported: FASTA (.fst, .fas, .fasta) and PHYLIP (.ph, .phy, .phylip).
    |                    eg.: [algndir]/ND5_sbc0.fas ; with icode 0 (universal code) will be employed in site, branch and clade models.
    | -T [treedir]    Input directory containing all the phylogenetic tree files (in NEWICK format) to be applied in several model types
    |                 (see option -m below). These are distinguished by its name, which must be given as the name of each hypothesis to
    |                 test or otherwise, for the only case of site-models, simply the matching gene names.
    |                    eg.: [treedir]/H_bc.nwk ; the hypothesis is named as 'H' and this tree is to be employed in branch and
    |                         clade models ('b' and 'c' options are according to option -m arguments found below).
    |                    eg.: [treedir]/ND5_s.nwk ; this tree is to be used for the case of site models for ND5 gene/alignment (in option -A).
    | -d [dir]        Input directory which will contain the final directory structure containing all input files.
    | -m [typmodels]  One or more of the following options in any order, separated by commas. eg.: -m c,s[0:1:2:3:7:8:8a],b,w.
    |    s[a:...:z]      Prepare to run the specified site models. Where a:...:z are any values/models from \(0,1,2,3,7,8,8a\),
    |                     separated by ':' and enclosed in brackets '[]'.
    |    b               Prepare to run branch models. Will run models M0, Trc and Tru.
    |    w               Prepare to run branch-site models. Will run models A (Alternative) and A1 (Null).
    |    c               Prepare to run clade models. Will run models CmC (Alternative) and M2a_rel (Null).
    | -j [projname]   Specify the project name. This is part of the directory structure to be created and will be 
    |                 the base directory name. Hereafter, it can be used to build additional runs for other
    |                 model types at this location.
    |                 ----------------------- not mandatory: ----------------------
    | -R [filename]   Results filename common to all PAML runs. By default, is the same occuring in templates files: $dfltcommonresfl
    | -f [tag:part]   Choose control files through a given filename portion, where tag is one of:
    |                      f: filename equals given part
    |                      s: filename starts with part
    |                      e: filename ends with part
    |                 By default, is $dfltoptf based on the common codeml control filename: $ctrlfl 
    | -p [tag:prog]   Program to run, where tag is one of:
    |                      l: from default location eg.: l:progname
    |                      f: given full path eg.: f:path/to/progname
    |                 For tag 'l' the location defaults to $ENV{HOME}/bin/ or the one chosen by the user during installation.
    |                 Use tag 'f' to give a different location for the intended PAML executable(s).
    |                 By default, is $dfltoptp based on the default PAML executables location : $efpath
    | -O [b[o]:c[o]]  Define the omega values to be tested for each model in option -m. By default, the omega values used in 
    |                 the case of branch and clade models are: $o
    |                 The values can be specified for one or both model types (b or c), in any order. To specify which values
    |                 to run for each model, type as in following examples: -O b[x,y,...,z]:c[x,y,...,z] or -O c[x,y,...,z]
    | -K [s[k]:w[k]]  Define the kappa values to be tested for each model in option -m. By default, the kappa values used in
    |                 the case of branch-site and site models are: $k
    |                 The values can be specified for one or both model types (s or w), in any order. To specify which values
    |                 to run for each model, type as in following examples: -K s[x,y,...,z]:w[x,y,...,z] or -K w[x,y,...,z]
    | -n [integer]    Specify the maximum number of CPUs/cores available (limit) to use for the current task.
    |                 By default will try to use the aproximate maximum number of cores available.
    |                 This option can be usefull to enable sharing of the CPU capacity between users or different (mmap) tasks.
    | -e {emailaddr}  Enable email notification, when mmap.pl finishes. The argument to this option, is optional and has two meanings:
    |                 If email address is not specified, it will default to the address defined during installation. Otherwise, it
    |                 will be given preference to the address here supplied. In case the address is not given in either case
    |                 (during installation and in this option), notification will not be sent.
    | -g [tag:ptext]  [Enable cmap.pl application (use both -g and -L)].
    |                 Control file target parameter line identified by its name eg.: -g f:icode
    |                 Where tag is one of the following:
    |                     f: Complete/full line
    |                     s: Starting with text  (may affect several lines/parameters)
    |                     c: Containing text     (may affect several lines/parameters)
    |                     e: Ending with text    (may affect several lines/parameters)
    | -L [paramval]   [Enable cmap.pl application (use both -g and -L)].
    |                 Value or text to be replaced in selected line(s)/parameter(s) from option -g.
    | --no-omap       Disable the execution of omap.pl application.
    |                 -----------------------------------
    | -h              This help
    | -v              Script version

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Jun 2015.

  $TCOLOR{bold}CITATION AND DOWNLOAD:$TCOLOR{endcolor}
    
    $citation
    $pweb

  $TCOLOR{bold}LICENSE AND COPYRIGHT:$TCOLOR{endcolor}

    LMAP: Lightweight Multigene Analyses in PAML
    lmap.pl is part of LMAP package
    Copyright (C) 2015 Emanuel Maldonado

  lmap.pl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  lmap.pl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with lmap.pl.  If not, see <http://www.gnu.org/licenses/>.

HELP
print @HELP;
    exit 1;
}

sub VERSION_MESSAGE {
    my $v = VERSION;
    my @VERS = <<VERS;
Use option -h for help and other information.

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}
    
    lmap.pl - Lightweight Multigene Analyses in PAML.

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Jun 2015.

VERS

print @VERS;
    exit 1;
}

