#!/usr/bin/perl -w 

#################################################
# Schedule a batch of runs for codeml           #
#           ------------------                  #
#   Written by Emanuel Maldonado - Jan 2014     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  mmap.pl is part of LMAP package
#  Copyright (C) 2015  Emanuel Maldonado
#
#  mmap.pl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  mmap.pl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with mmap.pl.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################

use Data::Dumper;
use strict;
use warnings;
use Getopt::Long;
Getopt::Long::Configure ("bundling");

use Sys::Info;
use Sys::Info::Constants qw( :device_cpu );
use threads;
use threads::shared;
use Thread::Semaphore;
use sigtrap qw(handler do_rexit INT QUIT TERM TSTP);
use Term::ReadKey;

use Cwd;
use File::Spec::Functions qw(:ALL);
use File::Copy;

#My modules
use FindBin;
use lib "$FindBin::Bin/LMAPlib";
use MyPAMLInfo qw(:CONSTANTS :ICODEML);
use MyNotify qw(:DEFAULT);
use MyUtil qw(:RPAML :CONST);

use vars qw ($EFPATH $SCNUM $thr $thrKey $thrPE $FKEY $RTKEY $TERM $EFL $CMNRESFL $semaphore $FTSTATUSFL
$WORKINGUSER %TCOLOR %COMPLETED %TIMEUSED @SCHEDULEDNXT %TITLETOFILE $MPID $PALL $INCSCNUM @NSCHNXT);

use constant SCREENP => qw(screen -d -m -L); #-L for automatic output logging
use constant {
    SHOWMAX => 10,
    STARTTIME => scalar(localtime()),
    MPIDTAG => "MPID:",
    VERSION => "1.0.1 Apr 24th, 2017"
};

$semaphore = Thread::Semaphore->new(); 

#Thread
share(%COMPLETED); 
share(%TIMEUSED); 
share($RTKEY);
share($FKEY);
share($TERM);
share($PALL);
share($INCSCNUM);
$FKEY = 1;
$TERM = 0; #signal-handler
$thr = "";
$thrKey = "";

$MPID = createMMAPID();
$PALL = 1; #show/hide other MPIDs ; by default show (=1)
$INCSCNUM = 0;

unew();
$EFPATH = getPBinPath();
if(wantColor()) {
    %TCOLOR = enableTCOLOR();
} else {
    %TCOLOR = disableTCOLOR();
}

my %opts;

GetOptions(
    "d=s" => \$opts{d}, 
    "R=s" => \$opts{R}, 
    "f=s" => \$opts{f}, 
    "p=s" => \$opts{p}, 
    "x!" => \$opts{x}, 
    "r!" => \$opts{r}, 
    "n=i" => \$opts{n},
    "l!" => \$opts{l}, 
    "e:s" => \$opts{e},
    "v!" => \$opts{v}, 
    "h!" => \$opts{h}, 
    ) or HELP_MESSAGE("*"x30);

%opts = cleanOpts(\%opts);

HELP_MESSAGE("") if($opts{h} || !%opts);
VERSION_MESSAGE() if($opts{v});

if(!$opts{d}) {
    HELP_MESSAGE("ERR:: Missing input directory: use -d option followed by path to directory.\n");
}

my $inputdir = $opts{d};
die "ERR:: [$inputdir] is not a directory!" if(!-d $inputdir);
$inputdir =~ s:/$::g;
$FTSTATUSFL = "mmapFinalTaskStatus_".pathtoTitle($inputdir).".csv";


if(!exists $opts{f}) {
    $opts{f} = CODEMLCTRLOPTF; #default
} elsif (!defined $opts{f}) {
    HELP_MESSAGE("ERR:: Argument missing in option -f. Please, see below. \n");
}

if(!exists $opts{p}) {
    $opts{p} = CODEMLDFLTOPTP; #default
} elsif(!defined $opts{p}) {
    HELP_MESSAGE("ERR:: Argument missing in option -p. Please see below. \n");
}

my ($ptag,$prog) = split(/:/,$opts{p});
die "ERR:: Unrecognized pattern [ $ptag ] in option -p !!\n" unless ($ptag eq "l" || $ptag eq "f");

if($ptag eq "f") {
    die "ERR:: Specified path [$prog] does not point to executable file..." if(!-f $prog || !-e $prog || !-B $prog);
    $EFPATH = (splitpath($prog))[1];
    die "ERR:: Could not access path to executable file [$EFPATH] ..." if(!-d $EFPATH);
    $prog = (splitpath($prog))[2];
}

#common results file name
if(exists $opts{R}) {
    if(!defined $opts{R}) {
	HELP_MESSAGE("ERR:: Missing common results file: use -R option followed by filename only.\n");
    }
}

$CMNRESFL = $opts{R} || COMMONRESFLN;


#common expression to all codeml control files names
my ($ttag,$fltag) = split(/:/,$opts{f});
die "ERR:: Unrecognized pattern [ $ttag ] in option -t !!\n" unless ($ttag eq "s" || $ttag eq "e" || $ttag eq "f");

#calculate how many might be already running from another hypothetical mmap or other
my $scnt = scalar(getScreenList(0));
print ":z: Found other $scnt screen instances already running.\n" if($scnt);

my $info = Sys::Info->new;
my $cpu = $info->device( CPU => my %options ); 

#** No CPU/cores are currently available to run...try again later or terminate running programs...
if(defined $opts{n}) {
    if(!int($opts{n})) {
	HELP_MESSAGE("ERR:: Argument of -n option is not a number ( > 0).\n");
    }
} else {
    $opts{n} = 0; #option -n was not specified : dummy value
}

#enable aproxmate better cpu-auto-scheduling based on current cpuload and/or screens running so as to determine aprox num of free CPUs/cores
my $cpuload = int($cpu->load) < $cpu->load ? int($cpu->load) +1 : int($cpu->load); #load ceiling
#other critical programs may be running beyond MMAP screens, so this is an attempt to find "load balance"
my $htorcores = $cpu->ht ? ($cpu->count < $cpu->ht ? $cpu->ht : $cpu->count) : $cpu->count; #avoid possible differences in different operating systems, since ht maybe diff from cores count
my $aproxcoresloaded = $scnt >= $cpuload ? $scnt : $cpuload; #decide which one is bigger <-- ceiling on both scnt; hence telling if the load is due to screens or something else
my $cpucnt = $aproxcoresloaded >= $htorcores ? 0 : $htorcores; #if in case aproxcoresloaded exceeds num cores then 0 cpus else keep cpu cnt

#give user specified cpus/runs, or;
#default: use the calculated aprox the num of free cores, based on num screens possibly running and other unknown tasks
$SCNUM = int($opts{n}) || #user
    ($cpucnt ? $cpucnt - $aproxcoresloaded : 0) || #default:give aprox free num cpus OR if aproxcoresloaded does exceeds total num cores then give 0 to die
    die "ERR:: No valid number of CPU cores selected (current CPU load: $aproxcoresloaded). Recommended to use option -n.\n";

#enable file log
$EFL = 0;
$EFL = 1 if($opts{l});

#validate email address
if(defined $opts{e}) {
    if(!isValidAddr($opts{e})) {
	HELP_MESSAGE("ERR:: Email address specified was not found correct, please try again...\n");
    }
    nnew($opts{e},catfile(getcwd(),""),$FTSTATUSFL,$MPID,STARTTIME,$EFL);
    $WORKINGUSER = getEmailTo(); #either file or option
}

printLicense($0);
my @scheduledfls = readinDirectoryStructure($inputdir,$opts{f},$opts{r},$opts{x});
startMMAP($prog,$htorcores,\@scheduledfls);
modifyLMAPCooky2($inputdir);

if($WORKINGUSER) {
    send();
}

exit 0;



############################
################ FUNCTIONS:
#######################

sub startMMAP { #MAIN
    my $pname = shift;
    my $htorcores = shift;
    my @scheduledfls = @{$_[0]};
    $NSCHNXT[0] = $NSCHNXT[1] = @scheduledfls; #1 fixed 0 variable
    my $flcnt = 0;
    my $scmon = 0;
    my $once = 1;
    my $schfl;
    my $runcnt;
    my $i = 0;
    for(;;) {
	$runcnt = scalar(grep {/^0$/} values %COMPLETED); #runs  0: running ; runs >0: finished
	
	while($i < @scheduledfls && $runcnt != $SCNUM) { #put as many running as needed to fill the SCNUM 'bin'
	    $schfl = $scheduledfls[$i];
	    if(-f $schfl) {
		if(!exists $COMPLETED{$schfl}) {
		    print ":: Starting execution for file: $schfl\n";
		    if(startProgram($schfl,$EFPATH.":".$pname) == 0) {
			$flcnt++;
			$i++;
			$NSCHNXT[0]--;
		    }
		}
	    } else {
		print "ERR:: Something wrong... $schfl is not a file!\n";
		next;
	    }
	    if(@scheduledfls - $i >= SHOWMAX) { #groups of SHOWMAX at most
		@SCHEDULEDNXT = @scheduledfls[$i..$i+SHOWMAX];
	    } elsif (@scheduledfls - $i > 0) { #or at least groups of whatever under SHOWMAX
		@SCHEDULEDNXT = @scheduledfls[$i..$#scheduledfls];
	    } else {
		@SCHEDULEDNXT = ();
	    }
	    $runcnt = scalar(grep {/^0$/} values %COMPLETED); 
	}

	if($TERM) { #signal handler
	    processExit();
	} else {
	    if($once) {
		$once = 0;
 		$thrKey = threads->create('getRTKey');
		$thr = threads->create({'context'=>'list'},'monitor',$SCNUM,$pname,$thrKey,1);
	    }
	    if(!defined $thrKey || !$thrKey->is_running()) {
		$thrKey = threads->create('getRTKey');
	    }
	    if(!defined $thr || !$thr->is_running()) {
		$thr = threads->create({'context'=>'list'},'monitor',$SCNUM,$pname,$thrKey,1);
	    }
	    if(!$TERM) {
		($scmon,$thrKey) = $thr->join();
	    } else {
		$thr->join();
	    }
	}
	if($INCSCNUM != 0) {
	    if($SCNUM + $INCSCNUM > 0 || $SCNUM + $INCSCNUM <= $htorcores) {
		$SCNUM += $INCSCNUM;
	    }
	    $INCSCNUM = 0;
	}
	next if(!defined $scmon || $TERM);
	last if($scmon == 0 && $i >= @scheduledfls); #finish when zero are running and no more to run
    }
    finalize();
    printHeader();
    $runcnt = scalar(grep {/^0$/} values %COMPLETED);
    if($runcnt == 0) {
	print "\n:z: All executions are finished successfully!\n";
    } elsif($TERM) {
	print "\n:z: MMAP interrupted: There were executions not terminated...\n";
    }
    print "\n:z: A total of $flcnt / $NSCHNXT[1] files were started in specified program:\n";
    printFinalStatus();
}

sub finalize {
    foreach my $thr (threads->list()) {
	if($thr->is_joinable()) {
	    $thr->join();
	}
	if($thr->is_running()) {
	    $thr->kill('KILL')->detach();
	}
    }
    ReadMode 0;
}

sub do_rexit { #signal handler
    print(":z: Caugth signal to quit... Press '9' to continue... \n"); 
    $TERM = 1;
}

sub processExit {
    local $SIG{INT} = 'IGNORE';
    local $SIG{QUIT} = 'IGNORE';
    local $SIG{TERM} = 'IGNORE';
    local $SIG{TSTP} = 'IGNORE';
    print "\n";
    my $sccnt = scalar(getScreenList(1));
    $semaphore->down();
    if($sccnt) {
	my $ans = processExit_QMD();
	if($ans =~ m/^[dD]/) {
	    $TERM = 0;
	    $semaphore->up();
	    return;
	}
	if($ans =~ m/^[mM]/) {
	    $sccnt = processManager();
	}
    }
    my $ans = processExit_ConfirmQ($sccnt);
    if($ans) {
	chomp $ans;
	if($ans =~ m/^[yY]/) {
	    if($sccnt) {
		processExit_TPP();
	    }
	    print "\nLeaving... Goodbye!\n"; 
	    finalize();
	    exit 1;
	}
    }
    $TERM = 0;
    $semaphore->up();
}

sub processExit_QMD {
    printHeader();
    print $TCOLOR{menu},"QUIT/MANAGER screen - Quit or Manager?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor};
    print "\n";
    my $ans = 0;
    do {
	print "\tDo you want to Quit or access Process Manager? (Type [$TCOLOR{qunln}Q$TCOLOR{endcolor}]uit/[$TCOLOR{qunln}M$TCOLOR{endcolor}]anager/[$TCOLOR{qunln}D$TCOLOR{endcolor}]ismiss)\n\t> ";
	$ans = <STDIN>;
	$ans =~ s/\s+//g;
    } while ($ans !~ m/^[dDqQmM]/);
    return $ans;
}

sub processExit_ConfirmQ {
    my $sccnt = shift;

    printHeader();
    print $TCOLOR{menu},"QUIT/MANAGER screen - Really Quit?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor};
    print "\n";
    my $ans = "";
    do {
	if($sccnt) {
	    print "\tThere are $sccnt programs still not terminated, do you really want to quit? (Type [$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	} else {
	    print "\tDo you really want to quit? (Type [$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	}
	$ans = <STDIN>;
	$ans =~ s/\s+//g;
    } while($ans !~ m/^[yYnN]/);
    return $ans;
}

sub processExit_TPP {
    printHeader();
    print $TCOLOR{menu},"QUIT/MANAGER screen - Pending Jobs\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor};
    print "\n";
    my $ans = "";
    do {
	print "\tDo you want to terminate pending programs? ([$TCOLOR{qunln}A$TCOLOR{endcolor}]ll/[$TCOLOR{qunln}S$TCOLOR{endcolor}]pawned (with ".$MPID." only)/[$TCOLOR{qunln}N$TCOLOR{endcolor}]one)\n\t> ";
	$ans = <STDIN>;
	$ans =~ s/\s+//g;
    } while($ans !~ m/^[AaSsNn]/);
    if($ans =~ m/^[AaSs]/) {
	killScreens($ans);
    }
}

sub killScreens {
    my $ans = shift;

    my @list = getScreenList(1);

    my $MPIDtag = MPIDTAG;
    my $MPIDWtag = $MPIDtag.$MPID;

    if($ans =~ m/^[sS]/) { #terminate only this tagged MPID
	@list = grep {/$MPIDWtag/} @list;
	print "\t MMAPID $MPID not found!" and return if(!@list);
    }

    foreach (@list)  {
	my ($pid, undef) = split(/\.$MPIDtag/,$_,2);
	$pid =~ s/\s+//g;
	$_ =~ s/^\s*//;
	print "\t[ $_ ]\n\t :Successfully terminated!\n" if(kill('KILL',$pid));
    }
    qx/screen -wipe/;
}

sub processManager {
    my $ans = "";
    my $sccnt;
    my $MPIDtag = MPIDTAG;
    do {
	my @sclist = getScreenList(1);
	$sccnt = scalar(@sclist);
	printHeader();
	print $TCOLOR{menu},"PROCESS MANAGER screen - Group or Single?\n",$TCOLOR{endcolor};
	print $TCOLOR{menu},"","="x23,"\n",$TCOLOR{endcolor};
	print "\n";

	print "\t","x"x80,"\n";
	print $TCOLOR{tblttl},"\tMMAPID\t\tPROCID\t\tFILE\t\t\t\t\t(",$sccnt,")$TCOLOR{endcolor}\n";
	print "\t","x"x80,"\n";
	
	#find all IDs present
	my %ids;
	foreach my $s (@sclist) {
	    if($s =~ m/(\d+)\.($MPIDtag([A-Z]{2}[0-9]{4}):.*)\s+\(\d+/) {
		if(!$TITLETOFILE{$2}) { #1: PID; 2: FILE; 3: MPID
		    $ids{$1} = $3."\t\t".$1."\t\tOther";
		} else {
		    $ids{$1} = $3."\t\t".$1."\t\t".$TITLETOFILE{$2}; #PID => MPID, $title
		}
	    }
	}
	foreach my $v (values %ids) {
	    if($v =~ m/$MPID/) {
		print "\t$TCOLOR{bgcyan}","$v","$TCOLOR{endcolor}\n";
	    } else {
		print "\t$v\n";
	    }
	}
	print "\t","x"x80,"\n\n";
	
	print "\tTo terminate processess type '$TCOLOR{qunln}G:$TCOLOR{endcolor}MMAPID' for a Group or '$TCOLOR{qunln}P:$TCOLOR{endcolor}PROCID' for a specific Process (or [$TCOLOR{qunln}D$TCOLOR{endcolor}]one):\n\t> ";
	$ans = <STDIN>;
	if($ans) {
	    chomp $ans;
	    if($ans =~ m/^P:(\d+)/) {
		if(kill('KILL',$1)) {
		    print "\t\t Process [",$ids{$1},"]\n\t\t :Successfully terminated!\n";
		} else {
		    print "\t\t PROCID $1 not found!\n";
		}
	    } elsif($ans =~ m/^G:([A-Z]{2}[0-9]{4})/) {
		local $MPID = $1; #temporary change value into $1
		killScreens("s"); #kill spawn targeting this chosen ID
	    } elsif ($ans =~ m/^[dD](one)?/) {
	    } elsif($ans) {
		print "\t Command $ans not found!\n";
		sleep 1;
	    }
	}
    } while($ans !~ m/^[dD](one)?/);
    return $sccnt;
}

sub getRTKey { #thread
    $RTKEY = undef;
    my $key;
    $SIG{'KILL'} = sub {
	threads->exit();
    };
    $semaphore->down();
    ReadMode 3;
    do {
	while(not defined ($key = ReadKey(-1))) {
	    if($TERM) {
		$key = 9;
		last;
	    }
	}
	$key =~ s/\s+//g;
    } while($key !~ m/^[12Ss9\+\-]/); #9: get out
    $semaphore->up();
    ReadMode 0;
    $RTKEY = $key;
}

sub monitor { #thread
    my $maxruns = $_[0];
    my $p = $_[1];
    my $thrKEY = $_[2];
    my $quiet = $_[3];

    $SIG{'KILL'} = sub { 
	if($thrKEY) {
	    if($thrKEY->is_running()) {
		$thrKEY->kill('KILL')->detach();
	    }
	}
	threads->exit(); 
    };

    my @out = getProgProcessInfo($p);
    while(1) {
	last if($TERM);
	my @list = getScreenList(2);

	if(!@list || $list[0] =~ m/^No\s*Sockets\s*found/i) {
	    markFinished([]) if(!@SCHEDULEDNXT && scalar(grep {/^0$/} values %COMPLETED) == 1);
	    print "No screen sockets are available!\n" if(@list);
	    return (0,$thrKEY);
	}

	my @scpassoc = getProgScreenAssoc($p,\@list,\@out);
	my %tstatus = getTaskStatus();

	if($thrKEY->is_joinable() && !$thrKEY->is_detached()) {
	    $thrKEY->join();
	    if($RTKEY =~ m/^[Ss]$/) { #toggle show complete screen list or only from curr mpid
		$PALL = $PALL ? 0 : 1;
		undef $RTKEY;
	    }
	    if($RTKEY =~ m/^\+$/) { #increase max num tasks/cpus
		$INCSCNUM++;
		undef $RTKEY;
	    } elsif($RTKEY =~ m/^\-$/) { #decrease max num tasks/cpus
		$INCSCNUM--;
		undef $RTKEY;
	    }
	    $FKEY = $RTKEY if(defined $RTKEY && $RTKEY =~ m/[12]/ && !$thrKEY->is_running());
	    $thrKEY = threads->create('getRTKey');
	}

	my $MPIDcnt = getMPIDRunsCount(\@list); # this count
	my $sccnt = scalar(@list);
	if($quiet) {
	    printHeader();
	    if(defined $FKEY && $FKEY == 2) {
		printTaskStatus(\%tstatus);
	    } elsif(defined $FKEY && $FKEY == 1) {
		my $cnt = $MPIDcnt != $sccnt ? $MPIDcnt."/".$sccnt : $sccnt;
		printRunStatus($cnt,\@scpassoc);
	    }
	    @out = getProgProcessInfo($p);
	}
	
	if($MPIDcnt < $maxruns) {
	    markFinished(\@list);
	    return ($MPIDcnt,$thrKEY);
	}
    }
    return ();
}

sub getMPIDRunsCount { #thread-sub
    my @sclist = @{$_[0]};

    my $MPIDwtag = MPIDTAG."".$MPID;
    
    my @runs = grep {/$MPIDwtag/} @sclist;

    return scalar(@runs);
}

sub markFinished { #thread-sub
    my @sclist = @{$_[0]};

    my $max = 0;
    my @scheduled = keys %COMPLETED;
    my $MPIDWtag = MPIDTAG."".$MPID;

    if(@sclist) { #non-zero screens ..
	my @runs = grep {/$MPIDWtag/} @sclist;
	
	foreach my $r (@runs) {
	    my $stitle = getScreenTitle($r);
	    if(exists $COMPLETED{$stitle}) {
		@scheduled = grep {!/$stitle/} @scheduled; #remove from scheduled which still are running (=common)
	    }
	}
    }

    map {$_ > $max ? $max = $_ : $max} values %COMPLETED;

    foreach (@scheduled) {
	if($COMPLETED{$_} == 0) { #update "running"
	    $COMPLETED{$_} = ++$max;		
	    $TIMEUSED{$max} = getPAMLTimeUsed($TITLETOFILE{$_});
	}
    }
}

sub getProgProcessInfo { #thread-sub
    my $p = shift;

    #http://unixhelp.ed.ac.uk/CGI/man-cgi?ps
    my %PROCSTATUS = (
	"R" => "$TCOLOR{bold}\[R: RUNNING\]$TCOLOR{endcolor}", 
	"S" => "$TCOLOR{yellow}\[S: WAITING FOR USER INPUT\]$TCOLOR{endcolor}",
	"Z" => "$TCOLOR{red}\[Z: WRONGLY TERMINATED\]$TCOLOR{endcolor}",
	"D" => "$TCOLOR{yellow}\[D: UNINTERRUPTIBLE SLEEP(IO)\]$TCOLOR{endcolor}",
	"T" => "$TCOLOR{lred}\[T: STOPPED (BY SIGNAL)\]$TCOLOR{endcolor}",
	"X" => "$TCOLOR{red}\[X: !DEAD!\]$TCOLOR{endcolor}"
	#"W" => "[Paging]" #not valid since kernel 2.6.xxx
	);

    my @proglist;
    my @pidlist;
    @pidlist = grep {/(\.\/)?$p/} qx/ps au/;
    @pidlist = grep {!/.*perl.*/} @pidlist;
    chomp @pidlist;
    foreach (@pidlist) {
	if($_ =~ m/\s+([DRSTXZ])([\<NLsl\+]+)\s+/) {
	    my $V = "  ".$PROCSTATUS{$1}."  ";
	    $_ =~ s:\s+[DRSTXZ][\<NLsl\+]+\s+:$V:;
	}
	push @proglist, $_;
    }

    return (@proglist);
}

sub getTaskStatus { #thread-sub
    my @FINISHED = ();
    my %NOTRUNNING = ();

    my %tstatus;
    $tstatus{Sc} = 0; #runcount
    $tstatus{Fc} = 0; #finishedcount
    foreach (keys %COMPLETED) {
    	if($COMPLETED{$_} > 0) {
    	    $NOTRUNNING{$_} = $COMPLETED{$_};
	    $tstatus{Fc}++;
    	} else {
	    $tstatus{Sc}++;
	}
    }
    @FINISHED = sort {$NOTRUNNING{$b} <=> $NOTRUNNING{$a}} keys %NOTRUNNING if(%NOTRUNNING);
    my $total = $tstatus{Fc} + $tstatus{Sc};

    @{$tstatus{S}} = ();
    @{$tstatus{F}} = ();
    if(@SCHEDULEDNXT) {
	for(my $i = 0; $i < SHOWMAX ; $i++) {
	    push @{$tstatus{S}}, sprintf("%s%i. %s%s","\t\t$TCOLOR{bold}",($total+$i+1),$SCHEDULEDNXT[$i],"$TCOLOR{endcolor}\n") if(defined $SCHEDULEDNXT[$i]);
	}
    }
    if(@FINISHED) {
	for(my $i = 0; $i < SHOWMAX ; $i++) {
	    if(defined $FINISHED[$i]) {
		push @{$tstatus{F}}, sprintf("%s%s%s  %i. %s%s","\t$TCOLOR{Bcyan}TU: ",$TIMEUSED{$COMPLETED{$FINISHED[$i]}},"$TCOLOR{endcolor}$TCOLOR{bold}",$COMPLETED{$FINISHED[$i]},$TITLETOFILE{$FINISHED[$i]},"$TCOLOR{endcolor}\n");
	    }
	}
    }

    return %tstatus;
} 

sub getProgScreenAssoc { #thread-sub
    my $p = shift;
    my @sclist = @{$_[0]};
    my @proglist = @{$_[1]}; 

    my @psres = grep {!/\s+SCREEN\s+/} qx/ps -eo 'user,pid,pcpu,pmem,ppid,stat,time,args'/;
    chomp @psres;
    my %SCassoc = ();
    my ($scpid,$pruser,$prpid,$prppid);
    for(my $i=0; $i < (@psres) ; $i++) { #find correspondence between screen and paml program
	if($psres[$i] =~ m/\s+(\.\/)?$p\s+/) { #middle of line : program
	    ($pruser,$prpid,undef,undef,$prppid) = split(/\s+/,$psres[$i],6); #Linux: ppid = scpid | MacOS: prg ppid => ('login' pid) ppid => scpid
	    $scpid = $prppid; #Linux
	    $scpid = getprocppid($prppid) if($^O eq "darwin"); #MacOSX
	    $SCassoc{$scpid} = $prpid;
	}
    }

    my @colored = ();
    foreach (0..@sclist-1) {
	($scpid,undef) = split(/\./,$sclist[$_],2);
	$scpid =~ s/\s+//g;
	my $cpid = $SCassoc{$scpid}; 
	next if(!$cpid);
	my @pstat = grep {/\s+$cpid\s+/} @proglist;
	#user   17694 99.3  0.0 191312 12744 pts/16   Rs+  16:41   3:03          |   \_ ./codeml codeml.ctl
	if(@pstat) { #avoid warns...
	    $pstat[0] =~ s/^\s+//;
	    $pstat[0] =~ s/^$ENV{USER}//;
	    $pstat[0] =~ s/^\s+//;
	    if($sclist[$_] =~ m/$MPID/) {
		unshift @colored, "   \\_\tPRGM:\t$pstat[0]\t".(getPAMLversion($sclist[$_]))."\n";
		unshift @colored, "SCREEN:\t$TCOLOR{bgcyan}$sclist[$_]$TCOLOR{endcolor}\n";
	    } else { #others
		if($PALL) {
		    unshift @colored, "   \\_\tPRGM:\t$pstat[0]\t".(getPAMLversion($sclist[$_]))."\n";
		    unshift @colored, "SCREEN:\t$sclist[$_]\n";
		}
	    }
	}
    }
    
    return @colored;
}


sub getprocppid { #thread-sub
    my $pid = shift;
    my @ppid = qx/ps -p $pid -o ppid/;
    if(@ppid == 1) {
	warn "WARNING:: Could not determine ppid of process $pid.\n";
	return "";
    }
    $ppid[1] =~ s/\s*//g;
    return $ppid[1];
}

sub printHeader { #thread-sub
    clear(0);
    print $TCOLOR{menu},"[ MMAP ",$TCOLOR{endcolor},$TCOLOR{bgcyan},MPIDTAG,$MPID,$TCOLOR{endcolor},$TCOLOR{menu},", Started on: ",STARTTIME,", Current time: ",scalar(localtime())," ]",$TCOLOR{endcolor},"\n";
    print $TCOLOR{menu},"[ MMAP USER: ",$WORKINGUSER," ]",$TCOLOR{endcolor},"\n" if($WORKINGUSER);
    print "\n";
}

sub printRunStatus { #thread-sub
    my $sccnt = shift;
    my @colored = @{$_[0]};

    print "\n";
    print $TCOLOR{menu},"RUN STATUS screen\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor};
    print "\t","x"x80,"\n";
    print "\t","::: $TCOLOR{tblttl}Status of the ",$sccnt," instances running$TCOLOR{endcolor}:\n";
    print "\t","x"x80,"\n\n";
    print foreach (@colored);
    print "\n\t","x"x80,"\n\n";
    printVersion(1);
}

sub printTaskStatus { #thread-sub
    my %tstatus = %{$_[0]};

    print "\n";
    print $TCOLOR{menu},"TASK STATUS screen\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor};

    my $flcntdown = join("/",@NSCHNXT);
    print "\t","x"x80,"\n";
    print "\t","::: $TCOLOR{tblttl}Files running next ($flcntdown)$TCOLOR{endcolor}:\n";
    print "\t","x"x80,"\n";
    if(scalar(@{$tstatus{S}}) == 0) {
	print "\t\t$TCOLOR{lred}No more files to be launched for execution.$TCOLOR{endcolor}\n";
    } else {
	print foreach (@{$tstatus{S}});
	print "\t\t","...","\n" if($NSCHNXT[0] > SHOWMAX);
    }
    print "\t","x"x80,"\n\n";

    $flcntdown = $tstatus{Fc}."/".$NSCHNXT[1];
    print "\t","x"x80,"\n";
    print "\t","::: $TCOLOR{tblttl}Files finished ($flcntdown)$TCOLOR{endcolor}:\n";
    print "\t","x"x80,"\n";
    if(scalar(@{$tstatus{F}}) == 0) {
	print "\t\t$TCOLOR{lred}No files have finished yet!$TCOLOR{endcolor}\n";
    } else {
	print foreach (@{$tstatus{F}});
        print "\t\t","...","\n" if($tstatus{Fc} > SHOWMAX);
    }
    print "\t","x"x80,"\n\n";
    printVersion(0);
}

sub printFinalStatus {
    my %DISPC = map {$COMPLETED{$_} => $_ } keys %COMPLETED; #invert COMPLETED
    my @display = sort {$b <=> $a} (keys %DISPC); 

    if($EFL) { #logging runs info 
	open(my $FH, ">", catfile(".",$FTSTATUSFL)) or die "ERR:: Could not open file for logging: $!";
	print $FH "Rank,File,Time Used\n";
	foreach (@display) {
	    print $FH "$_".",".$TITLETOFILE{$DISPC{$_}}.",".$TIMEUSED{$_}."\n" if(defined $_ && defined $TITLETOFILE{$DISPC{$_}} && defined $TIMEUSED{$_});
	}
	close($FH);
    }

    print "\n";
    print $TCOLOR{menu},"FINAL STATUS screen\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x20,"\n",$TCOLOR{endcolor};
    print "\t","x"x80,"\n";
    print "\t","::: $TCOLOR{tblttl}Runs completed$TCOLOR{endcolor}:\n"; 
    print "\t","x"x80,"\n";
    foreach (@display) {
	print "\t$TCOLOR{Bcyan}TU: ",$TIMEUSED{$_},$TCOLOR{endcolor},$TCOLOR{bold}," "x4,"$_",". ",$DISPC{$_},"$TCOLOR{endcolor}\n" if(defined $_ && defined $TIMEUSED{$_} && defined $DISPC{$_});
    }
    print "\t","x"x80,"\n\n";
    printVersion(2);
}

sub printVersion {
    my $pv = shift;
    if($pv == 0) {
	print $TCOLOR{menu},"="x27,$TCOLOR{endcolor},"\t"x6,$TCOLOR{menu},"="x27,$TCOLOR{endcolor},"\n  Press $TCOLOR{chscunln}1$TCOLOR{endcolor} for Run Status","\t  ","VERSION: ".VERSION,"\t"x2,"Ctrl-C or Ctrl-\\ to exit...\n",$TCOLOR{menu},"="x27,$TCOLOR{endcolor},"\t"x6,$TCOLOR{menu},"="x27,"\n",$TCOLOR{endcolor};
    } elsif($pv == 1) {
	print "\t"x4,"Press $TCOLOR{chscunln}s$TCOLOR{endcolor} to show/hide other MPIDs.\n";
	if($INCSCNUM == 0) {
	    print "\t"x2,"Press $TCOLOR{chscunln}+/-$TCOLOR{endcolor} to increase/decrease maximum number of instances running.\n\n";
	} else {
	    my $indec = "Decrement";
	    $indec = "Increment" if($INCSCNUM > 0);
	    print "\t"x4, "$indec in maximum number of tasks: $INCSCNUM\n\n";
	}
	print $TCOLOR{menu},"="x27,$TCOLOR{endcolor},"\t"x6,$TCOLOR{menu},"="x27,$TCOLOR{endcolor},"\n  Press $TCOLOR{chscunln}2$TCOLOR{endcolor} for Task Status","\t  ","VERSION: ".VERSION,"\t"x2,"Ctrl-C or Ctrl-\\ to exit...\n",$TCOLOR{menu},"="x27,$TCOLOR{endcolor},"\t"x6,$TCOLOR{menu},"="x27,"\n",$TCOLOR{endcolor};
    } else {
	print "\t"x4,"VERSION: ".VERSION."\n\n\n";
    }
    print "\n\n";
}

sub getPAMLTimeUsed {
    my $iflpath = shift;

    my $respath = catfile((splitpath($iflpath))[1],$CMNRESFL);

    if(!-f $respath || !-e $respath) {
	warn "WARNING:: Could not determine time used: Not a valid common results filename in [$respath]\n";
	return "--:--"; #undetermined
    }

    open(my $fh, "<", $respath);
    my @tuline = <$fh>;
    close $fh;

    @tuline = @tuline[$#tuline-10..$#tuline]; #search in last 10 lines
    @tuline = grep {/Time\s*used:/i} @tuline;
    
    if(!@tuline) { #if not found ; then there are most probably errors; hence return "possible error" indication;
	warn "WARNING:: Could not determine time used: Time not found in results file [$respath]\n";
	return "!ERR:PAMLEXC!";
    }

    chomp $tuline[0];
    $tuline[0] =~ s/^Time\s*used:\s*//i;
    return $tuline[0];  
}

sub getPAMLversion {
    my $scitem = shift;

    my $iflpath = $TITLETOFILE{getScreenTitle($scitem)};
    return "" if(!$iflpath);
    
    my $respath = catfile((splitpath($iflpath))[1],$CMNRESFL);

    if(!-f $respath || !-e $respath) {
	warn "WARNING:: Could not determine PAML version: Not a valid common results filename in [$respath]\n";
	return ""; #undetermined
    }

    open(my $fh, "<", $respath);
    my @vline = grep {/^CODONML/} <$fh>;
    close $fh;

    if(!@vline == 1) {
	return "";
    }

    $vline[0] =~ s/^CODONML\s*\(in\s//;
    $vline[0] =~ s/\).*\s//;
    $vline[0] =~ s/version\s*/v/;
    $vline[0] =~ s/paml/PAML/;
    #CODONML (in paml version 4.8a, July 2014)  align.txt
    return "[".$vline[0]."]";
}

sub getScreenTitle {
    my $screenrun = shift;

    $screenrun =~ s/^\s+\d+\.//;
    $screenrun =~ s/\s+\(.*//;

    return $screenrun;
}

sub getScreenList { 
    my $filter = shift;

    qx/screen -wipe/ if($filter != 2); #clean sockets
    my @list = qx/screen -ls/;
    chomp @list;
    return @list if(!@list || $list[0] =~ m/^No\s*Sockets\s*found/i);
    shift @list;
    pop @list;
    pop @list;
    if($filter) {
        @list = grep {/\.MPID\:/} @list;
    }
    return @list;
}

sub createMMAPID {

    my @ALPHA = ("A".."Z");
    my %alpha;
    my $i = 0;
    $alpha{$i++} = $_ foreach (@ALPHA);

    my $chari1 = int(rand(scalar(@ALPHA)));
    my $chari2 = int(rand(scalar(@ALPHA-1)));
    my $mpid = $alpha{$chari1}."".$alpha{$chari2}."".int(rand(10))."".int(rand(10))."".int(rand(10))."".int(rand(10));
    return $mpid;
}


sub pathtoTitle{
    my $title = shift;
    $title =~ s:^(\.\/)::; #if starts with ./
    $title =~ s:/:-:g; #all path-separator to -
    $title =~ s:\s:_:g; #all space to _
    return $title;
}

sub startProgram {
    my $inputfl = $_[0];
    my $program = $_[1];
    my $running = 1;
    my @screen = SCREENP;

    my (undef, $scnm, $ifl) = splitpath($inputfl); 
    $scnm =~ s:/$::;
    my $scdir = $scnm;
    my $l = '\$';
    $scnm =~ s/\$/$l/g;
    $l = '\#';
    $scnm =~ s/\#/$l/g;
    $l = '\&';
    $scnm =~ s/\&/$l/g;

    my $title = MPIDTAG."".$MPID.":".pathtoTitle($scnm);
    my ($efpath,$prog) = split(/:/,$program, 2);

    my $ppath = catfile($scnm,$prog); #path to exec on data loc
    if(!-e $ppath || !-B $ppath) { #verify if it is there
	$ppath = catfile($efpath, $prog); #redopath
	if(-e $ppath && -B $ppath) { #copy in place
	    my $dest = catfile($scnm, $prog);
	    copy($ppath,$dest) or die "ERR:: Could not copy file: $!\n";
	    chmod 0755, $dest;
	} else {
	    die "ERR:: Could not find executable file in path [$ppath] ... \nExiting...\n";
	}
    }
    chmod 0755, $ppath;

    push @screen, "-S $title", catfile(".",$prog), $ifl;
    my $pwd = getcwd();

    chdir catfile(".",$scdir) or die "ERR:: Could not change directory into $pwd, please verify your paths: $!"; 
    system(join(" ", @screen));
    if($? == -1) {
	die "ERR:: An error occured during the initialization of the screen program: $!\n";
    } elsif($? & 127) {
	printf "ERR:: screen unexpectedly terminated with signal %d, %s coredump\n",($? & 127),($? & 128) ? 'with':'without';
    } else {
	chdir "".$pwd or die "ERR:: Could not change directory into $pwd, please verify your paths: $!";
	$running = 0;
	$COMPLETED{$title} = $running;
	$TITLETOFILE{$title} = $inputfl;
    }
    return $running;
}

sub readinDirectoryStructure {
    my $inputdir = shift;
    my $ctltag = shift;
    my $recurse = shift;
    my $domissing = shift;
    my @DCOLL;
    my @content = [];
    @content = getDirContent($inputdir);
    
    my ($ttag,$fltag) = split(/:/,$ctltag);
    my @scheduledfls;
    while(1) {
	while($_  = shift @content) { 
	    if(-f $_ && $_ =~ m/$fltag/) {
		my $fln = (splitpath($_))[2];
		print ":: Gathering file: $_ ; $fln\n";
		if($ttag eq "s" && $fln =~ m/^($fltag)/) {
		    push @scheduledfls, $_;
		} elsif($ttag eq "e" && $fln =~ m/($fltag)$/) {
		    push @scheduledfls, $_;
		} elsif($ttag eq "f" && $fln eq $fltag) {
		    push @scheduledfls, $_;	
		}
	    } elsif (-d $_) {
		print ":: Adding Subdirectory $_\n";
		push @DCOLL, $_;
	    }
   
	    if(!@content && $recurse) { #after all content is passed through
		print ":: No more content on current directory\n";
		while(my $currdir = shift(@DCOLL)) {  # remove at the beggining
		    print ":: Going into directory $currdir...\n";
		    @content = getDirContent($currdir);
		    last if(@content);
		    print ":: Directory $currdir empty!\n" if(!@content);
		}
	    }
	}
	last if(!@content);
    }

    #lmap.pl mode only! limit running inputdir to those specified by user; avoid re-running others possibly already done
    setCookyLoc($inputdir);
    my @torun = readLMAPCooky();
    my @finaltorun = ();
    if(@torun) {
	foreach (@torun) {
	    my @tmptorun = ();
	    my $models = $C_MTYPESTAGS{$_};
	    @tmptorun = grep {/\/$models/} @scheduledfls; #only run these!
	    @finaltorun = (@finaltorun,@tmptorun);
	}
	@scheduledfls = (@finaltorun);
    }

    #option -x
    print "\n:z: ",scalar(@scheduledfls)," files scheduled to run in specified program.\n";  
    if($domissing) {
	@scheduledfls = scheduleOnlyMissing(\@scheduledfls);
	print "\n:z: ",scalar(@scheduledfls)," files not completed, scheduled to run in specified program.\n";
    }
    return @scheduledfls;
}

#filter from all scheduled, only those which were not completed in a previous mmap run
sub scheduleOnlyMissing { 
    my @schfls = @{$_[0]};

    my @notran;
    foreach (@schfls) {
	my $time = getPAMLTimeUsed($_);
	if($time !~ m/^\d+:\d+$/) { #file does not exist or not finished
	    push @notran, $_;
	}
    }
    return @notran;
}


sub HELP_MESSAGE {
    print shift;
    my $dfltcommonresfl = COMMONRESFLN;
    my $ctrlfl = CODEMLCTRLFLN;
    my $dfltoptf = CODEMLCTRLOPTF;
    my $dfltoptp = CODEMLDFLTOPTP;
    my $citation = CITATION;
    my $pweb = PROJWEBSITE;
    my $v = VERSION;
    my @HELP = <<HELP;

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    mmap.pl - Run multiple PAML instances.
   
  $TCOLOR{bold}SYNOPSIS:$TCOLOR{endcolor}
    
    mmap.pl -d [dirpath] {-r} {-n [int]} {-f [e:ctl]} {-p [l:progname]} {-R [resfilename]} {-l} {-e {emailaddr}}

  $TCOLOR{bold}DESCRIPTION:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | This application is intended to run and monitor multiple PAML (codeml) instances,
    | from an organized directory struture, containing all necessary data files.
    | This organized directory structure can be previously created by using the gmap.pl application.
    | The results from this/these runs can then be collected with the application imap.pl.

  $TCOLOR{bold}OPTIONS:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | -d [dir]        Input directory containing all necessary files.
    |                 ----------------------- not mandatory: ----------------------
    | -R [filename]   Results filename common to all PAML runs. By default, is the same occuring in templates files: $dfltcommonresfl
    | -f [tag:part]   Choose control files through a given filename portion, where tag is one of:
    |                      f: filename equals given part
    |                      s: filename starts with part
    |                      e: filename ends with part
    |                 By default, is $dfltoptf based on the common codeml control filename: $ctrlfl 
    | -p [tag:prog]   Program to run, where tag is one of:
    |                      l: from default location eg.: l:progname
    |                      f: given full path eg.: f:path/to/progname
    |                 For tag 'l' the location defaults to HOME/bin/ or the one chosen by the user during installation.
    |                 Use tag 'f' to give a different location for the intended PAML executable(s).
    |                 By default, is $dfltoptp based on the default PAML executables location : $EFPATH
    | -r              Work recursively.
    | -x              Use this option to rerun a previous directory, which may have files
    |                 not completed or PAML runs previously not initiated.
    | -n [integer]    Specify the maximum number of CPUs/cores available (limit) to use for the current task.
    |                 By default will try to use the aproximate maximum number of cores available.
    |                 This option can be usefull to enable sharing of the CPU capacity between users or different (mmap) tasks.
    | -l              Enable logging of all the finished runs in a CSV file located at the current working directory.
    | -e {emailaddr}  Enable email notification, when mmap.pl finishes. The argument to this option, is optional and has two meanings:
    |                 If email address is not specified, it will default to the address defined during installation. Otherwise, it
    |                 will be given preference to the address here supplied. In case the address is not given in either case,
    |                 (during installation and in this option), notification will not be sent.
    |                 -----------------------------------
    | -h              This help
    | -v              Application version

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2014.

  $TCOLOR{bold}CITATION AND DOWNLOAD:$TCOLOR{endcolor}
    
    $citation
    $pweb

  $TCOLOR{bold}LICENSE AND COPYRIGHT:$TCOLOR{endcolor}

    LMAP: Lightweight Multigene Analyses in PAML
    mmap.pl is part of LMAP package
    Copyright (C) 2015 Emanuel Maldonado

  mmap.pl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  mmap.pl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mmap.pl.  If not, see <http://www.gnu.org/licenses/>.

HELP
print @HELP;
    exit 1;
}

sub VERSION_MESSAGE {
    my $v = VERSION;
    my @VERS = <<VERS;
Use option -h for help and other information.

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    mmap.pl - Run multiple PAML instances.

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2014.

VERS

print @VERS;
    exit 1;
}


