#!/usr/bin/perl -w


#################################################
#  Interactively rearrange imap.pl information  #
#           ------------------                  #
#   Written by Emanuel Maldonado - Mar 2014     #
#############################################################################
#  LMAP: Lightweight Multigene Analyses in PAML
#  omap.pl is part of LMAP package
#  Copyright (C) 2015 Emanuel Maldonado
#
#  omap.pl is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  omap.pl is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with omap.pl.  If not, see <http://www.gnu.org/licenses/>.
#############################################################################


use Data::Dumper;
use strict;
use warnings;
use threads;
use Getopt::Long;
Getopt::Long::Configure ("bundling");

use Math::Cephes qw(:dists);
use Statistics::Distributions;

use sigtrap qw(handler do_rexit INT TERM TSTP HUP);
use sigtrap qw(handler dactivate_scroll QUIT USR1 USR2);

use Term::ReadKey;
use File::Spec::Functions qw(:ALL);

#My modules
use FindBin;
use lib "$FindBin::Bin/LMAPlib/";
use MyUtil qw(:OPAML :CONST);

use vars qw(@Menutable $Ncols @TTLcolumns @reset_MT $reset_Ncols @reset_TTLcolumns @FMenutable $FNcols @FTTLcolumns
$INFILE $OUTFILE $LEN %TTLHDR %BEB %HIDDENcols $SAVED $BEBSAVED $IGNBEB %UCOMMAND $DEFS_FILE $STROW $FSTROW $USTROW $KEY %TCOLOR
$S_ACTIVE $CtrlD_ON $TERM);

use constant {
    DLINESHOW => 30,
    DCOLSZ => 17,
    USERTBL => "U",
    FINALTBL => "F",
    CELLCTT_NFOUND => "Not Found!",
    CELLCTT_DASH => "-",
    CELLCTT_NAVAIL => "-na-",
    COLTTL_MK => "Mark", 
    COLTTL_IDX => "Index",
    COLTTL_MDL => "model",
    COLTTL_LNL => "lnl",
    COLTTL_NP => "np",
    COLTTL_FL => "File",
    COLTTL_TLRT => "LRTtest",
    COLTTL_DLRT => "deltaLnL",
    COLTTL_DF => "df",
    COLTTL_PV => "pvalue",
    COLTTL_CCL => "conclusion",
    COLTTL_BEB => "BEBsites",
    PHELPFILE => "omap_promptHelp.txt",
    VERSION => "1.0.0 Nov 20th, 2015"
};

unew();
if(wantColor()) {
    %TCOLOR = enableTCOLOR();
} else {
    %TCOLOR = disableTCOLOR();
}
$CtrlD_ON = 0; #ensure scroll ON by ^D only
$S_ACTIVE = 0;
$TERM = 0;
$DEFS_FILE = getPAMLRTDefsFl();

my %opts;

GetOptions(
    "i=s" => \$opts{i}, 
    "o=s" => \$opts{o}, 
    "help" => \$opts{help}, 
    "v!" => \$opts{v}, 
    "h!" => \$opts{h}, 
    ) or HELP_MESSAGE("*"x30);

%opts = cleanOpts(\%opts);

HELP_MESSAGE("") if($opts{h} || !%opts);
VERSION_MESSAGE() if($opts{v});

dumpPromptHelp() if($opts{help});

if(!$opts{i} || !-f $opts{i}) {
    HELP_MESSAGE("ERR:: Missing input file: use -i option followed by path to file.\n");
}

if(!$opts{o}) {
    HELP_MESSAGE("ERR:: Missing output file: use -o option followed by path to file.\n");
}

printLicense($0);

$INFILE = $opts{i};
$OUTFILE = $opts{o};

$BEBSAVED = $IGNBEB = 0;
$USTROW = $FSTROW = $STROW = 0;
start_menutable();
@Menutable = reset_menutable();
$Ncols = $reset_Ncols;
$FNcols = 0;

makeHdr();
my %sizes = calcColsizes();
my $tbl = USERTBL;
my $dflt_show = DLINESHOW;
my @UMenutable; #tmp table
my $Ucols;
my @UTTLcolumns;

my $ph = 1;
my $shflhdr = 0;
my $msg;
my $op = "";
my @ucmdmode = ();

#Mainloop
for (my $pcnt = 0;;$pcnt++) {
    print_menutable($dflt_show,$shflhdr,\%sizes) if($ph);
    printPrompt($msg,$ph,$tbl,$pcnt);
    $op = $msg = "";

    #input
    if(@ucmdmode) {
	$op = shift @ucmdmode; #user def commands
	print "$op\n";
    } else {
	if($S_ACTIVE) {
	    getUDKey($dflt_show,$shflhdr,$msg,$ph,$tbl,$pcnt,\%sizes); #non-blocking scrolling
	    $msg = "-scroll OFF!-";
	} else {
	    $op = <STDIN>;
	    if(!defined $op && !STDIN->error) { #received Ctrl-D signal
		$CtrlD_ON = 1; #to ensure scroll ON it is only done by this signal
		dactivate_scroll();
		$CtrlD_ON = 0;
		$msg = "-scroll ON!-";
		$op = "";
	    } elsif (STDIN->error) { #break-out
		$op = "";
	    } else {
		chomp $op;
		$op =~ s/^\s+//;
	    }
	}
    }
    $ph = 1; #back to basics!
    $shflhdr = 0;

    #switch
    if(exists $UCOMMAND{$op}) { #initialize user command
	@ucmdmode = split /;/, $UCOMMAND{$op};
	next;
    } elsif($op =~ m/^open\s+(([~\.]\/)?.*\.csv)/) {
	my $f = $1;
	$f =~ s:^~:$ENV{HOME}:;
	$msg = "Could not '$op': You must switch to User Table first (type: show u)!" and next if($tbl eq FINALTBL);
	$msg = "Could not '$op': $f does not exist!" and next if(!-e $f);
	$msg = "Could not '$op': $f is not a (CSV/Text file) file!" and next if(!-f $f);
	$msg = "Could not '$op': File $f is empty!" and next if(-s $f <= 1);
	open_menutable($f);
	makeHdr();
	%sizes = calcColsizes();
    } elsif($op =~ m:^ls(\s)*((([~\/\.]{1}[\/]?)|(((([~\.]?\/)?)[\w\d\.-])+[\/]?))\s*)?$:) {
	my $d = $2;
	$d =~ s:^~:$ENV{HOME}: if($d);
	list_directory($d,0);
	next;
    } elsif($op =~ m/^show\s+(\d+|[ufUF]{1})$/p || $op =~ m/^show\s+(\d+)\s+([ufUF]{1})$/p) {
	my $t = "";
	if($1 && $2) {
	    $dflt_show = $1;
	    $t = $2;
	} elsif(!$2) {
	    $t = $1;
	    $dflt_show = $t if ($t =~ m/\d+/);
	}
	$msg = "Could not '$op': Table ".uc($t)." already showing." if(uc($t) eq $tbl);
	$msg = "Could not '$op': Final Table is still empty." and next if(uc($t) eq FINALTBL && !@FMenutable);
	next if(!@UMenutable && uc($t) eq USERTBL);
	$tbl = show_menutable($t,$tbl);
	$msg = "\#Showing Final Table!" if($tbl eq FINALTBL);
	$msg = "\#Showing User Table!" if($tbl eq USERTBL);
	makeHdr();
	%sizes = calcColsizes();
    } elsif($op =~ m/^sort\s+C(\d+)\s+([adAD]{1})/) {
	$msg = "Could not '$op': No such column number." and next if(colRangeOffLimits($1));
	sort_menutable($1,$2);
	$SAVED = 0;
    } elsif($op =~ m/^delcl\s+C(\d+-\d+)/ || $op =~ m/^delcl\s+C((\d+)(,\d+)*)/) {
	$msg = "Could not '$op': No such column number(s) or range is off limits!" and next if(colRangeOffLimits($1));
	delcl_menutable($1);
	makeHdr();
	%sizes = calcColsizes();
	$SAVED = 0;
	$BEBSAVED = 0 if($tbl eq FINALTBL && !isColumnPresent(COLTTL_BEB));
    } elsif($op =~ m/^hide\s+C(\d+-\d+)/ || $op =~ m/^hide\s+C((\d+)(,\d+)*)/) {
	$msg = "Could not '$op': No such column number(s) or range is off limits!" and next if(colRangeOffLimits($1));
	hidecl_menutable($1);
	$SAVED = 0;
    } elsif($op =~ m/^mark\s+(\d+-\d+)/ || $op =~ m/^mark\s+((\d+)(,\d+)*)/ || $op =~ m/^mark\s*$/ ) { #one line ; set of lines x-y
	my $r = $1;
	$r = "1-".(scalar(@Menutable)-1) if(!$r);
	my $cnt = markrows_menutable($r);
	$msg = "\#".($cnt)." rows modified!";
	$SAVED = 0;
    } elsif($op =~ m/^markf\s+(\d+)\s+C(\d+)(\s+(\d+))?/) { #rows that match once for different values in length lines
	$msg = "Could not '$op': No such column number." and next if(colRangeOffLimits($2));
	my $cnt = markfst_menutable($1,$2,$3||$#Menutable-1);
	$msg = "\#Found ".($cnt)." rows!";
	$SAVED = 0;
    } elsif($op =~ m/^findr\s+\"(.*)\"\s+C(\d+)/) { # mark all rows containing "values" in column CX
	$msg = "Could not '$op': No such column number." and next if(colRangeOffLimits($2));
	my $cnt = markbyval_menutable($1,$2);
	$msg = "\#Found ".($cnt)." rows!";
	$SAVED = 0;
    } elsif($op =~ m/^copy\s*(\d+)?/) { # copy marked rows to final result
	$msg = "Could not '$op': operation need to be performed in User Table showing and not empty." and next if($tbl ne USERTBL);
	$msg = "Could not '$op': Table is currently not marked."  and next if(!isMenutable_marked());
	$msg = "Could not '$op': Final Table is currently not empty and containing a different number of columns." and next if($tbl eq USERTBL && $FNcols >0 && $FNcols != $Ncols);
	copyfrom_menutable($1||1);
	$SAVED = 0;
    } elsif($op =~ m/^move\s*/) { # move marked rows to final result
	$msg = "Could not '$op': operation need to be performed in User Table showing and not empty." and next if($tbl ne USERTBL);
	$msg = "Could not '$op': Table is currently not marked."  and next if(!isMenutable_marked());
	$msg = "Could not '$op': Final Table is currently not empty and containing a different number of columns." and next if($tbl eq USERTBL && $FNcols >0 && $FNcols != $Ncols);
	movefrom_menutable();
	$SAVED = 0;
    } elsif($op =~ m/^mrup\s*(\d+)?/) { # move rows up X positions
	$msg = "Could not '$op': Table is currently not marked."  and next if(!isMenutable_marked());
	mrowsup_menutable($1||1);
	$SAVED = 0;
    } elsif($op =~ m/^mrdn\s*(\d+)?/) { # move rows down X positions
	$msg = "Could not '$op': Table is currently not marked."  and next if(!isMenutable_marked());
	mrowsdown_menutable($1||1);
	$SAVED = 0;
    } elsif($op =~ m/^delrs\s*/) { # delete marked rows
	$msg = "Could not '$op': Table is currently not marked."  and next if(!isMenutable_marked());
	delrs_menutable();
	$STROW = @Menutable <= $dflt_show ? 0 : @Menutable - $dflt_show; #recenter table
	$SAVED = 0;
    } elsif($op =~ m/^plrt\s+(\d+(\.\d+)?)/) { # perform lrt pairwise
	$msg = "Could not '$op': LRT operation need to be performed in Final Table showing and not empty." and next if($tbl ne FINALTBL || !@FMenutable);
	$msg = "Could not '$op': Unable to perform LRT test in Final Table without model, np and lnl columns." and next unless(isColumnPresent(COLTTL_LNL) && isColumnPresent(COLTTL_MDL) && isColumnPresent(COLTTL_NP));
	$msg = "Could not '$op': Unable to perform LRT test: Please remove BEBsites column first!" and next if(isColumnPresent(COLTTL_BEB));
	plrt_fmenutable($1);
	makeHdr();
	%sizes = calcColsizes();
	$SAVED = 0;
    } elsif($op =~ m/^pbeb\s*/) { # putbeb in additional column
	$msg = "Could not '$op': BEB does not exist!" and next if(!%BEB);
	$msg = "Could not '$op': BEB can only be placed in Final Table showing and not empty." and next if($tbl ne FINALTBL || !@FMenutable);
	$msg = "Could not '$op': BEB column already exists!" and next if(isColumnPresent(COLTTL_BEB));
	pbeb_fmenutable();
	makeHdr();
	%sizes = calcColsizes();
	$SAVED = 0;
	$BEBSAVED = 1;
	$IGNBEB = 0;
    } elsif($op =~ m/^save(\!)?\s*(.+\..+)?/) { # save FMenutable to file specified or be default to -o option
	$msg = "Could not '$op': save performed only in Final Table showing and not empty." and next if($tbl ne FINALTBL || !@FMenutable);
	if(%BEB) {
	    if(!ignoreBEB()) {
		$msg = "\#Content not saved.";
		next;
	    }
	}
	my $ign = "";
	$ign = " (BEB ignored)" if($IGNBEB);
	my $cnt = save_fmenutable(0||$2);
	if($cnt == scalar(@Menutable)-1) {
	    $msg = "\#All rows saved to ".($2||$OUTFILE)."$ign\.";
	} else {
	    $msg = "\#$cnt marked rows saved to ".($2||$OUTFILE)."$ign\.";
	}
	$OUTFILE = $2 if($1 && $2); #keep track of last saved file...
	$SAVED = 1;
	$IGNBEB = 0;
    } elsif($op =~ m/^fh\s*/) { 
	$shflhdr = 1;
    } elsif($op =~ m/^empty\s*/) { 
	if($tbl eq FINALTBL) { 
	    @Menutable = ();
	    $Ncols = 0;
	    @TTLcolumns = @reset_TTLcolumns;
	} else {
	    @FMenutable = ();
	    $FNcols = 0;
	    @FTTLcolumns = @reset_TTLcolumns;
	}
	$msg = "\#Final Table emptied!";
	$SAVED = 0;
	$BEBSAVED = 0;
	$IGNBEB = 0;
    } elsif($op =~ m/^reset\s*/) {
	$msg = "Could not '$op': reset only in User Table showing." and next if($tbl eq FINALTBL);
	@Menutable = ();
	@Menutable = reset_menutable();
	@TTLcolumns = @reset_TTLcolumns;
	%HIDDENcols = ();
	makeHdr();
	%sizes = calcColsizes();
	$SAVED = 0;
	$BEBSAVED = 0;
	$IGNBEB = 0;
    } elsif($op =~ m/^defsave(\s+([\d\w]+\.[\d\w]+))?/) {
	$msg = "Could not '$op': no commands defined yet." and next if(!%UCOMMAND);
	my $fl =$1||$DEFS_FILE;
	saveCommandDefinitions($fl);
	$msg = "\#User Commands saved to $fl.";
    } elsif($op =~ m/^defload(\s+([\d\w]+\.[\d\w]+))?/) {
	my $fl =$1||$DEFS_FILE;
	$msg = "Could not '$op': file $fl does not exist." and next if(defined $fl && !-e $fl);
	loadCommandDefinitions($fl);
	$msg = "\#User Commands loaded from $fl.";
    } elsif($op =~ m/^defdel\s*/) {
	if(unlink $DEFS_FILE) {
	    $msg = "\#User Commands default file ($DEFS_FILE) deleted.";
	} else {
	    $msg = "Could not '$op': $!";
	}
    } elsif($op =~ m/^def\s+(\w+)=([\s\w\d\"\$\[\]\^,-]+(;[\s\w\d\"\$\[\]\^,-]+)*)/ || $op =~ m/^def\s+(\w+)\s*$/ || $op =~ m/^def\s*$/) {
	my ($a1,$a2) = ($1,$2);
	if($a1 && !$a2) {
	    if(exists $UCOMMAND{$a1}) {
		print "\n\t[ User Definition for command $a1: '$UCOMMAND{$a1}' ]\n";
		print "\nPress ENTER to continue: ";
		<STDIN>;
		next;
	    } else {
		$msg = "\#User Command '$a1' is not yet defined. You can define it by typing $a1=command1;...;commandn.";
	    }
	    next;
	}
	if(!$a1 && !$a2) {
	    $msg = "Could not '$op': no commands defined yet." and next if(!%UCOMMAND);
	    print "\n\tListing User Defined Commands:\n";
	    my $j = 0;
	    print "\t| ";
	    foreach (keys %UCOMMAND) {
		print "$_\t";
		print "\n\t| " if(++$j % 5 ==0);
	    }
	    print "\n\nPress ENTER to continue: ";
	    <STDIN>;
	    next;
	}
	my $rezn = invalidCmdDefinition($a1,$a2);
	$msg = "Could not '$op': $rezn" and next if($rezn);
	$msg = "\#User Command '$a1' Defined!";
	$msg = "\#User Command '$a1' ReDefined!" if(exists $UCOMMAND{$a1});
	$UCOMMAND{$a1} = $a2;
    } elsif($TERM || $op =~ m/^quit\s*/ || $op =~ m/^q\s*/) {
	last if(do_quit());
    } elsif($op =~ m/^\?\s*/) {
	promptHelp();
	$ph = 0;
    } else {
	$msg = "Could not '$op': Command not found!" if($op);
    }
}

exit 0;

##########################
############### Functions
#######################

sub dactivate_scroll { #signal-handler :working
    if($CtrlD_ON) {
	$S_ACTIVE = 1; #ON
	return;
    }
    $S_ACTIVE = 0; #OFF
}

sub do_rexit { #signal-handler
    $TERM = 1;
    STDIN->flush;
    STDIN->close() || warn "WARNING: STDIN close failed: $!";
}

sub do_quit {
    clear(0);
    local $SIG{INT} = 'IGNORE';
    local $SIG{QUIT} = 'IGNORE';
    local $SIG{TERM} = 'IGNORE';
    local $SIG{TSTP} = 'IGNORE';
    print "\n\n";
    print $TCOLOR{menu},"QUIT screen - Really Quit?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x13,"\n",$TCOLOR{endcolor},"\n";
    open(STDIN,"<","/dev/tty") or die "ERR:: Could not reopen STDIN: $!";
    my $ans;
    do {
	if(!$SAVED && @FMenutable) {
	    print "\tFinal Table appears to be not empty, do you really want to quit, without saving? ([$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	} else {
	    print "\tDo you really want to quit? ([$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	}
	$ans = <STDIN>;
	chomp $ans;
    } while ($ans !~ m/^[yYnN]/);
    if($ans =~ m/[yY](es)?/) {
	print "\nLeaving... Goodbye!\n";
	return 1;
    }
    $TERM = 0;
    return 0;
}

sub getUDKey { #non-blocking STDIN for scrolling up/down
    my $howmany = shift;
    my $fhs = shift;
    my $msg = shift;
    my $ph = shift;
    my $tbl = shift;
    my $pcnt = shift;
    my %sizes = %{$_[0]};

    local $SIG{INT} = 'IGNORE';
    local $SIG{TERM} = 'IGNORE';
    local $SIG{TSTP} = 'IGNORE';

    ReadMode 3;
    do {
	$KEY = "";
	while(not defined ($KEY = ReadLine(-1))) {
	    last if(!$S_ACTIVE);
	}
	if($KEY) {
	    $KEY =~ s/\s+//g;
	    scrollUP($howmany,$fhs,"-scroll up-",$ph,$tbl,$pcnt,\%sizes) if($KEY =~ m/^(\e\[A)+|(\e\[H)/); #UP / top
	    scrollDOWN($howmany,$fhs,"-scroll down-",$ph,$tbl,$pcnt,\%sizes) if($KEY =~ m/^(\e\[B)+|(\e\[F)/); #DOWN / bottom       
	}
	STDIN->flush();
    } while($S_ACTIVE);
    ReadMode 0;
}

sub scrollUP {
    my $howmany = shift;
    my $fhs = shift;
    my $msg = shift;
    my $ph = shift;
    my $tbl = shift;
    my $pcnt = shift;
    my %sizes = %{$_[0]};
    if($KEY =~ m/^(\e\[H)/) {
	$STROW = 0;
	$msg =~ s/up/top/;
    } else {
	my $len = length($KEY)/3; #arrow UP
	$STROW-- while($len-- && $STROW != 0);
    }
    print_menutable($howmany,$fhs,\%sizes);
    printPrompt($msg,$ph,$tbl,$pcnt);
}

sub scrollDOWN {
    my $howmany = shift;
    my $fhs = shift;
    my $msg = shift;
    my $ph = shift;
    my $tbl = shift;
    my $pcnt = shift;
    my %sizes = %{$_[0]};
    if($KEY =~ m/^(\e\[F)/) {
	$STROW = $#Menutable-$howmany;
	$STROW = 0 if($STROW < 0);
	$msg =~ s/down/bottom/;
    } elsif($STROW < $#Menutable-$howmany) {
	my $len = length($KEY)/3; #arrow DOWN
	$STROW++ while($len-- && $STROW <= $#Menutable-$howmany);
    }
    print_menutable($howmany,$fhs,\%sizes);
    printPrompt($msg,$ph,$tbl,$pcnt);
}

sub start_menutable {
    $SAVED = 0;
    %BEB = ();

    print "\n\n\n:z: Initializing User Table. Please wait...\n\n";

    my @pifile = readac($INFILE);

    my $header = shift @pifile;
    my ($c1,undef) = split(/,lnl/i,$header); #locate...
    my @c = split(/,/,$c1);  #count...
    @TTLcolumns = split(/,/,$header,@c+2); #split by its location...
    pop @TTLcolumns if($TTLcolumns[-1] !~ m/lnl$/i); #lnl must be the last or penultimate(s) in order to grab all the columns, but the BEB!!

    my $len = grep {/^[\w\/,]+/} @pifile;
    $LEN = length($len);

    my @columns;
    my $row = 1;
    my $flpath;

    for(my $j = 0; $j < @pifile ; $j++) {
	if($pifile[$j] =~ m/^([\w\d\/\.]+),/) {
	    getBEB($row,$j,\@pifile);
	    $flpath = $1;
	} else {
	    next;
	}

	$flpath = (splitpath($flpath))[1];
	$flpath =~ s:(\/)$::; 
	my @flpathelem = splitdir($flpath); 

	my @rest = split(/,/,$pifile[$j]);
	shift @rest;
	@rest = @rest[0..@TTLcolumns-2];
	
	@columns = (@flpathelem,@rest);
	
	my @f = ("[ ]", sprintf("%"."-$LEN"."s","$row."),@columns);

	push @reset_MT, [@f];
	$row++;
    }
    $reset_Ncols = @columns;
    @reset_TTLcolumns = @FTTLcolumns = @TTLcolumns;    
}

sub makeHdr {
    #permanent hdr
    my @ttl;
    shift @Menutable if($Menutable[0][1] eq "C0");
    $ttl[$_] = "C".$_ foreach (0..$Ncols);
    unshift @Menutable, [" "x3,sprintf("%"."-"."s",$ttl[0]),@ttl[1..$#ttl]];
    #file hdr map Cnumber->descriptive ttl
    %TTLHDR = ();
    %TTLHDR = map { my $d = $Ncols-$#TTLcolumns; $ttl[$_] => $TTLcolumns[$_ >= $d ? $_-$d : 0] } reverse(0..$Ncols);
    $TTLHDR{$ttl[0]} = COLTTL_IDX;
}

sub ignoreBEB {
    if($BEBSAVED) {
	return 1;
    }
    clear(0);
    print "\n\n";
    print $TCOLOR{menu},"SAVE TABLE screen - Really Save?\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"","="x17,"\n",$TCOLOR{endcolor},"\n";
    my $ans;
    do {
	print "\tThe BEB column exists and appears to be not placed in Final Table, do you really want to save and ignore BEB? ([$TCOLOR{qunln}Y$TCOLOR{endcolor}]es/[$TCOLOR{qunln}N$TCOLOR{endcolor}]o)\n\t> ";
	$ans = <STDIN>;
	chomp $ans;
    } while ($ans !~ m/^[yYnN]/);
    if($ans =~ m/[yY](es)?/) {
	$IGNBEB = 1; #make ignore
	return 1;
    }
    return 0;
}

sub getBEB {
    my $row = shift ;
    my $index = shift ;
    my @file = @{$_[0]};
    #BEB (always) last column, after lnl
    my $beb ="";
    for(my $i = $index ; $i < @file ; $i++) {
	if($file[$i] =~ m/,\"(Bayes.*)$/) {
	    $beb = "\"".$1."\n";
	    $i++;
	    while(1) {
		if($file[$i] =~ m/(.*)"$/) {
		    $beb .= $1."\"\n";
		    last;
		} else {
		    $beb .= $file[$i]."\n";
		}
		$i++;
	    }
	    $BEB{$row."."} = $beb;
	    last;
	} else {
	    last;
	}
    }
}

sub open_menutable {
    my $f = shift;
    $INFILE = $f;
    @reset_MT = ();
    $reset_Ncols = 0;
    @reset_TTLcolumns = ();
    @Menutable = ();
    start_menutable();
    @Menutable = reset_menutable();
    @TTLcolumns = @reset_TTLcolumns;
    $Ncols = $reset_Ncols;
    $STROW = $USTROW = $FSTROW = 0;
}

sub sort_menutable {
    my ($column,$direction) = @_;

    my $offset = 2;
    $column += $offset-1;
    my @hdr = shift @Menutable;

    my $type = "a";
    my $value = $Menutable[0][$column];
    $value =~ s/\s*//g;
    if($value =~ m/^-?(\d+(\.\d+)?)$/) {
	$type = "n";
    }

    my @MT;
    if($direction =~ m/[aA]/) { #asc
	if($type eq "a") { #alpha
	    @MT = sort {
		@{$a}[$column] cmp @{$b}[$column];
	    } @Menutable;
	} else {
	    @MT = sort {
		@{$a}[$column] <=> @{$b}[$column];
	    } @Menutable;
	}
    } else { # desc
	if($type eq "a") { #alpha
	    @MT = sort {
		@{$b}[$column] cmp @{$a}[$column];
	    } @Menutable;
	} else {
	    @MT = sort {
		@{$b}[$column] <=> @{$a}[$column];
	    } @Menutable;
	}
    }

    @Menutable = @MT;
    unshift @Menutable, @hdr;
}

sub show_menutable {
    my $t = shift;
    my $tbl = shift;
    if(uc($t) eq FINALTBL) { #final table
	@UMenutable = @Menutable;
	$Ucols = $Ncols;
	@UTTLcolumns = @TTLcolumns;
	$USTROW = $STROW;
	@Menutable = @FMenutable;
	$Ncols = $FNcols;
	@TTLcolumns = @FTTLcolumns;
	$STROW = $FSTROW;
	$tbl = FINALTBL;
    } elsif (uc($t) eq USERTBL) { #inicial/origin table
	if($tbl eq FINALTBL) {
	    @FMenutable = @Menutable;
	    $FNcols = $Ncols;
	    @FTTLcolumns = @TTLcolumns;
	    $FSTROW = $STROW;
	}
	@Menutable = @UMenutable;
	$Ncols = $Ucols;
	@TTLcolumns = @UTTLcolumns;
	$STROW = $USTROW;
	$tbl = USERTBL;
    }
    return $tbl;
}


sub markrows_menutable {
    my $range = shift @_;

    my @range = getRange($range);
    my $cnt = 0;

    if(@range == scalar(@Menutable) -1) { #whole table -> mark
	foreach $_ (@Menutable) {
	    next if($_->[1] =~ m/C0/);
	    if($_->[0] =~ m/\[\s\]/) {
		$_->[0] =~ s/\[\s\]/\[x\]/;
		$cnt++;
	    } else {
		$_->[0] =~ s/\[x\]/\[ \]/;
		$cnt++;
	    }
	}

    } else { #mark \w+
	foreach my $n (@range) {
	    foreach $_ (@Menutable) {
		if($_->[1] =~ m/^$n\./) {
		    if($_->[0] =~ m/\[\s\]/) {
			$_->[0] =~ s/\[\s\]/\[x\]/;
			$cnt++;
			last;
		    } else {
			$_->[0] =~ s/\[x\]/\[ \]/;
			$cnt++;
			last;
		    }
		}
	    }
	}
    }
    return $cnt;
}

sub markbyval_menutable {
    my $findv = shift;
    my $column = shift;

    my $offset = 2;
    $column += $offset-1;
    my $cnt = 0;

    foreach (@Menutable) {
	if($_->[$column] =~ m/$findv/) {
	    if($_->[0] =~ m/\[\s\]/) {
		$_->[0] =~ s/\[\s\]/\[x\]/;
		$cnt++;
	    } else {
		$_->[0] =~ s/\[x\]/\[ \]/;
		$cnt++;
	    }
	}
    }
    return $cnt;
}

sub markfst_menutable {
    my $maxmatches = shift;
    my $column = shift;
    my $fornlines = shift;

    my $cnt = 0;
    my $offset = 2;
    $column += $offset-1;

    my @hdr = shift @Menutable;

    my %mymatches; #memory; which values have been matched, to ensure singletons
    my $once = 0;
    my $findv = $Menutable[0][$column];
    foreach (@Menutable) {
	if($findv ne $_->[$column] && !exists $mymatches{$_->[$column]}) {
	    $findv = $_->[$column]; #new value
	    $once = 0;
	}
	if(!$once) {
	    if($_->[$column] =~ m/$findv/) {
		if($_->[0] =~ m/\[\s\]/) {
		    $_->[0] =~ s/\[\s\]/\[x\]/;
		    $cnt++;
		} else {
		    $_->[0] =~ s/\[x\]/\[ \]/;
		    $cnt++;
		}
	    }
	    $once = 1;
	    $mymatches{$findv} = 1;
	    $maxmatches--;
	}
	last unless($maxmatches);
	last unless($fornlines);
	$fornlines--;
    }
    unshift @Menutable, @hdr;
    return $cnt;
}


sub isMenutable_marked {
    my @check = grep {$_->[0] =~ m/\[x\]/} @Menutable;
    return scalar(@check);
}

sub mrowsup_menutable {
    my $mvpos = shift ;

    my $many = isMenutable_marked();
    for(my $i = 1 ; $i < @Menutable ;) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) {
	    return if($i == 1);
	    my @row = splice @Menutable, $i, 1;
	    my $to = $i-$mvpos-1;
	    $to = 0 if ($to <= 0);
	    my $from = $i-$mvpos;
	    $from = 1 if ($from <= 0);
	    @Menutable = (@Menutable[0..$to],@row,@Menutable[$from..$#Menutable]);
	    $Menutable[$i-$mvpos]->[0] =~ s/\[x\]/\[ \]/; #comment to disable autom deselection after moving
	} else {
	    $i++;
	}
    }
}

sub mrowsdown_menutable {
    my $mvpos = shift ;

    my $s = @Menutable-1;
    for(my $i = $s ; $i >= 1 ;) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) { 
	    return if($i == $s); #already last
	    my @row = splice @Menutable, $i, 1;
	    my $to = $i+$mvpos-1;
	    $to = $s-1 if ($to >= $s);
	    my $from = $i+$mvpos;
	    if ($from > $#Menutable) {
		@Menutable = (@Menutable[0..$to],@row);
	    } else {
		@Menutable = (@Menutable[0..$to],@row,@Menutable[$from..$#Menutable]);
	    }
	    $Menutable[$i+$mvpos]->[0] =~ s/\[x\]/\[ \]/; #comment to disable autom deselection after moving
	} else {
	    $i--;
	}
    }
}

sub delcl_menutable  { 
    my $range = shift @_;

    my @range = getRange($range);
    @range = sort {$b <=> $a} (grep {!/^0$/} @range); #cannot hide/rem column 0
    
    my $colttl = COLTTL_FL;
    shift @Menutable;
    foreach my $col (@range) { #remove from last to first
	my $val = delete $TTLHDR{"C".$col};
	if($val eq $colttl) {
	    my $many = map {$_ =~ /$colttl/} values %TTLHDR; 
	    @TTLcolumns = grep {!/^$colttl$/} @TTLcolumns if(!$many); #removes only if there is no File word in values
	} else { 
	    @TTLcolumns = grep {!/^\Q$val\E$/} @TTLcolumns;
	}

	foreach $_ (@Menutable) {
	    splice @{$_}, $col+1, 1;
	}
	
	$Ncols--;
    }
}

sub hidecl_menutable {
    my $range = shift ;

    my @range = getRange($range);
    @range = grep {!/^0$/} @range; #cannot hide/rem column 0
    my @hidden = map {$_ += 2 -1} @range;

    foreach (@hidden) {
	if(exists $HIDDENcols{$_}) { 
	    delete $HIDDENcols{$_}; #removed
	    next;
	} else {
	    $HIDDENcols{$_} = 1; #added
	}
    }
}

sub isColumnHidden {
    my $col = shift;

    if(exists $HIDDENcols{$col}) {
	return 1;
    }
    return 0;
}

sub colRangeOffLimits {
    my $range = shift ;
    
    my @range = getRange($range);
    if($range[0] < 1 || $range[$#range] > $Ncols) {
	return 1;
    }
    return 0;
}

sub getRange {
    my $range = shift ;

    my @range;
    if($range =~ m/^\d+$/) {
	@range = ($range);
    } elsif ($range =~ m/(\d+)-(\d+)/) {
	@range = ($2..$1);
	@range = ($1..$2) if($1 < $2);
    } elsif ($range =~ m/\d+(,\d+)*/) {
	@range = split(/,/,$range);
    }

    return sort {$a <=> $b} @range;
}

sub delrs_menutable {
    
    for(my $i = 1 ; $i < @Menutable ;) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) { 
	    splice @Menutable, $i, 1;
	} else {
	    $i++;
	}
    }
}

sub copyfrom_menutable {
    my $numcopies = shift;
    for(my $i = 1 ; $i < @Menutable ; $i++) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) { 
	    push @FMenutable, updateIndex($_,$Menutable[$i]) foreach (0..$numcopies-1);
	}
    }

    unshift @FMenutable, [@{$Menutable[0]}] if($FMenutable[0]->[1] !~ m/C0/);
    $FNcols = $Ncols;
}

sub updateIndex {
    my $ri = shift;
    my @row = @{$_[0]};

    my $ori = $row[1];
    $ori =~ s/\s+//g;

    my @lofcpy = grep {$_->[1] =~ m/\Q$ori\E/} @FMenutable;

    my $hIndex = 0;
    foreach (@lofcpy) {
    	my $i = $_->[1];
    	$i =~ s/\s+//g;
	($_,$i) = split(/\./,$i,2);
	$i =~ s//0/;
    	if($i > $hIndex) {
    	    $hIndex = $i;
    	}
    }
    $ri = ++$hIndex;

    if($ri) { #updates indexes on copies as x.1 x.2 etc... 0 means original
	$row[1] = $ori.$ri;
    }

    return [@row];
}

sub movefrom_menutable {
    for(my $i = 1 ; $i < @Menutable ;) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) { 
	    my @row = splice @Menutable, $i, 1;
	    push @FMenutable, @row;
	} else {
	    $i++;
	}
    }

    unshift @FMenutable, [@{$Menutable[0]}] if($FMenutable[0]->[1] !~ m/C0/);
    $FNcols = $Ncols;
}

sub reset_menutable {
    my @MT = ();
    for(my $i = 0 ; $i < @reset_MT ; $i++) {
	my @row = @{$reset_MT[$i]};
	push @MT, [@row]; #create (new) references
    }

    $Ncols = $reset_Ncols;
    @TTLcolumns = @reset_TTLcolumns;
    $STROW = 0;
    return @MT;
}

sub plrt_fmenutable {
    my $conf = shift;

    my ($cmi, $cli, $cfi);
    $cmi = getColumnIndex(COLTTL_MDL) if(isColumnPresent(COLTTL_MDL));
    $cli = getColumnIndex(COLTTL_LNL) if(isColumnPresent(COLTTL_LNL));
    $cfi = getColumnIndex(COLTTL_FL) if(isColumnPresent(COLTTL_FL));

    if(isMenutable_marked()) {
	plrtmarked_fmenutable($conf,$cmi,$cli,$cfi);
	return;
    }

    for(my $i = 1 ; $i < @Menutable ; $i+=2) {
	if(defined $Menutable[$i]->[$cli+1]) {
	    next if ($Menutable[$i]->[$cli+1] eq CELLCTT_DASH); #from this is assumed that LRT was done for this pair
	}
	my $m1 = $Menutable[$i]->[$cmi]; #null!
	my $m2 = $Menutable[$i+1]->[$cmi]; #alternate!
	next if(!$m2); #pairs!
	
	my $res = calculateLRT($conf,$m1,$m2,[ join(",",@{$Menutable[$i]}[$cfi..$Ncols+1]),join(",",@{$Menutable[$i+1]}[$cfi..$Ncols+1]) ]);
	my @res = split /,/, $res;
	
	#add columns, LRT test, $deltaLnL, $df, $pvalue, $conclusion
	push @{$Menutable[$i]}, (CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH);
	push @{$Menutable[$i+1]}, ("$m1 vs. $m2",@res);
    }

    if(!isColumnPresent(COLTTL_TLRT)) {
	$Ncols+=5;
	push @TTLcolumns, COLTTL_TLRT, COLTTL_DLRT, COLTTL_DF, COLTTL_PV, COLTTL_CCL."($conf)";
    }
}

sub plrtmarked_fmenutable { #intended for branch-models to calc lrt in potentially 3 models
    my ($conf,$cmi,$cli,$cfi) = @_;

    my $cri = getColumnIndex(COLTTL_TLRT);
    my ($fst,$snd) = getpRowIndex(0);    
    while($snd >= 1) {
	my $m1 = $Menutable[$fst]->[$cmi]; #null!
	my $m2 = $Menutable[$snd]->[$cmi]; #alternate!
	if($cri >= 1) {
	    $Ncols-=5;
	    splice @{$Menutable[$fst]}, $Ncols+2, 5;
	    splice @{$Menutable[$snd]}, $Ncols+2, 5;
	}
	
	my $res = calculateLRT($conf,$m1,$m2,[ join(",",@{$Menutable[$fst]}[$cfi..$Ncols+1]),join(",",@{$Menutable[$snd]}[$cfi..$Ncols+1]) ]);
	my @res = split /,/, $res;
	if($cri >= 1) {
	    $Ncols+=5; #put back
	}
	
	push @{$Menutable[$fst]}, (CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH);
	push @{$Menutable[$snd]}, ("$m1 vs. $m2",@res);
	($fst,$snd) = getpRowIndex($snd);
    }
    if(!isColumnPresent(COLTTL_TLRT)) {
	$Ncols+=5;
	push @TTLcolumns, COLTTL_TLRT, COLTTL_DLRT, COLTTL_DF, COLTTL_PV, COLTTL_CCL."($conf)";
    }   
}

#get pair of consecutive rows selected
sub getpRowIndex { 
    my $start = shift;

    my ($i1,$i2);
    for(my $i = $start+1 ; $i < @Menutable ; $i++) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) {
	    $i2 = $i if($i1);
	    $i1 = $i if(!$i2);
	} else {
	    my $last = $#{$Menutable[$i]};
	    push @{$Menutable[$i]}, (CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH,CELLCTT_DASH) if($Menutable[$i]->[$last] !~ m/^[-H10]+$/);
	}
	last if($i1 && $i2);
    }
    return (-1,-1) if(!$i1 && !$i2);
    return ($i1,$i2);
}


sub pbeb_fmenutable {
    if(isColumnPresent(COLTTL_CCL)) { #put considering the conclusion res of the previous lrt test
	my $cci = getColumnIndex(COLTTL_CCL); #conclusion column!
	for(my $i = 1 ; $i < @Menutable ; $i++) {
	    my $k = $Menutable[$i]->[1];
	    $k =~ s/\s+//g;
	    $k =~ s/\.\d+$/\./;
	    if($Menutable[$i]->[$cci] =~ m/H1/) { #if conclusion column in every 2nd row is having H1 as result then put BEB
		push @{$Menutable[$i]}, $BEB{$k} if($BEB{$k}); #alternate!
		push @{$Menutable[$i]}, CELLCTT_NFOUND if(!$BEB{$k}); #alternate!
	    } else {
		push @{$Menutable[$i]}, CELLCTT_DASH; #null or H0
	    }
	}
    } else { #or in every row
	for(my $i = 1 ; $i < @Menutable ; $i++) { 
	    my $k = $Menutable[$i]->[1];
	    $k =~ s/\s+//g; 
	    $k =~ s/\.\d+$/\./;
	    if(exists $BEB{$k}) {
		push @{$Menutable[$i]}, $BEB{$k} if($BEB{$k}); #alternate!
		push @{$Menutable[$i]}, CELLCTT_NFOUND if(!$BEB{$k}); #alternate!
	    } else {
		push @{$Menutable[$i]}, CELLCTT_DASH; #null!
	    }
	}
    }
    $Ncols++;
    push @TTLcolumns, COLTTL_BEB;
}

sub isColumnPresent {
    my $col = shift @_;
    return 0 || grep {/$col/i} @TTLcolumns;
}

sub getColumnIndex {
    my $colname = shift @_;

    my $ci = -1;
    my $max = -1;
    while(my ($k,$v) = each %TTLHDR) { 
	if($v =~ m/$colname/i) {
	    $ci = (split /C/,$k)[1];
	    $ci++;
	}
	$max = $ci if ($ci > $max);
    }
    return $max;
}

sub save_fmenutable {
    my $newfl = shift;

    my $cf = getColumnIndex(COLTTL_FL);
    my $fh;
    if($newfl) {
	$fh = openOUTFLb(join(",",@TTLcolumns)."\n",$newfl,$OUTFILE);
    } else {
	$fh = openOUTFLb(join(",",@TTLcolumns)."\n",0,$OUTFILE);
    }

    my @hdr = shift @Menutable;

    my $cnt;
    my @marked = @Menutable;
    if(isMenutable_marked()) { #if marked save only marked else save all
	@marked = grep {$_->[0] =~ m/\[x\]/} @Menutable;
    }
    $cnt = scalar(@marked);
    foreach my $r (@marked) {
	my @row = @{$r};
	shift @row; #remove marks and indices;
	shift @row;
	my $file = catfile(catdir(@row[0..$cf-2]),"");
	print $fh $file.",".join(",",@row[$cf-1..$#row]),"\n";
    }
    unshift @Menutable, @hdr;
    close $fh;
    return $cnt;
}

sub loadCommandDefinitions {
    my $deffl = shift;

    %UCOMMAND = ();
    my @ctt = readac($deffl||$DEFS_FILE);
    chomp @ctt;
    my $entry;
    foreach (@ctt) {
	if($entry eq "cmd") {
	    my ($k,$v) = split(/=/,$_);
	    $UCOMMAND{$k} = $v;
	}
	if($_ eq "COMMANDS") {
	    $entry = "cmd";
	}
    }
}

sub saveCommandDefinitions {
    my $outfl = shift;

    my $fh = openOUTFLb("COMMANDS\n",$outfl||$DEFS_FILE,$OUTFILE);
    while(my ($k,$v) = each %UCOMMAND) {
	print $fh $k."=".$v."\n";
    }
    close $fh;
}

sub invalidCmdDefinition { # ALWAYS update below %lists if new commands arise
    my $cmdnm = shift;
    my $cmd = shift;

#Command names not allowed in 'def' command for user command alias (25)
    my %ALL_COMMANDS = (sort => 1, show => 1, delcl => 1, mark => 1, markf => 1, findr => 1, copy => 1, move => 1, reset => 1, quit => 1, "q" => 1,
			open => 1, ls => 1, hide => 1, fh => 1, mrup => 1, mrdn => 1, delrs => 1, plrt => 1, pbeb => 1, save => 1, empty => 1,
			def => 1, defdel => 1, defload => 1, defsave => 1);

    if(exists $ALL_COMMANDS{$cmdnm}) {
	return "Command name '$cmdnm' not allowed: already a predefined word!";
    }

#Commands allowed in 'def' command for user command alias (15)
    my %ALWD_COMMANDS = (sort => 1, show => 1, delcl => 1, mark => 1, markf => 1, findr => 1, copy => 1, move => 1, 
			 mrup => 1 , mrdn => 1, delrs => 1, plrt => 1, pbeb => 1, save => 1, empty => 1);
    
    my @subcmds = split(/;/,$cmd);
    foreach (@subcmds) {
	$_ = substr $_, 0, index($_," ") if($_ =~ m/\s+/);
	if(!exists $ALWD_COMMANDS{$_}) {
	    return "Command '$_' not allowed in user definition!";
	}
    }
    return;
}


sub print_menutable {
    my $howmany = shift @_;
    my $showttl = shift @_;
    my %sizes = %{$_[0]};

    clear(0);

    my $prefSize = DCOLSZ;
    my $coltxt;

    print $TCOLOR{underln}," FILE ",$TCOLOR{endcolor},": $INFILE ";
    print $TCOLOR{underln}," SAVE TO ",$TCOLOR{endcolor},": $OUTFILE\n\n";

    my @hdr;
    if($showttl) {
	@hdr = printFileHeaders(1,[0]);
	%sizes = calcColsizes();
    }

    print "   ...\n" unless($STROW == 0);
    for(my $i = $STROW || 0; $i < @Menutable ; $i++) {
	showTableColors(1,$i);
	for(my $j =0; $j < @{$Menutable[$i]}; $j++) {
	    next if(isColumnHidden($j));
	    my $maxlen = $sizes{$j};
	    my $len = length($Menutable[$i][$j]);
	    if($maxlen > $prefSize) {
		$maxlen = $prefSize+3+2;
		$coltxt = substr($Menutable[$i][$j],0,$prefSize);
		$coltxt .= "..." if($i != 0 && $len >= $maxlen);
		my $format = "%"."-$maxlen"."s";
		printf($format,$coltxt);
	    } else {
		$maxlen += 2;
		my $format = "%"."-$maxlen"."s";
		printf($format,$Menutable[$i][$j]);
	    }
	}
	showTableColors(0,$i);
	print "\n";
	last if($i == $howmany+$STROW);
    }
    if($showttl) {
	printFileHeaders(0,\@hdr);
    }
    print "   ...\n" unless($howmany+$STROW >= @Menutable-1);
    printStatusBars($howmany);
}

sub showTableColors {
    my $stend = shift;
    my $i = shift;
    if($stend) {
	if($Menutable[$i]->[0] =~ m/\[x\]/) { 
	    print $TCOLOR{bggreen};
	} else {
	    print $TCOLOR{Bred} if($i == 0); 
	    print $TCOLOR{bgwhite} if($i != 0 && $i % 2 ==0); 
	}
    } else {
	print $TCOLOR{endcolor};
    }
}

sub printFileHeaders {
    my $onoff = shift ;
    my @hdr = @{$_[0]};

    if($onoff) {
	my @nhdr = @{shift @Menutable};
	my @ttlhdr;
	my $fst = shift @nhdr;
	foreach (@nhdr) {
	    push @ttlhdr, $TTLHDR{$_};
	}
	unshift @Menutable, [COLTTL_MK,@ttlhdr];
	unshift @nhdr, $fst;
	return @nhdr;
    } else {
	shift @Menutable;
	unshift @Menutable, [@hdr];
    }
    return 0;
}

sub printStatusBars {
    my $howmany = shift;
    my $status = "[".(scalar(@Menutable)-1)."] rows, [".scalar(keys %TTLHDR)."] columns, [".isMenutable_marked()."] rows marked, [$howmany] rows showing, [".scalar(keys %HIDDENcols)."] columns hidden\n";

    print $TCOLOR{menu},"="x(length($status)),"\n",$TCOLOR{endcolor};
    print $TCOLOR{Bcyan},$status,$TCOLOR{endcolor};
    print $TCOLOR{menu},"="x(length($status)),"\n",$TCOLOR{endcolor};

    my $msg = "SCROLLING  ".$TCOLOR{red}."ON".$TCOLOR{endcolor}.$TCOLOR{bold}.": Ctrl+d , ".$TCOLOR{red}."OFF".$TCOLOR{endcolor}.$TCOLOR{bold}.": Ctrl+\\ \nScroll Keys: Arrows [UP / DOWN], Top: [HOME], Bottom: [END] ";
    print $TCOLOR{bold},$msg,"\n",$TCOLOR{endcolor};
    print $TCOLOR{menu},"="x(65),"\n",$TCOLOR{endcolor};
}

sub printPrompt {
    my ($msg, $ph, $tbl, $pcnt) = @_;
    if($msg && $ph) {
	$msg = " \{".$TCOLOR{cyan}.$msg.$TCOLOR{endcolor}.$TCOLOR{bold}."\}";
    } else {
	$msg = $TCOLOR{bold};
    }
    my $tblp = $TCOLOR{red}.$tbl.$TCOLOR{endcolor}.$TCOLOR{bold};
    print "( type '?' for help ; 'q' to quit )\n";
    print $TCOLOR{bold},"[$pcnt $tblp OMAP$msg PROMPT\]\$ ",$TCOLOR{endcolor};
}

sub calculateLRT {
    my $confidence = shift @_;
    my $m0 = shift @_;
    my $m1 = shift @_;
    my @pairmodels = @{$_[0]};

    my $res = "";
    my ($mdl0, $mdl1, $np0, $np1, $lnl0, $lnl1);
    if($pairmodels[0] =~ m/,($m0),(\d+),.*,(-?\d+\.\d+)$/) { #null
	$mdl0 = $1;
	$np0 = $2;
	$lnl0 = $3;
    }
    if($pairmodels[1] =~ m/,($m1),(\d+),.*,(-?\d+\.\d+)$/) { #alternate
	$mdl1 = $1;
	$np1 = $2;
	$lnl1 = $3;
    }

    my $df = $np1 - $np0;
    my $deltaLnL = 2 * ($lnl1 - $lnl0);
    if($deltaLnL < 0) {
	$deltaLnL *= -1;
    }

    my $pvalue;
    my $criticval;
    my $conclusion = "H0";
    if(eval("require Math::Cephes;")) { #most precise
	eval {$pvalue = chdtrc($df,$deltaLnL);};
	warn "WARNING:: Error in chdtrc function(Math::Cephes):: $@ !\n" if($@); # Complemented Chi-square distribution
	$pvalue = 0.0 if($@ =~ m/underflow/);
	eval {$criticval = chdtri($df,$confidence);};
	warn "WARNING:: Error in chdtri function(Math::Cephes):: $@ !\n" if($@); # Inverse of complemented Chi-square distribution
	if($deltaLnL > $criticval) {
	    $conclusion = "H1";
	}
	return "$deltaLnL,$df,$pvalue,$conclusion";
    } elsif(eval("require Statistics::Distributions;")) { #less precise <=> chi PAML binary
	eval {$pvalue = Statistics::Distributions::chisqrprob($df,$deltaLnL);};
	warn "WARNING:: Error in chisqrprob function(Statistics::Distributions::chisqrprob):: $@ !\n" if($@); # Complemented Chi-square distribution
	$pvalue = 0.0 if($@ =~ m/underflow/);
	eval {$criticval = Statistics::Distributions::chisqrdistr($df,$confidence);};
	warn "WARNING:: Error in chisqrdistr function(Statistics::Distributions::chisqrdistr:: $@ !\n" if($@); # Inverse of complemented Chi-square distribution
	if($deltaLnL > $criticval) {
	    $conclusion = "H1";
	}
	return "$deltaLnL,$df,$pvalue,$conclusion";
    }
    warn "WARNING:: Could not find either one of the following modules installed: Math::Cephes and/or Statistics::Distributions : $@ !\n" if($@);
    return "$deltaLnL,$df,$pvalue,".CELLCTT_NAVAIL;
}

sub calcColsizes {
    my %sizes;

    $sizes{$_} = 0 foreach (0..scalar(@{$Menutable[0]}));

    for(my $i = 0; $i < @Menutable ; $i++) {
	
	for(my $j = 0; $j < @{$Menutable[$i]} ; $j++) {
	    my $len = length($Menutable[$i][$j]);
	    $sizes{$j} = $len if($sizes{$j} < $len);
	}
    }
    return %sizes;
}

sub dumpPromptHelp {
    %TCOLOR = disableTCOLOR();
    open(FOUT, ">", catfile(".",PHELPFILE));
    select(FOUT); #switch from default STDOUT to FOUT
    promptHelp(); #print to FOUT
    close FOUT;
    select(STDOUT); #switch back to STDOUT
    %TCOLOR = enableTCOLOR();
    print ":: Prompt help printed to file ",catfile(".",PHELPFILE),"\n";
    exit 1;
}

sub promptHelp {
    clear(0) if(select() !~ m/FOUT/);
    printLicense($0);
    print "\t",$TCOLOR{menu},"OMAP PROMPT HELP - COMMANDS\n";
    print "="x80,$TCOLOR{endcolor},"\n\n";
    print $TCOLOR{bold},"Options in [ ] - means, options are not mandatory; | - means, alternative options; ( ) - means, grouping",$TCOLOR{endcolor},"\n\n";
    printf("%s%-10s %-15s %-20s%s\n", $TCOLOR{bold},"Command","Options","Description",$TCOLOR{endcolor});
    printf("%-10s %-15s %-20s\n", "==========","===============","======================");
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"open",$TCOLOR{endcolor},$TCOLOR{yellow},"filename.csv","(Only in User Table! see command show) Open CSV file, discarding any current values.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"ls",$TCOLOR{endcolor},$TCOLOR{eopen},"[path|dir]","List contents on given directory. If no option is given, lists current directory.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"reset",$TCOLOR{endcolor},$TCOLOR{yellow},"","Reset contents in User Table to initial contents.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"delcl(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"C(a|a-z|a,d,h)","Remove column(s). Either a single column or a range of columns specififed by from-to values or any columns comma-separated.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"hide",$TCOLOR{endcolor},$TCOLOR{yellow},"C(a|a-z|a,d,h)","(Un)Hide column(s). Either a single column or a range of columns specififed by from-to values or any columns comma-separated.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"sort(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"Cn (A|a)|(D|d)","Sort column C[number] [A|a]scending or [D|d]escending. This is performed according to its contents type.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"show(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"(n|t)|(n t)","Show i) (number of rows n) or (Table t) or ii) number of rows (n) in Table (t). (t = [U|u]ser or [F|f]inal).",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"mark(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"[a|a-z|a,d,h]","(Un)Mark rows. Either a single row, a range of rows specififed by from-to values or any rows comma-separated;",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","If no option is given, the selection will (un)mark all rows in Table.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"markf(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"r Cn [l]","(Un)Mark the first r rows which contain distinct values in column C[number];",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{yellow},"","Search within l rows limit or, by default, in full Table length.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"findr(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"\"value\" Cn","(Un)Mark all rows containing the quoted specified value in column C[number];",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","The characters '\^','\$','\|','\[' and '\]', can be used in the value to mean either 'start', 'end', 'or', or 'set of alternative characters', respectively;",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","eg.: findr \"\^value\" C4 ; findr \"value\$\" C4 ; findr \"val1\|val2\|val3\" C4 ; findr \"\[vV\]alue\" C4.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"copy(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"[n]","(Only in User Table! see command show) Copy any marked rows, n times each, from User Table to Final Table. If not given n, number of copies defaults to 1;",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{yellow},"","Copies are differentiated by 'index.subindex' in column 'Index', 2nd from the left. Use command 'findr' to (un)mark any copies.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"move(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"","(Only in User Table! see command show) Move any marked rows from User Table to Final Table.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"mrup(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"[n]","Move any marked rows n positions up in current Table. If not given n, number of positions defaults to 1.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"mrdn(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"[n]","Move any marked rows n positions down in current Table. If not given n, number of positions defaults to 1.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"delrs(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"","Delete any marked rows in current Table.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"fh",$TCOLOR{endcolor},$TCOLOR{eopen},"","Show file headers in current Table.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"empty(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"","Empty Final Table contents.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"plrt(*)",$TCOLOR{endcolor},$TCOLOR{eopen},"significance","(Only in Final Table! see command show) Perform LRT test for every pair of rows, in Final Table, from top to bottom;",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","The null model should be placed on the row above the row of the alternative model. In case rows are marked, this operation is performed preferencially in the selection.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"pbeb(*)",$TCOLOR{endcolor},$TCOLOR{yellow},"","(Only in Final Table! see command show) Create an additional column containing the BEB information considering the LRT test conclusion if exists;",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{yellow},"","otherwise, BEB will appear in all rows regardless of the LRT conclusion. This column can only be placed, if initially existed in the input file.",$TCOLOR{endcolor});
    printf("%s%-10s%s %-15s %s%-20s%s\n", $TCOLOR{bold},"save[!](*)",$TCOLOR{endcolor},"[file.csv]",$TCOLOR{eopen},"(Only in Final Table! see command show) Save Final Table contents to default output file (from initial option -o) or if specified, to 'file.csv';",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","Append exclamation mark at the end of command, as in 'save\! file.csv', to save contents and make the specified CSV file the default;",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","The contents will be appended to the end of the default file, whenever the file exists; regardless if the file is empty or is a new file;",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","Only marked rows (see command mark, markf, findr), are saved whenever found in Table; or otherwise, all is saved.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"def",$TCOLOR{endcolor},$TCOLOR{yellow},"[CMD[=LIST]]","Define an alias command to perform several operations at once; two other uses can take place:",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{yellow},"","1) The user chooses a command name (CMD) (not found in this HELP) and a LIST of commands, where LIST=(cmd1;...;cmdn), following previous rules for each command;",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{yellow},"","2) Type 'def CMD' to show the definition for the specified command;",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{yellow},"","3) Type 'def' only to show the list of all the defined commands;",$TCOLOR{endcolor});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{lred},"","(*) The commands allowed in the LIST are here marked by an asterisc.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"defsave",$TCOLOR{endcolor},$TCOLOR{eopen},"[file.name]","Save all defined commands to file. File is by default $DEFS_FILE, if none is given.",$TCOLOR{eclose});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"defload",$TCOLOR{endcolor},$TCOLOR{yellow},"[file.name]","Load all defined commands from file. File is by default $DEFS_FILE, if none is given.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"defdel",$TCOLOR{endcolor},$TCOLOR{eopen},"","Delete $DEFS_FILE file.",$TCOLOR{eclose});
    printf("%-10s %s%-15s %-20s%s\n", "",$TCOLOR{eopen},"","This file may contain several definitions including redefinitions of the same command. Hence, the last definitions in file prevail.",$TCOLOR{eclose});
    printf("%-10s %-15s %-20s\n", "----------","---------------","----------------------");
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"q[uit]",$TCOLOR{endcolor},$TCOLOR{yellow},"","Checks if any unsaved data are there in Final Table, before exiting.",$TCOLOR{endcolor});
    printf("%s%-10s%s %s%-15s %-20s%s\n", $TCOLOR{bold},"?",$TCOLOR{endcolor},$TCOLOR{eopen},"","This Help.",$TCOLOR{eclose});
    printf("%-10s %-15s %-20s\n", "==========","===============","======================");

    print "\nVERSION: ".VERSION."\n";
    print $TCOLOR{menu},"="x80,$TCOLOR{endcolor},"\n\n";
    print "Type any command or press \[$TCOLOR{qunln}ENTER$TCOLOR{endcolor}\] to go back to Table.\n\n";
}


sub HELP_MESSAGE {
    print shift;
    my $citation = CITATION;
    my $pweb = PROJWEBSITE;
    my $v = VERSION;
    my $phelp = PHELPFILE;
    my @HELP = <<HELP;

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    omap.pl - Organize CSV files from imap.pl and perform LRT tests.
   
  $TCOLOR{bold}SYNOPSIS:$TCOLOR{endcolor}
    
    omap.pl -i [infile.csv] -o [outfile.csv]

  $TCOLOR{bold}DESCRIPTION:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | This application is intended to interactively organize PAML (codeml) results extracted and 
    | collected into a CSV file with the application imap.pl. These results can be organized
    | through several steps required and performed by the user. omap.pl provides several
    | interactive commands that help the user easily achieve this purpose and to finally
    | perform the necessary LRT tests.
    | Information about the interactive commands is provided interactively, by typing '?'.

  $TCOLOR{bold}OPTIONS:$TCOLOR{endcolor}
    |  ----------------------------------------------------------------------------
    | -i [infile]     CSV results file from imap.pl.
    | -o [outfile]    Choose a name for the resulting CSV file to create or append results.
    |                 -----------------------------------
    | --help          Dump the interactive commands help to a file located in current directory, named '$phelp'.
    | -h              This help
    | -v              Application version

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2014.

  $TCOLOR{bold}CITATION AND DOWNLOAD:$TCOLOR{endcolor}
    
    $citation
    $pweb

  $TCOLOR{bold}LICENSE AND COPYRIGHT:$TCOLOR{endcolor}

    LMAP: Lightweight Multigene Analyses in PAML
    omap.pl is part of LMAP package
    Copyright (C) 2015 Emanuel Maldonado

  omap.pl is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  omap.pl is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with omap.pl.  If not, see <http://www.gnu.org/licenses/>.
   
HELP

print @HELP;
    exit 1;
}

sub VERSION_MESSAGE {
    my $v = VERSION;
    my @VERS = <<VERS;
Use option -h for help and other information.

  $TCOLOR{bold}NAME:$TCOLOR{endcolor}

    omap.pl - Organize CSV files from imap.pl and perform LRT tests.

  $TCOLOR{bold}VERSION:$TCOLOR{endcolor} $v

  $TCOLOR{bold}AUTHOR:$TCOLOR{endcolor}

    Written by Emanuel Maldonado, <emaldonado\@ciimar.up.pt>. Since Mar 2014.

VERS

print @VERS;
    exit 1;
}

